//---------------------------------------------------------------------------
// File:    c:\cbproj\AudioIO\ASIO\ASIO_dummy.c
// Author:  Wolfgang Buescher (DL4YHF)
// Date:    2011-10-31
// Purpose:
//   ASIO-"dummy" (not really an ASIO "driver", but tries to mimick one),
//   half-way compatible with Steinberg's ASIO specification V2 .
//   [ ASIO Specification V2.1 ,     Steinberg Audio Stream I/O API
//               (c) 1997 - 2005, Steinberg Media Technologies GmbH ] .
//
//   This was originally just the C source code to a simple COM object,
//   compiled into an ordinary dynamic link library (DLL).
//     Based on an article titled "COM in plain C" By Jeff Glatt,
//     locally saved as c:/cbproj/COM_in_plain_C/info/com_in_c1.asp.htm .
//   Later used as for the optional ASIO-part in DL4YHF's Audio-I/O-Library,
//    which contains a native DLL interface for Spectrum Lab,
//    an interface to WINAMP, and an ASIO-dummy: cbproj/AudioIO/ASIO/*.c .
//
// Notes:
//  - One of the main problems was that Steinberg decided to use Microsoft's
//    "thiscall" calling convention, even if there's no reason to use it,
//    not even for COM server (which unfortunately is the framework
//    for the interface between an audio-processing application (=host)
//    and the ASIO driver). Writing C++ code with Borland C++Builder,
//    compatible with the 'thiscall' calling convention seemed impossible.
//    Thus the important 'wrapper' (AsioDriverThiscalls_BCB.c) for all functions
//    in an ASIO driver which use this proprietary calling convention .
//  - Some red tape for the COM-interface in the DLL
//    ( DllGetClassObject, DllCanUnloadNow, DllRegisterServer, DllUnregisterServer,
//      and possibly DllMain ) obviously use the __stdcall calling convention.
//    Since all C compilers (Borland, MS, GCC) support it, there's no need
//    for a wrapper written in assembler code.
//  - Other COM-related red tape uses the STDMETHODCALLTYPE calling convention.
//    This also seems to (!) boil down to __stdcall .
//    (for details, read "Simplifying the Concept of COM", which used to be at
//     http://www.codeguru.com/cpp/com-tech/activex/misc/article.php/c5509 )
//    This applies at least to the first three pointers in a COM-object's
//    VTable, which are QueryInterface, AddRef, and Release .
//  - ASIO-specific stuff in the 'ASIO-driver' (which, on a PC, unfortunately
//    is a COM-object, but only suitable to be used in a SINGLE INSTANCE)
//    unfortunately uses Microsoft's proprietary 'thiscall' calling convention,
//    as explained by Ross Bencina in the "IASIOThiscallResolver", see
//     http://www.rossbencina.com/code/iasio-thiscall-resolver .
//  - Sometimes, after recompilation of the DLL project, the ASIO "driver" DLL
//    must be registered again. No idea how, but windoze seems to recognize
//    that the DLL has been tampered with, and refuses to use it again (?) .

#include "switches.h" // project-specific compilation switches,
                      // located in a project specific directory, like
                      // \cbproj\AudioIO\in_AudioIO\SWITCHES.H
                      // (the actual path is defined in the PROJECT OPTIONS)
#ifndef SWI_ASIO_SUPPORTED
# error "Missing definition of SWI_ASIO_SUPPORTED in SWITCHES.H !"
#endif
#if( SWI_ASIO_SUPPORTED ) // defined in SWITCHES.H (project-specific options)

#define L_TEST_THISCALL 0 /* 1=yes, 0=no=normal compilation into a DLL */

#include <initguid.h>  // must(?) be included BEFORE windows.h ?
         // "initguid.h" only #defines INITGUID
         //          and then #includes <guiddef.h> .

#include <windows.h>
#include <objbase.h>
#include <tchar.h>

#include "ASIO_dummy.h"           // header for THIS module ("ASIO-driver-dummy")
#include "ASIO_Register.h"        // strange stuff to register and unregister an ASIO driver
#include "AsioDriverThiscalls.h"  // stupid 'thiscall'-wrappers, written in assembler

#include "aio_libs.h"  // includes all headers required for the IMPLEMENTATION
                       // of an 'audio-I/O-DLL' (not the DLL/host-interface ! )


#if TESTWAVES  // optionally fill 'unconnected' inputs with sinewaves ?
# include <math.h>
# ifndef  C_PI
#  define C_PI 3.141592654
# endif
#endif // TESTWAVES ?



// Purely internal consts+defines (those from Steinberg's "asiosmpl" will be thrown out soon)
static const double twoRaisedTo32 = 4294967296.0;
static const double twoRaisedTo32Reciprocal = 1.0 / 4294967296.0;
static DWORD __stdcall ASIOThread (void *param);
static HANDLE ASIOThreadHandle = 0; // EEEK !
static BOOL   done = FALSE;       // EEEEEK !
// ex: TMyAsioDriver *pTheDriver = NULL;  // we don't need this single-instance-bullshit anymore


// Internal forward references (to keep the less interesting stuff near the end)
BOOL AsioPrivate_inputOpen( TMyAsioDriver *pDriver );
void AsioPrivate_inputClose(TMyAsioDriver *pDriver );
void AsioPrivate_input(TMyAsioDriver *pDriver );
BOOL AsioPrivate_outputOpen(TMyAsioDriver *pDriver );
void AsioPrivate_outputClose(TMyAsioDriver *pDriver );
void AsioPrivate_output(TMyAsioDriver *pDriver );
void AsioPrivate_bufferSwitch( TMyAsioDriver *pDriver );
void AsioPrivate_bufferSwitchX( TMyAsioDriver *pDriver );
ASIOError AsioPrivate_outputReady(TMyAsioDriver *pDriver );
void AsioPrivate_timerOn(TMyAsioDriver *pDriver );
void AsioPrivate_timerOff(TMyAsioDriver *pDriver );
// TEST functions (for the 'dummy driver' which is just a stupid oscillator):
void getNanoSeconds (ASIOTimeStamp* ts);

//---------------------------------------------------------------------------
// Implementation of the dummy-ASIO-driver in "plain C"
// Some of these functions are called through a wrapper
// which converts the Microsoft-specific 'thiscall' into 'stdcall' .
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
void AsioDrv_Construct( TMyAsioDriver *pDriver ) // replaces a contructor (in C++)
  // ex: AsioSample::AsioSample () : AsioDriver ()
  // ...but THIS ONE inherits NOTHING from AsioDriver because it's plain C...
  // Should be called from our IClassFactory object's 'CreateInstance' method, somehow.
{
  long i;
  if( pDriver==NULL )
   { return;
   }

  // Do NOT set the entire TMyAsioDriver-struct to ZERO here (using memset),
  // because that struct also contains the 'dummy VTABLE' which has
  // already been set in classCreateInstance() ! !

  pDriver->nSamplesPerBlock = kAsioSamplesPerBlock;  // number of AUDIO SAMPLE POINTS per block
  pDriver->inputLatency  = pDriver->nSamplesPerBlock;  // typically...
  pDriver->outputLatency = pDriver->nSamplesPerBlock * 2;
    // typically nSamplesPerBlock * 2; try to get 1 by offering direct buffer
    // access, and using asioPostOutput for lower latency
  pDriver->dblSamplePosition = 0;
  pDriver->dblSampleRate = 44100.;
  pDriver->active  = FALSE;
  pDriver->started = FALSE;
  pDriver->timeInfoMode = FALSE;
  pDriver->tcRead = FALSE;
  for (i = 0; i < kNumInputs; i++)
   {
     pDriver->inputBuffers[i] = NULL;
     pDriver->inMap[i] = 0;
#   if TESTWAVES
     pDriver->dblSinewavePhase[i] = 0.0;
#   endif
   }
  for (i = 0; i < kNumOutputs; i++)
   {
     pDriver->outputBuffers[i] = NULL;
     pDriver->outMap[i] = 0;
   }
  pDriver->pCallbacks = NULL;
  pDriver->nActiveInputs = pDriver->nActiveOutputs = 0;
  pDriver->toggle = 0;
}


//---------------------------------------------------------------------------
void AsioDrv_Destruct( TMyAsioDriver *pDriver ) // replaces a destructor (in C++)
  // ex: AsioSample::~AsioSample()
{
  if( pDriver!=NULL )
   { AsioDrv_stop(pDriver);
     AsioPrivate_outputClose(pDriver);
     AsioPrivate_inputClose(pDriver);
     AsioDrv_disposeBuffers(pDriver);
   }
}


//---------------------------------------------------------------------------
ASIOBool  __stdcall AsioDrv_init(TMyAsioDriver * pDriver, void * sysHandle)
{
  /**************************************************************************
    Purpose:
      Initialize the AudioStreamIO.
    Parameter:
      info: pointer to an ASIODriver structure:
        - asioVersion:
            - on input, the host version. *** Note *** this is 0 for earlier asio
            implementations, and the asioMessage callback is implemeted
            only if asioVersion is 2 or greater. sorry but due to a design fault
            the driver doesn't have access to the host version in ASIOInit :-(
            added selector for host (engine) version in the asioMessage callback
            so we're ok from now on.
            - on return, asio implementation version.
              older versions are 1
              if you support this version (namely, ASIO_outputReady() )
              this should be 2 or higher. also see the note in
              ASIO_getTimeStamp() !
        - version: on return, the driver version (format is driver specific)
        - name: on return, a null-terminated string containing the driver's name
        - error message: on return, should contain a user message describing
          the type of error that occured during ASIOInit(), if any.
        - sysRef: platform specific (WB: not sure if sysHandle is older or newer)
    Returns:
      If neither input nor output is present ASE_NotPresent
      will be returned.
      ASE_NoMemory, ASE_HWMalfunction are other possible error conditions
   **************************************************************************/

  if( pDriver==NULL )
   {  return ASIOFalse; // result is 'ASIOBool', not the compiler's native 'bool' !
   }
  // Nobody knows where my 'sysRef' has gone .. now they call is 'sysHandle'
  // ... and Steinberg's own ASIO driver demo ignores it, too. So do we:
  (void)sysHandle;

  if (pDriver->active)
   {  return ASIOTrue;
   }
  memset( pDriver->sz123ErrorMessage, 0, 124/*sizeof TMyAsioDriver::sz123ErrorMessage)*/ );

  // Note: Isn't it a bit strange that the ASIO driver's "init" method
  //       already opens the "input" and the "output",
  //       long before the application (host) has created the audio buffers ?
  if( AsioPrivate_inputOpen(pDriver) )
   {
     if( AsioPrivate_outputOpen(pDriver))
      {
         pDriver->active = TRUE;
         return ASIOTrue;
      }
   }
  AsioPrivate_timerOff(pDriver);        // de-activate 'hardware'
  AsioPrivate_outputClose(pDriver);
  AsioPrivate_inputClose(pDriver);
  strcpy( pDriver->sz123ErrorMessage, "ASIO Driver open Failure!");
  return ASIOFalse;
}

//---------------------------------------------------------------------------
void __stdcall AsioDrv_getDriverName(TMyAsioDriver * pDriver,
                      char *name) // [out] max 32 bytes incl. terminating zero
{
  if( name!=NULL )
   { // ex: strcpy (name, "Sample ASIO");  // too bad .
     // Note that this is NOT the name which appears in the ASIO driver list !
     // From Steinberg's documentation it's not clear WHAT EXACTLY this function should return.
     // Obviously NOT the name of the DLL ("driver") but something more descriptive.
     strcpy( name, MY_REGISTRY_ASIO_KEY);    // max 31 chars !
     // -> something like "ASIO Sample Driver", see asiosmpl.h
     // (at least, this not such a stupid 'phantasy name' : THIS ONE exists in the registry,
     //  for example : HKEY_LOCAL_MACHINE.SOFTWARE.ASIO."ASIO Sample Driver" = KEY name)
     // return ASE_OK;   // Steinberg's (not COM's!!) way of saying "all ok"
     // Nope, Steinberg decided that getDriverName does NOT return a value !
   }
  // return ASE_InvalidParameter; // see Steinberg's error codes in ASIO.H
}

//---------------------------------------------------------------------------
long __stdcall AsioDrv_getDriverVersion(TMyAsioDriver * pDriver)
{
  // ?? return 1; // This is what Steinberg used in 'their' example from ASIO SDK V2.1!
  // But shouldn't this be "currently 2", as in ASIO.H : struct ASIODriverInfo.asioVersion ?
  return 2; // GUESSWORK :  Since we support the 'new' buffer-switch with timecode,
            //  we let the host know we support ASIO V2 .
}

//---------------------------------------------------------------------------
void __stdcall AsioDrv_getErrorMessage(TMyAsioDriver * pDriver,
                  char *string) // out: max. 123 characters, plus trailing zero
{
  if( pDriver!=NULL )
   { strcpy( string, pDriver->sz123ErrorMessage );
   }
  else
   { strcpy( string, "Error: THIS==NULL !");
   }
}

//---------------------------------------------------------------------------
ASIOError __stdcall AsioDrv_start(TMyAsioDriver * pDriver)
{
  /**************************************************************************
   Purpose (of ASIOStart) :
      Start input and output processing synchronously.
      This will
      - reset the sample counter to zero
      - start the hardware (both input and output)
        The first call to the hosts' bufferSwitch(index == 0) then tells
        the host to read from input buffer A (index 0), and start
        processing to output buffer A while output buffer B (which
        has been filled by the host prior to calling ASIOStart())
        is possibly sounding (see also ASIOGetLatencies())
    Parameter:
      None.
    Returns:
      If neither input nor output is present, ASE_NotPresent
      will be returned.
      If the hardware fails to start, ASE_HWMalfunction will be returned.
    Notes:
      There is no restriction on the time that ASIOStart() takes
      to perform (that is, it is not considered a realtime trigger).
   **************************************************************************/

  if( pDriver->pCallbacks!=NULL )
   {
     pDriver->started = FALSE;
     pDriver->dblSamplePosition = 0;
     pDriver->theSystemTime.lo = pDriver->theSystemTime.hi = 0;
     pDriver->toggle = 0;

     AsioPrivate_timerOn(pDriver);         // activate 'hardware'
     pDriver->started = TRUE;
     return ASE_OK; // Steinberg's (not COM's!) way to say "all ok",
                    // but only if the result-type is ASIOError (not ASIOBool)
   }
  return ASE_NotPresent;
}

//---------------------------------------------------------------------------
ASIOError __stdcall AsioDrv_stop(TMyAsioDriver * pDriver)
{
  /**************************************************************************
   Quote from a sourcecode floating around in the internet (from SDK V2.1 ?) :
   Purpose (of ASIOStop) :
      Stops input and output processing altogether.
    Parameter:
      None.
    Returns:
      If neither input nor output is present ASE_NotPresent
      will be returned.
    Notes:
      On return from ASIOStop(), the driver must in no
      case call the hosts' bufferSwitch() routine.
   **************************************************************************/

  pDriver->started = FALSE;
  AsioPrivate_timerOff(pDriver);        // de-activate 'hardware'
  return ASE_OK;

}

//---------------------------------------------------------------------------
ASIOError __stdcall AsioDrv_getChannels(TMyAsioDriver * pDriver, long *numInputChannels, long *numOutputChannels)
{
  /**************************************************************************
   Quote from a sourcecode floating around in the internet (from SDK V2.1 ?) :
   Purpose (of ASIOGetChannels) :
      Returns number of individual input/output channels.
    Parameter:
      numInputChannels will hold the number of available input channels
      numOutputChannels will hold the number of available output channels
    Returns:
      If no input/output is present ASE_NotPresent will be returned.
      If only inputs, or only outputs are available, the according
      other parameter will be zero, and ASE_OK is returned.
   **************************************************************************/

  if( numInputChannels!=NULL )
   { *numInputChannels = kNumInputs;
   }
  if( numOutputChannels!=NULL)
   { *numOutputChannels = kNumOutputs;
   }
  return ASE_OK;   // Steinberg's (not COM's!!) way of saying "all ok"
}

//---------------------------------------------------------------------------
ASIOError __stdcall AsioDrv_getLatencies(TMyAsioDriver * pDriver,
        long *inputLatency,  // [out] 'age' of the first sample frame
                             //        in the input buffer when the hosts reads it
        long *outputLatency) // [out] time between the buffer switch,
              //  and the time when the next play buffer will start to sound
{
  /**************************************************************************
   Quote from a sourcecode floating around in the internet (from SDK V2.1 ?) :
   Purpose (of ASIOGetLatencies) :
      Returns the input and output latencies. This includes
      device specific delays, like FIFOs etc.
    Parameter:
      inputLatency will hold the 'age' of the first sample frame
      in the input buffer when the hosts reads it in bufferSwitch()
      (this is theoretical, meaning it does not include the overhead
      and delay between the actual physical switch, and the time
      when bufferSitch() enters).
      This will usually be the size of one block in sample frames, plus
      device specific latencies.

      outputLatency will specify the time between the buffer switch,
      and the time when the next play buffer will start to sound.
      The next play buffer is defined as the one the host starts
      processing after (or at) bufferSwitch(), indicated by the
      index parameter (0 for buffer A, 1 for buffer B).
      It will usually be either one block, if the host writes directly
      to a dma buffer, or two or more blocks if the buffer is 'latched' by
      the driver. As an example, on ASIOStart(), the host will have filled
      the play buffer at index 1 already; when it gets the callback (with
      the parameter index == 0), this tells it to read from the input
      buffer 0, and start to fill the play buffer 0 (assuming that now
      play buffer 1 is already sounding). In this case, the output
      latency is one block. If the driver decides to copy buffer 1
      at that time, and pass it to the hardware at the next slot (which
      is most commonly done, but should be avoided), the output latency
      becomes two blocks instead, resulting in a total i/o latency of at least
      3 blocks. As memory access is the main bottleneck in native dsp processing,
      and to acheive less latency, it is highly recommended to try to avoid
      copying (this is also why the driver is the owner of the buffers). To
      summarize, the minimum i/o latency can be acheived if the input buffer
      is processed by the host into the output buffer which will physically
      start to sound on the next time slice. Also note that the host expects
      the bufferSwitch() callback to be accessed for each time slice in order
      to retain sync, possibly recursively; if it fails to process a block in
      time, it will suspend its operation for some time in order to recover.
    Returns:
      If no input/output is present ASE_NotPresent will be returned.
   **************************************************************************/

  return 0;
}

//---------------------------------------------------------------------------
ASIOError __stdcall AsioDrv_getBufferSize(TMyAsioDriver * pDriver,
         long *minSize,       // [out] the minimum buffer size
         long *maxSize,       // [out] the maxium possible buffer size
         long *preferredSize, // [out] size which best fits performance and hardware requirements
         long *granularity)   // [out] granularity at which buffer sizes may differ
{
  /**************************************************************************
   Purpose (of ASIOGetBufferSize) :
      Returns min, max, and preferred buffer sizes for input/output
    Parameter:
      minSize will hold the minimum buffer size
      maxSize will hold the maxium possible buffer size
      preferredSize will hold the preferred buffer size (a size which
      best fits performance and hardware requirements)
      granularity will hold the granularity at which buffer sizes
      may differ. Usually, the buffer size will be a power of 2;
      in this case, granularity will hold -1 on return, signalling
      possible buffer sizes starting from minSize, increased in
      powers of 2 up to maxSize.
    Returns:
      If no input/output is present ASE_NotPresent will be returned.
    Notes:
      When minimum and maximum buffer size are equal,
      the preferred buffer size has to be the same value as well; granularity
      should be 0 in this case.
   **************************************************************************/

# if(L_TEST_THISCALL) // to test the 'thiscall' calling-convention-converter with DebugTestMain.c :
    *minSize      = 1;
    *maxSize      = 2;
    *preferredSize= 3;
    *granularity  = 4;
# else // normal compilation / "functional" ASIO dummy driver :
    *minSize = *maxSize = *preferredSize = pDriver->nSamplesPerBlock; // allow this size only
    *granularity = 0;  // .. and nothing else !
# endif
    return ASE_OK;
}

//---------------------------------------------------------------------------
ASIOError __stdcall AsioDrv_canSampleRate(TMyAsioDriver * pDriver, ASIOSampleRate sampleRate)
{
  /**************************************************************************
   Quote from a sourcecode floating around in the internet (from SDK V2.1 ?) :
   Purpose (of ASIOCanSampleRate) :
      Inquires the hardware for the available sample rates.
    Parameter:
      sampleRate is the rate in question.
    Returns:
      If the inquired sample rate is not supported, ASE_NoClock will be returned.
      If no input/output is present ASE_NotPresent will be returned.
   **************************************************************************/

  switch( (long)(sampleRate+0.5) )
   { case   8000:
     case  11025:
     case  16000:
     case  22050:
     case  32000:
     case  44100:
     case  48000:
     case  96000:
     case 192000:
        return ASE_OK;
     default:
        break;
   }
  return ASE_NoClock; // Steinberg's way to say "sample rate NOT supported"

} // end AsioDrv_canSampleRate()

//---------------------------------------------------------------------------
ASIOError __stdcall AsioDrv_getSampleRate(TMyAsioDriver * pDriver, ASIOSampleRate *sampleRate)
{
  /**************************************************************************
   Quote from a sourcecode floating around in the internet (from SDK V2.1 ?) :
   Purpose (of ASIOGetSampleRate) :
      Get the current sample Rate.
    Parameter:
      currentRate (/"sampleRate"?) will hold the current sample rate on return.
    Returns:
      If sample rate is unknown, sampleRate will be 0 and ASE_NoClock will be returned.
      If no input/output is present ASE_NotPresent will be returned.
   **************************************************************************/

  // half-hearted implementation, ignores the spec quoted above:
  if( pDriver != NULL )
   { *sampleRate = pDriver->dblSampleRate;
     return ASE_OK;
   }

  return ASE_NoClock; // Steinway's berg to say "sample rate UNKNOWN"
}

//---------------------------------------------------------------------------
ASIOError __stdcall AsioDrv_setSampleRate(TMyAsioDriver * pDriver, ASIOSampleRate sampleRate)
{
  /**************************************************************************
   Quote from a sourcecode floating around in the internet (from SDK V2.1 ?) :
   Purpose (of ASIOSetSampleRate) :
      Set the hardware to the requested sample Rate. If sampleRate == 0,
      enable external sync.
    Parameter:
      sampleRate: on input, the requested rate
    Returns:
      If sampleRate is unknown ASE_NoClock will be returned.
      If the current clock is external, and sampleRate is != 0,
      ASE_InvalidMode will be returned
      If no input/output is present ASE_NotPresent will be returned.
   **************************************************************************/
  ASIOError ae = AsioDrv_canSampleRate(pDriver,sampleRate);
  if( ae != ASE_OK )
   { return ae;
   }
  if (sampleRate != pDriver->dblSampleRate)
   {
     pDriver->dblSampleRate = sampleRate;
     pDriver->asioTime.timeInfo.sampleRate = sampleRate;
     pDriver->asioTime.timeInfo.flags |= kSampleRateChanged;
     if( (pDriver->pCallbacks!=NULL) && pDriver->pCallbacks->sampleRateDidChange )
      {   pDriver->pCallbacks->sampleRateDidChange (pDriver->dblSampleRate);
      }
   }
  return ASE_OK;

}

//---------------------------------------------------------------------------
ASIOError __stdcall AsioDrv_getClockSources(TMyAsioDriver * pDriver, ASIOClockSource *clocks, long *numSources)
{
  /**************************************************************************
   Quote from a sourcecode floating around in the internet (from SDK V2.1 ?) :
   Purpose (of ASIOGetClockSources) :
      Get the available external audio clock sources
    Parameter:
      clocks points to an array of ASIOClockSource structures:
        - index: this is used to identify the clock source
          when ASIOSetClockSource() is accessed, should be
          an index counting from zero
        - associatedInputChannel: the first channel of an associated
          input group, if any.
        - associatedGroup: the group index of that channel.
          groups of channels are defined to seperate for
          instance analog, S/PDIF, AES/EBU, ADAT connectors etc,
          when present simultaniously. Note that associated channel
          is enumerated according to numInputs/numOutputs, means it
          is independant from a group (see also ASIOGetChannelInfo())
          inputs are associated to a clock if the physical connection
          transfers both data and clock (like S/PDIF, AES/EBU, or
          ADAT inputs). if there is no input channel associated with
          the clock source (like Word Clock, or internal oscillator), both
          associatedChannel and associatedGroup should be set to -1.
        - isCurrentSource: on exit, ASIOTrue if this is the current clock
          source, ASIOFalse else
        - name: a null-terminated string for user selection of the available sources.
      numSources:
          on input: the number of allocated array members
          on output: the number of available clock sources, at least
          1 (internal clock generator).
    Returns:
      If no input/output is present ASE_NotPresent will be returned.
   **************************************************************************/

  clocks->index = 0;
  clocks->associatedChannel = -1;
  clocks->associatedGroup = -1;
  clocks->isCurrentSource = ASIOTrue;
  strcpy( clocks->name, "Internal" );
  *numSources = 1;
  return ASE_OK;
}

//---------------------------------------------------------------------------
ASIOError __stdcall AsioDrv_setClockSource(TMyAsioDriver * pDriver,
           long iClockSourceIndex ) // [in], once called "reference"
{
  /**************************************************************************
   Quote from a sourcecode floating around in the internet (from SDK V2.1 ?) :
   Purpose (of ASIOSetClockSource) :
      Set the audio clock source
    Parameter:
      index (ex: "reference"??) as obtained from an inquiry to ASIOGetClockSources()
    Returns:
      If no input/output is present ASE_NotPresent will be returned.
      If the clock can not be selected because an input channel which
      carries the current clock source is active, ASE_InvalidMode
      *may* be returned (this depends on the properties of the driver
      and/or hardware).
    Notes:
      Should *not* return ASE_NoClock if there is no clock signal present
      at the selected source; this will be inquired via ASIOGetSampleRate().
      It should call the host callback procedure sampleRateHasChanged(),
      if the switch causes a sample rate change, or if no external clock
      is present at the selected source.
   **************************************************************************/

  if(iClockSourceIndex==0)
   {
     pDriver->asioTime.timeInfo.flags |= kClockSourceChanged;
     return ASE_OK;
   }
  return ASE_NotPresent;
}

//---------------------------------------------------------------------------
ASIOError __stdcall AsioDrv_getSamplePosition(TMyAsioDriver * pDriver, ASIOSamples *sPos, ASIOTimeStamp *tStamp)
{
  /**************************************************************************
   Quote from a sourcecode floating around in the internet (from SDK V2.1 ?) :
   Purpose (of ASIOGetSamplePosition) :
      Inquires the sample position/time stamp pair.
    Parameter:
      sPos will hold the sample position on return. The sample
      position is reset to zero when ASIOStart() gets called.
      tStamp will hold the system time when the sample position
      was latched.
    Returns:
      If no input/output is present, ASE_NotPresent will be returned.
      If there is no clock, ASE_SPNotAdvancing will be returned.
    Notes:
      In order to be able to synchronise properly,
      the sample position / time stamp pair must refer to the current block,
      that is, the engine will call ASIOGetSamplePosition() in its bufferSwitch()
      callback and expect the time for the current block. thus, when requested
      in the very first bufferSwitch after ASIO_Start(), the sample position
      should be zero, and the time stamp should refer to the very time where
      the stream was started. it also means that the sample position must be
      block aligned. the driver must ensure proper interpolation if the system
      time can not be determined for the block position. the driver is responsible
      for precise time stamps as it usually has most direct access to lower
      level resources. proper behaviour of ASIO_GetSamplePosition() and ASIO_GetLatencies()
      are essential for precise media synchronization!
   **************************************************************************/

  double dblSamplePosition2, dblSamplePosition;

  do  // may have to read this TWICE when interrupted by the worker thread:
   { dblSamplePosition2 = (volatile double)pDriver->dblSamplePosition;
     tStamp->lo = pDriver->theSystemTime.lo;
     tStamp->hi = pDriver->theSystemTime.hi;
     dblSamplePosition = (volatile double)pDriver->dblSamplePosition;
   }while(dblSamplePosition != dblSamplePosition2);

  if ( pDriver->dblSamplePosition >= twoRaisedTo32)
   {
     sPos->hi = (unsigned long)(dblSamplePosition * twoRaisedTo32Reciprocal);
     sPos->lo = (unsigned long)(dblSamplePosition - (sPos->hi * twoRaisedTo32));
   }
  else
   {
     sPos->hi = 0;
     sPos->lo = (unsigned long)dblSamplePosition;
   }
  return ASE_OK;
}

//---------------------------------------------------------------------------
ASIOError __stdcall AsioDrv_getChannelInfo(TMyAsioDriver * pDriver,
        ASIOChannelInfo *info) // [in,out] structure defined in Steinberg's asio.h
{
  /**************************************************************************
   ASIO_GetChannelInfo :
      Retreives information about the nature of a channel.
    Parameter:
      info: pointer to a ASIOChannelInfo structure with
        - channel: on input, the channel index of the channel in question.
        - isInput: on input, ASIOTrue if info for an input channel is
          requested, else output
        - channelGroup: on return, the channel group that the channel
          belongs to. For drivers which support different types of
          channels, like analog, S/PDIF, AES/EBU, ADAT etc interfaces,
          there should be a reasonable grouping of these types. Groups
          are always independant form a channel index, that is, a channel
          index always counts from 0 to numInputs/numOutputs regardless
          of the group it may belong to.
          There will always be at least one group (group 0). Please
          also note that by default, the host may decide to activate
          channels 0 and 1; thus, these should belong to the most
          useful type (analog i/o, if present).
        - type: on return, contains the sample type of the channel
        - isActive: on return, ASIOTrue if channel is active as it was
          installed by ASIOCreateBuffers(), ASIOFalse else
        - name:  describing the type of channel in question. Used to allow
          for user selection, and enabling of specific channels. examples:
          "Analog In", "SPDIF Out" etc . MAXIMUM 31 characters plus trailing zero.
       [ Here, in DL4YHF's Audio-I/O-Library, the name will consist of the
         STREAM NAME, followed by the channel number within that stream.
         Example: Stream = VLF6 ->  channel names = VLF6.1, VLF6.2 ]   
    Returns:
      If no input/output is present ASE_NotPresent will be returned.
    Notes:
      If possible, the string should be organised such that the first
      characters are most significantly describing the nature of the
      port, to allow for identification even if the view showing the
      port name is too small to display more than 8 characters, for
      instance.
   **************************************************************************/

  long i;

  if (info->channel < 0 || (info->isInput ? info->channel >= kNumInputs : info->channel >= kNumOutputs))
   { return ASE_InvalidParameter;
   }

  info->type = ASIOSTInt16LSB;  // preferred by all little-endian machines ("Intel")
  info->channelGroup = 0;
  info->isActive = ASIOFalse;

  strcpy(info->name, "Reserve");
  if (info->isInput)
   {
     for (i = 0; i < pDriver->nActiveInputs; i++)
      {
        if (pDriver->inMap[i] == info->channel)
         {
           info->isActive = ASIOTrue;
           wsprintf(info->name, "In%d", (int) i+1 );
           break;
         }
      }
   }
  else
   {
     for (i = 0; i < pDriver->nActiveOutputs; i++)
      {
        if (pDriver->outMap[i] == info->channel)
         {
           info->isActive = ASIOTrue;
           wsprintf(info->name, "Out%d", (int) i+1 );
           break;
         }
      }
   }
  return ASE_OK;

}

//---------------------------------------------------------------------------
ASIOError __stdcall AsioDrv_createBuffers(TMyAsioDriver * pDriver,
                       ASIOBufferInfo *bufferInfos,
                       long numChannels,
                       long bufferSize,
                       ASIOCallbacks *callbacks)
{
  /**************************************************************************
   ASIOCreateBuffers :
      Allocates input/output buffers for all input and output channels to be activated.
    Parameter:
      bufferInfos is a pointer to an array of ASIOBufferInfo structures:
        - isInput: on input, ASIOTrue if the buffer is to be allocated
          for an input, output buffer else
        - channelNum: on input, the index of the channel in question
          (counting from 0)
        - buffers: on exit, 2 pointers to the halves of the channels' double-buffer.
          the size of the buffer(s) of course depend on both the ASIOSampleType
          as obtained from ASIOGetChannelInfo(), and bufferSize
      numChannels is the sum of all input and output channels to be created;
        thus bufferInfos is a pointer to an array of numChannels ASIOBufferInfo
        structures.
      bufferSize selects one of the possible buffer sizes as obtained from
        ASIOGetBufferSizes().   (WB: here AsioDrv_getBufferSize() ? )
      callbacks is a pointer to an ASIOCallbacks structure.
    Returns:
      If not enough memory is available ASE_NoMemory will be returned.
      If no input/output is present ASE_NotPresent will be returned.
      If bufferSize is not supported, or one or more of the bufferInfos elements
      contain invalid settings, ASE_InvalidMode will be returned.
    Notes:
      If individual channel selection is not possible but requested,
      the driver has to handle this. namely, bufferSwitch() will only
      have filled buffers of enabled outputs. If possible, processing
      and buss activities overhead should be avoided for channels which
      were not enabled here.
   **************************************************************************/

  ASIOBufferInfo *info = bufferInfos;
  long i;
  int  input_index  = 0;
  int  output_index = 0;
  BOOL notEnoughMem = FALSE;

  if( (pDriver->nActiveInputs>0) || (pDriver->nActiveOutputs>0) )
   { AsioDrv_disposeBuffers(pDriver);  // get rid of 'em old buffers
   }

  pDriver->nActiveInputs  = 0; // later set to input_index
  pDriver->nActiveOutputs = 0; // later set to output_index
  pDriver->nSamplesPerBlock = bufferSize;
  for (i = 0; i < numChannels; i++, info++)
   {
     if (info->isInput)
      {
        if (info->channelNum < 0 || info->channelNum >= kNumInputs)
           goto error;
        pDriver->inMap[input_index] = info->channelNum;
        // Create a double buffer for this 'input'-channel :
        pDriver->inputBuffers[input_index] = malloc( pDriver->nSamplesPerBlock * 2 * sizeof(short) );
        if( pDriver->inputBuffers[input_index] )
         {
           info->buffers[0] = pDriver->inputBuffers[input_index];
           info->buffers[1] = pDriver->inputBuffers[input_index] + pDriver->nSamplesPerBlock;
         }
        else
         {
           info->buffers[0] = info->buffers[1] = 0;
           notEnoughMem = TRUE;
         }
        input_index++; // ex: activeInputs++;
        if( input_index > kNumInputs )  // oops.. went too far ?!
         { input_index--;
           pDriver->nActiveInputs  = input_index;  // important to free other bufs
           pDriver->nActiveOutputs = output_index;
error:
           AsioDrv_disposeBuffers(pDriver);
           return ASE_InvalidParameter;
         }
      }
     else    // output
      {
        if (info->channelNum < 0 || info->channelNum >= kNumOutputs)
           goto error;
        pDriver->outMap[output_index] = info->channelNum;
        // Create a double buffer for this 'input'-channel :
        pDriver->outputBuffers[output_index] = malloc( pDriver->nSamplesPerBlock * 2 * sizeof(short) );
        if( pDriver->outputBuffers[output_index] != NULL)
         {
           info->buffers[0] = pDriver->outputBuffers[output_index];
           info->buffers[1] = pDriver->outputBuffers[output_index] + pDriver->nSamplesPerBlock;
         }
        else
         {
           info->buffers[0] = info->buffers[1] = 0;
           notEnoughMem = TRUE;
         }
        output_index++;  // ex: activeOutputs++;
        if (output_index > kNumOutputs)  // oops.. went too far ?!
         {
           output_index--;
           pDriver->nActiveInputs  = input_index;  // important to free other bufs
           pDriver->nActiveOutputs = output_index;
           AsioDrv_disposeBuffers(pDriver);
           return ASE_InvalidParameter;
         }
      }
   }
  pDriver->nActiveInputs  = input_index;  // should be 'ready for use' now
  pDriver->nActiveOutputs = output_index;
  if (notEnoughMem)                       // .. no, it's not !
   {
     AsioDrv_disposeBuffers( pDriver );
     return ASE_NoMemory;
   }

  pDriver->pCallbacks = callbacks; // WB: this may be dangerous,
                // because 'callbacks' points into the CALLER's var-space,
                // so we can only HOPE that the caller keeps it valid
                // (and doesn't, by accident, use a local variable to pass in
                // the addresses of 'his' callback functions)
  if (pDriver->pCallbacks->asioMessage(kAsioSupportsTimeInfo, 0, 0, 0) )
   { pDriver->timeInfoMode = TRUE;
     pDriver->asioTime.timeInfo.speed = 1.0;
     pDriver->asioTime.timeInfo.systemTime.hi = pDriver->asioTime.timeInfo.systemTime.lo = 0;
     pDriver->asioTime.timeInfo.samplePosition.hi = pDriver->asioTime.timeInfo.samplePosition.lo = 0;
     pDriver->asioTime.timeInfo.sampleRate = pDriver->dblSampleRate;
     pDriver->asioTime.timeInfo.flags = kSystemTimeValid | kSamplePositionValid | kSampleRateValid;

     pDriver->asioTime.timeCode.speed = 1.0;
     pDriver->asioTime.timeCode.timeCodeSamples.lo = pDriver->asioTime.timeCode.timeCodeSamples.hi = 0;
     pDriver->asioTime.timeCode.flags = kTcValid | kTcRunning ;
   }
  else
   { pDriver->timeInfoMode = FALSE;
   }
  return ASE_OK;

}

//---------------------------------------------------------------------------
ASIOError __stdcall AsioDrv_disposeBuffers(TMyAsioDriver * pDriver)
{
  /**************************************************************************
   Quote from a sourcecode floating around in the internet (from SDK V2.1 ?) :
   Purpose (of ASIODisposeBuffers) :
      Releases all buffers for the device.
    Parameter:
      None.
    Returns:
      If no buffer were ever prepared, ASE_InvalidMode will be returned.
      If no input/output is present ASE_NotPresent will be returned.
    Notes:
      This implies ASIOStop().
   **************************************************************************/

  long i;

  pDriver->pCallbacks = NULL;  // does the caller really want this side-effect ?
  AsioDrv_stop( pDriver );
  for (i = 0; i < pDriver->nActiveInputs && i<kNumInputs; i++)
   { if( pDriver->inputBuffers[i] != NULL )
      { free( pDriver->inputBuffers[i] );
      }
     pDriver->inputBuffers[i] = NULL;
   }
  pDriver->nActiveInputs = 0;
  for (i = 0; i < pDriver->nActiveOutputs && i<kNumOutputs; i++)
   { if( pDriver->outputBuffers[i] != NULL )
      { free( pDriver->outputBuffers[i] );
      }
     pDriver->outputBuffers[i] = NULL;
   }
  pDriver->nActiveOutputs = 0;
  return ASE_OK;
}

//---------------------------------------------------------------------------
ASIOError __stdcall AsioDrv_controlPanel(TMyAsioDriver * pDriver)
{
  /**************************************************************************
   Quote from a sourcecode floating around in the internet (from SDK V2.1 ?) :
   Purpose (of ASIOControlPanel) :
      request the driver to start a control panel component
      for device specific user settings. This will not be
      accessed on some platforms (where the component is accessed
      instead).
    Parameter:
      None.
    Returns:
      If no panel is available ASE_NotPresent will be returned.
      Actually, the return code is ignored.
    Notes:
      if the user applied settings which require a re-configuration
      of parts or all of the enigine and/or driver (such as a change of
      the block size), the asioMessage callback can be used (see
      ASIO_Callbacks).
   **************************************************************************/

  // When used in "in_AudioIO.dll", hosted by Spectrum Lab,
  //  the program gets here from SL's "Audio I/O" configuration screen
  //  through the following functions (Call Stack):
  // ShowAsioControlPanel() [still inside Spectrum Lab]
  //  -> AsioWrap_ShowControlPanel() [ "   "   "  ]
  //      -> ASIOControlPanel()  [this is Steinberg's "host"-API]
  //          -> IASIOThisCallResolver::controlPanel() [stdcall->thiscall]
  //              -> AsioThiscall_controlPanel() [IN THE DLL, thiscall->stdcall]
  //                  -> AsioDrv_controlPanel()  [IN THE DLL, ASIO_dummy.c]
  //                      -> AIO_ShowControlPanel() [opens a NON-MODAL window]
  // Unfortunately, the documentation by Steinberg is quite sparce;
  //  they don't even mention if ASIOControlPanel() should block the caller
  //   (until the operator closes the control panel, LIKE A MODAL DIALOG)
  //  or if in this case, the "ASIO driver" shall create a NON-BLOCKING
  //  window (with its own message loop, and all the windoze red tape).
  // Since Steinberg don't pass the caller's window handle, WB GUESSED
  // that ASIOControlPanel() should just open a non-modal window.
  // This seems to be confirmed by a modification-note in ASIO4ALL :
  // > GUI moved into its own thread completly. This has been necessary
  // > because existing host software as a whole is too broken on average
  // > as to be able to chase individual developers and have them fix it.
  // > Never mind the intstalled base out there. As a direct result,
  // > the ASIO control panel does not appear as modal dialog anymore.
  // Ok, we don't use an extra thread for the GUI, but we don't block here, too:
  if( AIO_ShowControlPanel( // NON-MODAL (not blocking) !
        NULL,        // [in] handle of the caller's main window . NOT AVAILABLE FROM ASIO-API !!
        NULL,        // [out,optional] handle of the DLL's control panel
        AIO_DISPLAY_SHOW_WINDOW, // [in] iDisplayOptions
        NULL,        // [in,optional] pCallbackUserDataForControlPanel
        NULL )      // [in,optional] address of a callback function, may be NULL
       == AIO_NO_ERROR )
   { return ASE_OK;  // even if Steinberg say they ignore the result, RETURN SOMETHING
   }
  else
   { return ASE_NotPresent; // no panel is available
   }
} // end AsioDrv_controlPanel()

//---------------------------------------------------------------------------
ASIOError __stdcall AsioDrv_future(TMyAsioDriver * pDriver, long selector,void *params)
{
  /**************************************************************************
   Quote from a sourcecode floating around in the internet (from SDK V2.1 ?) :
   Purpose (of ASIOFuture) :
      various
    Parameter:
      selector: operation Code as to be defined. zero is reserved for
      testing purposes.
      params: depends on the selector; usually pointer to a structure
      for passing and retreiving any type and amount of parameters.
    Returns:
      the return value is also selector dependant. if the selector
      is unknown, ASE_InvalidParameter should be returned to prevent
      further calls with this selector. on success, ASE_SUCCESS
      must be returned (note: ASE_OK is *not* sufficient!)
   **************************************************************************/

  // ASIOTransportParameters* tp = (ASIOTransportParameters*)params;
  switch (selector)
   {
     case kAsioEnableTimeCodeRead:   pDriver->tcRead = TRUE;
                                     return ASE_SUCCESS;
     case kAsioDisableTimeCodeRead:  pDriver->tcRead = FALSE;
                                     return ASE_SUCCESS;
     case kAsioSetInputMonitor:      return ASE_SUCCESS; // for testing!!!
     case kAsioCanInputMonitor:      return ASE_SUCCESS; // for testing!!!
     case kAsioCanTimeInfo:          return ASE_SUCCESS;
     case kAsioCanTimeCode:          return ASE_SUCCESS;
   }

  return ASE_InvalidParameter; // not 'ASE_NotPresent' which Steinberg use in THEIR OWN demo !
}

//---------------------------------------------------------------------------
ASIOError __stdcall AsioDrv_outputReady(TMyAsioDriver * pDriver)
{
  /**************************************************************************
   Quote from a sourcecode floating around in the internet (from SDK V2.1 ?) :
   Purpose (of ASIO_outputReady) :
      this tells the driver that the host has completed processing
      the output buffers. if the data format required by the hardware
      differs from the supported asio formats, but the hardware
      buffers are DMA buffers, the driver will have to convert
      the audio stream data; as the bufferSwitch callback is
      usually issued at dma block switch time, the driver will
      have to convert the *previous* host buffer, which increases
      the output latency by one block.
      when the host finds out that ASIOOutputReady() returns
      true, it will issue this call whenever it completed
      output processing. then the driver can convert the
      host data directly to the dma buffer to be played next,
      reducing output latency by one block.
      another way to look at it is, that the buffer switch is called
      in order to pass the *input* stream to the host, so that it can
      process the input into the output, and the output stream is passed
      to the driver when the host has completed its process.
    Parameter:
        None
    Returns:
      only if the above mentioned scenario is given, and a reduction
      of output latency can be acheived by this mechanism, should
      ASE_OK be returned. otherwise (and usually), ASE_NotPresent
      should be returned in order to prevent further calls to this
      function. note that the host may want to determine if it is
      to use this when the system is not yet fully initialized, so
      ASE_OK should always be returned if the mechanism makes sense.
    Notes:
      please remeber to adjust ASIOGetLatencies() according to
      whether ASIOOutputReady() was ever called or not, if your
      driver supports this scenario.
      also note that the engine may fail to call ASIO_OutputReady()
      in time in overload cases. as already mentioned, bufferSwitch
      should be called for every block regardless of whether a block
      could be processed in time.
   **************************************************************************/

  return 0;
}


//---------------------------------------------------------------------------
// private methods
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// input
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
BOOL AsioPrivate_inputOpen(TMyAsioDriver *pDriver)
{
  return TRUE;
}


//---------------------------------------------------------------------------------------------
void AsioPrivate_inputClose (TMyAsioDriver *pDriver )
{
}

//---------------------------------------------------------------------------
void AsioPrivate_input(TMyAsioDriver *pDriver)
{
  // Performs the actual 'input' processing.
  //   In a real ASIO driver, this would fill the active INPUT buffers
  //    ( pDriver->inputBuffers[0 ... <pDriver->nActiveInputs-1> ] )
  //   with samples received from the A/D converter.
  //   Here, in this dummy, each channel is filled with a coherent sinewave,
  //   with 1000 Hz for the 1st channel, with 2000 Hz for the 2nd, etc etc.
  //
  // Call stack:
  //   ASIOThread()
  //    -> AsioSample::bufferSwitch()
  //        |-> AsioSample::input()
  //        |-> AsioSample::output()
  //        |-> AsioSample::bufferSwitchX()           [for ASIO V2, or ...
  //        |-> AsioSample::callbacks->bufferSwitch    for older ASIO ]
  //
#if TESTWAVES
  long iInput, iSample;
  short *in = 0;
  double dblPhase, dblPhaseInc;

  for (iInput = 0; iInput < pDriver->nActiveInputs; ++iInput)
   {
     in = pDriver->inputBuffers[iInput];
     dblPhase = fmod( pDriver->dblSinewavePhase[iInput], 2.0*C_PI );
     dblPhaseInc = 2.0 * C_PI * (double)(1000*(iInput+1)) / pDriver->dblSampleRate;
     if( in != NULL )
      {
        if (pDriver->toggle)
                in += pDriver->nSamplesPerBlock;
        for( iSample=0; iSample<pDriver->nSamplesPerBlock; ++iSample )
         { *in++ = (short)(32767.0 * sin( dblPhase ) );
           dblPhase += dblPhaseInc;
         }
        pDriver->dblSinewavePhase[iInput] = dblPhase;
      }
   }
#endif
}

//---------------------------------------------------------------------------
// output
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
BOOL AsioPrivate_outputOpen(TMyAsioDriver *pDriver )
{
    return TRUE;
}

//---------------------------------------------------------------------------
void AsioPrivate_outputClose(TMyAsioDriver *pDriver )
{
}

//---------------------------------------------------------------------------------------------
void AsioPrivate_output(TMyAsioDriver *pDriver)
{
}

//---------------------------------------------------------------------------------------------
void AsioPrivate_bufferSwitch( TMyAsioDriver *pDriver )
{
  if (pDriver->started && (pDriver->pCallbacks!=NULL) )
    {
        getNanoSeconds(&pDriver->theSystemTime);         // latch system time
        AsioPrivate_input(pDriver);
        AsioPrivate_output(pDriver);
        pDriver->dblSamplePosition += pDriver->nSamplesPerBlock;
        if (pDriver->timeInfoMode)
            AsioPrivate_bufferSwitchX (pDriver);
        else
            pDriver->pCallbacks->bufferSwitch( pDriver->toggle, ASIOFalse);
        pDriver->toggle = pDriver->toggle ? 0 : 1;
    }
}

//---------------------------------------------------------------------------------------------
// asio2 buffer switch
void AsioPrivate_bufferSwitchX( TMyAsioDriver *pDriver )
{
  long offset;
  AsioDrv_getSamplePosition( (void*)pDriver,
      &pDriver->asioTime.timeInfo.samplePosition,
      &pDriver->asioTime.timeInfo.systemTime);
  offset = pDriver->toggle ? pDriver->nSamplesPerBlock : 0;
  if (pDriver->tcRead)  // TRUE when application requested 'kAsioEnableTimeCodeRead' ...
   { // Create a fake time code, which is 10 minutes ahead of the card's sample position
     // Please note that for simplicity here time code will wrap after 32 bit are reached
     pDriver->asioTime.timeCode.timeCodeSamples.lo = (unsigned long)(pDriver->asioTime.timeInfo.samplePosition.lo + 600.0 * pDriver->dblSampleRate);
     pDriver->asioTime.timeCode.timeCodeSamples.hi = 0;
   }
  if( pDriver->pCallbacks != NULL )
   {  pDriver->pCallbacks->bufferSwitchTimeInfo (&pDriver->asioTime, pDriver->toggle, ASIOFalse);
   }
  pDriver->asioTime.timeInfo.flags &= ~(kSampleRateChanged | kClockSourceChanged);
}

//---------------------------------------------------------------------------------------------
ASIOError AsioPrivate_outputReady(TMyAsioDriver *pDriver)
{
/* ASIOOutputReady:
	  this tells the driver that the host has completed processing
	  the output buffers. if the data format required by the hardware
	  differs from the supported asio formats, but the hardware
	  buffers are DMA buffers, the driver will have to convert
	  the audio stream data; as the bufferSwitch callback is
	  usually issued at dma block switch time, the driver will
	  have to convert the *previous* host buffer, which increases
	  the output latency by one block.
	  when the host finds out that ASIOOutputReady() returns
	  true, it will issue this call whenever it completed
	  output processing. then the driver can convert the
	  host data directly to the dma buffer to be played next,
	  reducing output latency by one block.
	  another way to look at it is, that the buffer switch is called
	  in order to pass the *input* stream to the host, so that it can
	  process the input into the output, and the output stream is passed
	  to the driver when the host has completed its process.
	Parameter:
		None
	Returns:
	  only if the above mentioned scenario is given, and a reduction
	  of output latency can be acheived by this mechanism, should
	  ASE_OK be returned. otherwise (and usually), ASE_NotPresent
	  should be returned in order to prevent further calls to this
	  function. note that the host may want to determine if it is
	  to use this when the system is not yet fully initialized, so
	  ASE_OK should always be returned if the mechanism makes sense.	  
	Notes:
	  please remeber to adjust ASIOGetLatencies() according to
	  whether ASIOOutputReady() was ever called or not, if your
	  driver supports this scenario.
	  also note that the engine may fail to call ASIO_OutputReady()
	  in time in overload cases. as already mentioned, bufferSwitch
      should be called for every block regardless of whether a block
      could be processed in time.
*/

    return ASE_NotPresent;
}


//------------------------------------------------------------------------------------------
void getNanoSeconds (ASIOTimeStamp* ts)
{
	double nanoSeconds = (double)((unsigned long)timeGetTime ()) * 1000000.;
	ts->hi = (unsigned long)(nanoSeconds / twoRaisedTo32);
	ts->lo = (unsigned long)(nanoSeconds - (ts->hi * twoRaisedTo32));
	// Got a LINKER ERROR when trying to build the 'ASIO driver' example 
	// with Microsoft Visual C++ 2010 Express :
    // 1> wintimer.obj : error LNK2019: unresolved external symbol __imp__timeGetTime@0 
	//    referenced in function "void __cdecl getNanoSeconds(struct ASIOTimeStamp *)"
	//    (?getNanoSeconds@@YAXPAUASIOTimeStamp@@@Z)
	// Cure: Add winmm.lib to the project (somewhere under Options..Linker..)
}

//------------------------------------------------------------------------------------------
void AsioPrivate_timerOn(TMyAsioDriver *pDriver)
{
	DWORD asioId;

	// pTheDriver = pDriver;  // EEEEK ! A static single instance, not much better than a GLOBAL VARIABLE !
         // (hey Steinberg, guess what 'lpParameter' in CreateThread is for ! )

	done = FALSE;
	ASIOThreadHandle = CreateThread(
                         0,           // pointer to thread security attributes
                         0,           // initial thread stack size, in bytes
                         ASIOThread,  // pointer to thread function (& is useless here)
                      (void*)pDriver, // void *param; argument for new thread ("user data" !)
                         0,           // creation flags
                         &asioId);    // pointer to returned thread identifier
}

//------------------------------------------------------------------------------------------
void AsioPrivate_timerOff(TMyAsioDriver *pDriver)
{
	done = TRUE;   // seems to be a 'polite attempt' to stop the ASIO worker thread
	if (ASIOThreadHandle)
	 {	WaitForSingleObject(ASIOThreadHandle, 1000);
    }
	ASIOThreadHandle = 0;
}

//---------------------------------------------------------------------------
static DWORD __stdcall ASIOThread (void *param)
  // Hey, did you know what Microsoft's intention for the 'param' was,
  //  when they created the Win32 thread API ?   Harr harr...
  // AVOIDING SUCH BULLSHIT LIKE STATIC/GLOBAL VARIABLES to communicate
  // between thread-data and the "outside world" !
  //
{
  TMyAsioDriver *pDriver = (TMyAsioDriver *)param; // eliminate Steinberg's dreadful 'theDriver'
  double dbl, secondsPerBlock, secondsInBufferSwitch;
  LONGLONG i64PerfCounterBeforeBufferSwitch, i64PerfCounterAfterBufferSwitch;
  LONGLONG i64PerfCounterFrequency, i64Diff;
  LONGLONG i64PerfCounterOfPreviousBufferSwitch = -1;
  double dblCurrCallingInterval; // number of seconds between the current and the previous buffer-switch
  double dblTimeEarlyLate = 0.0; // negative=LATE, positive=EARLY, 0=perfect. Unit: SECONDS
  do
	{ if( pDriver==NULL )
      { break;   // oops.. this should never happen
      }
	  if(  (pDriver->active) && (pDriver->started) )
		{
         QueryPerformanceCounter( (LARGE_INTEGER*)&i64PerfCounterBeforeBufferSwitch );
			AsioPrivate_bufferSwitch( pDriver );
         QueryPerformanceCounter( (LARGE_INTEGER*)&i64PerfCounterAfterBufferSwitch );
         secondsPerBlock = pDriver->nSamplesPerBlock / pDriver->dblSampleRate;
			// ex: Sleep( 1000.0 * secondsPerBlock );  // replaced 2011-10, reasoning below .
         // Note: the number of milliseconds which this thread REALLY 'slept' in Sleep()
         // may be unreliable. When the test-sinewaves were fed to a real soundcard,
         // periodic short dropouts were audible, because Spectrum Lab's
         // buffer ran empty. A CRUDE solution: Wait a bit shorter here than necessary,
         // and tolerate that the HOST (when called from bufferSwitch) blocks this thread ?
         // Only if, in the long run, the high-resolution timer values indicate
         // that we're permanently feeding 'too many samples per second'
         // into the ASIO host, slow down .
         // This should avoid problems with hosts which perform their own buffering
         // in much larger chunks than here.
         QueryPerformanceFrequency( (LARGE_INTEGER*)&i64PerfCounterFrequency );
         if( i64PerfCounterFrequency > 0 )   // does this system support this timer/counter at all ?
          { secondsInBufferSwitch    = (double)(i64PerfCounterAfterBufferSwitch-i64PerfCounterBeforeBufferSwitch)
                                     / (double)i64PerfCounterFrequency;
            // Typical values (seen in debugger sessions, calculation of FOUR sine waves using the FPU) :
            //  secondsInBufferSwitch = 0.000949; 0.000942;
            //  secondsPerBlock       = 0.0232 s [=1024 / 44100Hz]
            //  dblTimeEarlyLate      = 0.00713;  0.00789;
            if( i64PerfCounterOfPreviousBufferSwitch > 0 )
             {  // is it possible to do the early/late calculation ?
               dblCurrCallingInterval= (double)(i64PerfCounterBeforeBufferSwitch-i64PerfCounterOfPreviousBufferSwitch)
                                     / (double)i64PerfCounterFrequency;
               dblTimeEarlyLate += ( secondsPerBlock/*"ideal"*/ - dblCurrCallingInterval/*"measured"*/ );
               // Because we cannot trust this windows-timer-thingy,
               // and we cannot trust the accuracy of the 'real' soundcard used for output,
               // make the integrator a bit leaky:
               // and allow a slip of 1 / 100th of the calling cycle per calling cycle,
               // i.e. avoid building up an 'early/late' integral if the sample rate
               // is 48000 + 480 Hz instead of 48000 Hz :
               dbl = secondsPerBlock * 0.01;
               if( dblTimeEarlyLate < -secondsPerBlock )
                {  dblTimeEarlyLate += dbl;   // slip
                }
               if( dblTimeEarlyLate >  secondsPerBlock )
                {  dblTimeEarlyLate -= dbl;   // slip
                }
               // Hard limit (clip) the early-late integral to +/- 5 seconds.
               // More than that makes no sense at all !
               if( dblTimeEarlyLate < -5.0 )
                {  dblTimeEarlyLate = -5.0;   // clip
                }
               if( dblTimeEarlyLate >  5.0 )
                {  dblTimeEarlyLate =  5.0;   // clip
                }
             }
          } // end if( i64PerfCounterFrequency > 0 )
         else
          { secondsInBufferSwitch = 0.0;  // n/a ...
            dblCurrCallingInterval = secondsPerBlock; // assume the calling intervals are 'perfect'
          }
         i64PerfCounterOfPreviousBufferSwitch = i64PerfCounterBeforeBufferSwitch;

         // How long does this thread have to 'sleep' now, to stay in sync ?
         dbl = secondsPerBlock/*ideal*/ - secondsInBufferSwitch/*time already spent in ASIO-HOST*/;
         // If we are LATE (dblTimeEarlyLate negative), wait a little less now:
         dbl = dbl + 0.1 * dblTimeEarlyLate;  // all units in SECONDS !
               // Future plan: Let the input-buffer-usage (of a larger FIFO)
               //              have an effect on the early/late calculation !
               //    For example, when the planned TCPIP/ASIO-gateway detects
               //    that the input buffer usage tends to increase,
               //    slightly increase the 'calling frequency'
               //    of the ASIO host here .

         // Limit the 'time to sleep' to some reasonable value:
         if( dbl > 1.1*secondsPerBlock )
          {  dbl = 1.1*secondsPerBlock;
          }
         if( dbl >= 1e-3 )  // dare to wait at all ?
          { Sleep( (unsigned long)( 1000.0 * dbl) );  // Sleep-interval in MILLISECONDS !
            // Typical values (seen in debugger sessions, at this point) :
            //  secondsPerBlock      = 0.0232 s [=1024 / 44100Hz]
            //  Sleep-time here: dbl = 0.0229 s [=SLIGHTLY less than the above]
          }
         else
          { // If the interval is too short to wait here,
            // the 'early/late' interval will INCREASE, and sooner or later
            // there will be enough time to SLEEP here (at least one millisecond)
          }
		}
    else
		{
			double a = 1000. / 44100.;  // eeeek !
			Sleep ((long)(a * (double)kAsioSamplesPerBlock));
		}
	} while (!done);
	return 0;
}

//---------------------------------------------------------------------------------------------
double AsioSamples2double( ASIOSamples* samples )
{
    double a = (double)(samples->lo);
    if (samples->hi)
        a += (double)(samples->hi) * twoRaisedTo32;
    return a;
}




//---------------------------------------------------------------------------
// COM-related red tape...
//---------------------------------------------------------------------------


// A count of how many objects our DLL has created (by some
// app calling our IClassFactory object's CreateInstance())
// which have not yet been Release()'d by the app
static DWORD        OutstandingObjects;

// A count of how many apps have locked our DLL via calling our
// IClassFactory object's LockServer()
static DWORD        LockCount;











//---------------------------------------------------------------------------
// The TMyAsioDriver object (formerly IExample object)
//---------------------------------------------------------------------------

#if(0)  // from Jeff Glatt's "COM in plain C", we don't use it for the ASIO driver !
// In our .H file, we use a macro which defines our IExample struct
// as so:
//
// typedef struct {
//    IExampleVtbl  *lpVtbl;
// } IExample;
//
// In other words, the .H file defines our IExample to have nothing
// but a pointer to its VTable. And of course, every COM object must
// start with a pointer to its VTable.
//
// But we actually want to add some more members to our IExample.
// We just don't want any app to be able to know about, and directly
// access, those members. So here we'll define a MyRealIExample that
// contains those extra members. The app doesn't know that we're
// really allocating and giving it a MyRealIExample object. We'll
// lie and tell it we're giving a plain old IExample. That's ok
// because a MyRealIExample starts with the same VTable pointer.
//
// We add a DWORD reference count so that this IExample
// can be allocated (which we do in our IClassFactory object's
// CreateInstance()) and later freed. And, we have an extra
// 80 char buffer, which is used by some of the functions we'll
// add to TMyAsioDriver (formerly IExample) .
typedef struct {
    IMyAsioDriverVtbl    *lpVtbl;
    DWORD           count;
    char            buffer[80];
} MyRealIExample;
#else  // for the ASIO driver, use 'TMyAsioDriver' also instead of 'MyRealIExample' !
#endif // (0)

// Here are TMyAsioDriver's functions (formerly IExample's functions) .
//
// Every COM object's interface must have the 3 functions QueryInterface(),
// AddRef(), and Release().
//
// I also chose to add 2, extra functions to IExample, which a program
// will call with the names GetString and SetString.

// TMyAsioDriver's QueryInterface()
static HRESULT STDMETHODCALLTYPE QueryInterface(TMyAsioDriver *this, REFIID vTableGuid, void **ppv)
{
    // Check if the GUID matches IExample VTable's GUID. We gave the C variable name
    // IID_MyAsioDriver to our VTable GUID. We can use an OLE function called
    // IsEqualIID to do the comparison for us. Also, if the caller passed a
    // IUnknown GUID, then we'll likewise return the TMyAsioDriver, since it can
    // masquerade as an IUnknown object too
  if (!IsEqualIID(vTableGuid, &IID_IUnknown) && !IsEqualIID(vTableGuid, &IID_MyAsioDriver))
   {
     // We don't recognize the GUID passed to us. Let the caller know this,
     // by clearing his handle, and returning E_NOINTERFACE.
     *ppv = 0;
     return(E_NOINTERFACE);
   }

  // Fill in the caller's handle
  *ppv = this;

  // Increment the count of callers who have an outstanding pointer to this object
  this->lpVtbl->AddRef(this);

  return(NOERROR);
}

// TMyAsioDriver's AddRef()
static ULONG STDMETHODCALLTYPE AddRef(TMyAsioDriver *pDriver)
{ // Increment TMyAsioDriver's reference count, and return the updated value.
  return( ++pDriver->count );
}

// TMyAsioDriver's Release()
static ULONG STDMETHODCALLTYPE Release(TMyAsioDriver *pDriver)
{
  // Decrement TMyAsioDriver's reference count. If 0, then we can safely free
  // this TMyAsioDriver now.
  // WB: This is probably overkill for ASIO, because one ASIO-driver
  //     can only be instantiated ONCE (at least by Steinberg's design in SDK V2.1) !
  if( --pDriver->count==0 )
   {
     GlobalFree(pDriver);
     // ex: InterlockedDecrement(&OutstandingObjects);
     // Borland warned about suspicious pointer conversion above. Why ?
     // in winbase.h : WINBASEAPI LONG WINAPI InterlockedDecrement( IN OUT LONG volatile *lpAddend );
     // But 'OutstandingObjects' is a simple stupid DWORD. Thus:
     InterlockedDecrement( (LONG volatile *)&OutstandingObjects);
     return 0;
   }
  return pDriver->count;
}

// ================== The following are my own extra functions added to TMyAsioDriver

#if(0)
// IExample's SetString(). This copies the passed string to IExample's
// buffer[]
static HRESULT STDMETHODCALLTYPE SetString(IExample *this, char *str)
{
    DWORD  i;

    // Make sure that caller passed a buffer
    if (!str) return(E_POINTER);

    // Copy the passed str to IExample's buffer
    i = lstrlen(str);
    if (i > 79) i = 79;
    CopyMemory(((MyRealIExample *)this)->buffer, str, i);
    ((MyRealIExample *)this)->buffer[i] = 0;

    return(NOERROR);
}
#endif // (0)


#if(0)
// IExample's GetString(). This retrieves IExample's buffer[],
// and stores its contents in a buffer passed by the caller
static HRESULT STDMETHODCALLTYPE GetString(IExample *this, char *buffer, DWORD length)
{
    DWORD  i;

    // Make sure that caller passed a buffer
    if (!buffer) return(E_POINTER);

    if (length)
    {
        // Let's copy IExample's buffer to the passed buffer
        i = lstrlen(((MyRealIExample *)this)->buffer);
        --length;
        if (i > length) i = length;
        CopyMemory(buffer, ((MyRealIExample *)this)->buffer, i);
        buffer[i] = 0;
    }
    return(NOERROR);
}
#endif // (0)

// Here's TMyAsioDriver's VTable. It never changes so we can declare it static.
// But first some background info from Ross Bencina's ASIO-Thiscall-Resolver:
// >  FINDING FUNCTION POINTERS FROM AN IASIO POINTER
// >
// > The first field of a COM object is a pointer to its vtble. Thus a pointer
// > to an object implementing the IASIO interface also points to a pointer to
// > that object's vtbl. The vtble is a table of function pointers for all of
// > the virtual functions exposed by the implemented interfaces.
// >
// > If we consider a variable declared as a pointer to IASO:
// >
// > IASIO *theAsioDriver
// >
// > theAsioDriver points to:
// >
// > object implementing IASIO
// >  {
// >        IASIOvtbl *vtbl
// >        other data
// >  }
// > in other words, theAsioDriver points to a pointer to an IASIOvtbl
// >
// > vtbl points to a table of function pointers:
// >
// >  IASIOvtbl ( interface IASIO : public IUnknown )
// >    {
// >    (IUnknown functions)
// >    0   virtual HRESULT STDMETHODCALLTYPE (*QueryInterface)(REFIID riid, void **ppv) = 0;
// >    4   virtual ULONG STDMETHODCALLTYPE (*AddRef)() = 0;
// >    8   virtual ULONG STDMETHODCALLTYPE (*Release)() = 0;
// >
// > (IASIO functions, using the annoying MSVC's THISCALL calling convention
//    because Steinberg didn't specify one of the standard calling conventions):
// >    12	virtual ASIOBool (*init)(void *sysHandle) = 0;
// >    16	virtual void (*getDriverName)(char *name) = 0;
// >    20	virtual long (*getDriverVersion)() = 0;
//    ...
// >  The numbers in the left column show the byte offset of each function ptr
// >  from the beginning of the vtbl.
// They were used in iasiothiscallresolver.cpp (i.e. for the implementation
// of an ASIO HOST in Borland C++ or GCC), so we must do something similar here.
// The prefix AsioThiscall_ indicates the WRAPPER FUNCTIONS, written in
// assembly language to convert the 'thiscall' parameter passing into something
// which all non-Microsoft-C-compilers can handle: The good old __stdcall .

static const IMyAsioDriverVtbl IMyAsioDriver_Vtbl =
{
  // Methods from the IUnknown interface :
  QueryInterface,
  AddRef,
  Release,

  // Methods from the IASIO interface, must use the dreadful
  // 'thiscall' calling convention (which Borland doesn't support)
  // for the outside world, to keep this thing halfway compatible with Steinberg.
  // As explained above, these functions begin with AsioThiscall_ to emphasize
  // that these function are just WRAPPERS, written in assembler,
  // to convert the annoying 'thiscall' into standard C function calls (__stdcall).
  // Note that, due to 'thiscall' these functions receive the 'this' pointer
  // in register ECX (NOT on the stack as __stdcall !)  .
  // The following THISCALL-WRAPPERS are declared in AsioDriverThiscalls.h :
  AsioThiscall_init,          // virtual ASIOBool init(void *sysHandle);
  AsioThiscall_getDriverName, // virtual void getDriverName(char *name);
  AsioThiscall_getDriverVersion, // virtual long getDriverVersion();
  AsioThiscall_getErrorMessage,  // virtual void getErrorMessage(char *string);
  AsioThiscall_start,         // virtual ASIOError start();
  AsioThiscall_stop,          // virtual ASIOError stop();
  AsioThiscall_getChannels,   // virtual ASIOError getChannels(long *numInputChannels, long *numOutputChannels);
  AsioThiscall_getLatencies,  // virtual ASIOError getLatencies(long *inputLatency, long *outputLatency);
  AsioThiscall_getBufferSize, // virtual ASIOError getBufferSize(long *minSize, long *maxSize, long *preferredSize, long *granularity);
  AsioThiscall_canSampleRate, // virtual ASIOError canSampleRate(ASIOSampleRate sampleRate);
  AsioThiscall_getSampleRate, // virtual ASIOError getSampleRate(ASIOSampleRate *sampleRate);
  AsioThiscall_setSampleRate, // virtual ASIOError setSampleRate(ASIOSampleRate sampleRate);
  AsioThiscall_getClockSources,  // virtual ASIOError getClockSources(ASIOClockSource *clocks, long *numSources);
  AsioThiscall_setClockSource,   // virtual ASIOError setClockSource(long reference);
  AsioThiscall_getSamplePosition,// virtual ASIOError getSamplePosition(ASIOSamples *sPos, ASIOTimeStamp *tStamp);
  AsioThiscall_getChannelInfo,   // virtual ASIOError getChannelInfo(ASIOChannelInfo *info);
  AsioThiscall_createBuffers, // virtual ASIOError createBuffers(ASIOBufferInfo *bufferInfos, long numChannels, long bufferSize, ASIOCallbacks *callbacks);
  AsioThiscall_disposeBuffers,// virtual ASIOError disposeBuffers();
  AsioThiscall_controlPanel,  // virtual ASIOError controlPanel();
  AsioThiscall_future,        // virtual ASIOError future(long selector,void *opt);
  AsioThiscall_outputReady    // virtual ASIOError outputReady();
};



//---------------------------------------------------------------------------
// The IClassFactory object
//---------------------------------------------------------------------------

// Since we only ever need one IClassFactory object, we declare
// it static. The only requirement is that we ensure any
// access to its members is thread-safe
static IClassFactory    MyIClassFactoryObj;

// IClassFactory's AddRef()
static ULONG STDMETHODCALLTYPE classAddRef(IClassFactory *pClassFactory)
{
  // Someone is obtaining my IClassFactory, so inc the count of
  // pointers that I've returned which some app needs to Release()
  // ex: InterlockedIncrement(&OutstandingObjects);
  InterlockedIncrement( (volatile long*)&OutstandingObjects);

  // Since we never actually allocate/free an IClassFactory (ie, we
  // use just 1 static one), we don't need to maintain a separate
  // reference count for our IClassFactory. We'll just tell the caller
  // that there's at least one of our IClassFactory objects in existance
  return(1);
}

// IClassFactory's QueryInterface()
static HRESULT STDMETHODCALLTYPE classQueryInterface(IClassFactory *pClassFactory, REFIID factoryGuid, void **ppv)
{
  // Make sure the caller wants either an IUnknown or an IClassFactory.
  // In either case, we return the same IClassFactory pointer passed to
  // us since it can also masquerade as an IUnknown
  if (IsEqualIID(factoryGuid, &IID_IUnknown) || IsEqualIID(factoryGuid, &IID_IClassFactory))
   {
     // Call my IClassFactory's AddRef
     pClassFactory->lpVtbl->AddRef(pClassFactory);

     // Return (to the caller) a ptr to my IClassFactory
     *ppv = pClassFactory;

     return(NOERROR);
   }

  // We don't know about any other GUIDs
  *ppv = 0;
  return(E_NOINTERFACE);
}

// IClassFactory's Release()
static ULONG STDMETHODCALLTYPE classRelease(IClassFactory *pClassFactory)
{
  // One less object that an app has not yet Release()'ed
  return(InterlockedDecrement( (volatile long*)&OutstandingObjects));
}

// IClassFactory's CreateInstance() function. It is called by
// someone who has a pointer to our IClassFactory object and now
// wants to create and retrieve a pointer to our TMyAsioDriver
static HRESULT STDMETHODCALLTYPE classCreateInstance(
          IClassFactory *pClassFactory,
          IUnknown *punkOuter,
          REFIID vTableGuid,
          void **objHandle)
{
  HRESULT        hr;
  TMyAsioDriver *pDriver;  // address of a new driver instance (allocated here)

  // Assume an error by clearing caller's handle
  *objHandle = 0;

  // We don't support aggregation in this example
  if (punkOuter)
      hr = CLASS_E_NOAGGREGATION;
  else
   {
     // Allocate our TMyAsioDriver object :
     if ( (pDriver = (TMyAsioDriver *)GlobalAlloc(GMEM_FIXED, sizeof(TMyAsioDriver)))==NULL )
      {
        hr = E_OUTOFMEMORY;
      }
     else
      {
        // WB: Don't trust that a newly allocated memory block is 'zeroed'.
        //     To avoid leaving anything to fate, ZERO the memory here:
        memset( pDriver, 0, sizeof(TMyAsioDriver) );

        // Store TMyAsioDriver's VTable in the object .
        pDriver->lpVtbl = (IMyAsioDriverVtbl *)&IMyAsioDriver_Vtbl;

        // Increment the reference count so we can call Release() below and
        // it will deallocate only if there is an error with QueryInterface()
        pDriver->count = 1;

        // EX: Initialize any other members we added to the TMyAsioDriver.
        //   pDriver->buffer[0] = 0;
        // (For the 'ASIO driver in plain C', we'll leave this to
        //  the ASIO driver's "constructor", which is AsioDrv_Construct() )
        AsioDrv_Construct( pDriver ); // replaces the contructor
          // (which, in C++, would automagically be called via 'new')


        // Fill in the caller's handle with a pointer to the TMyAsioDriver we just
        // allocated above. We'll let TMyAsioDriver's QueryInterface do that, because
        // it also checks the GUID the caller passed, and also increments the
        // reference count (to 2) if all goes well
        hr = IMyAsioDriver_Vtbl.QueryInterface(pDriver, vTableGuid, objHandle);

        // Decrement reference count. NOTE: If there was an error in QueryInterface()
        // then Release() will be decrementing the count back to 0 and will free the
        // TMyAsioDriver for us. One error that may occur is that the caller is asking for
        // some sort of object that we don't support (ie, it's a GUID we don't recognize)
        IMyAsioDriver_Vtbl.Release(pDriver);

        // If success, inc static object count to keep this DLL loaded
        if (!hr) InterlockedIncrement( (volatile long*)&OutstandingObjects);
      }
   }
  return(hr);
}

// IClassFactory's LockServer(). It is called by someone
// who wants to lock this DLL in memory
static HRESULT STDMETHODCALLTYPE classLockServer(IClassFactory *pClassFactory, BOOL flock)
{
    if (flock) InterlockedIncrement( (volatile long*)&LockCount);
    else InterlockedDecrement( (volatile long*)&LockCount);

    return(NOERROR);
}

// IClassFactory's VTable
static const IClassFactoryVtbl IClassFactory_Vtbl =
{ classQueryInterface,
  classAddRef,
  classRelease,
  classCreateInstance,
  classLockServer
};



//---------------------------------------------------------------------------
// Miscellaneous functions
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
__declspec(dllexport) HRESULT PASCAL DllGetClassObject(
                         REFCLSID objGuid,     // [in]
                         REFIID factoryGuid,   // [in]
                         void **factoryHandle) // [out]
  // This is called by the OLE functions CoGetClassObject() or
  // CoCreateInstance() in order to get our DLL's IClassFactory object
  // (and return it to someone who wants to use it to get ahold of one
  // of our TMyAsioDriver objects). Our IClassFactory's CreateInstance() can
  // be used to allocate/retrieve our TMyAsioDriver object.
  //
  //  NOTE: After we return the pointer to our IClassFactory, the caller
  //        will typically call its CreateInstance() function.
  //
  // Note (WB): Other authors use 'STDAPI' instead of 'HRESULT PASCAL' !
  //
// ex: HRESULT PASCAL DllGetClassObject(REFCLSID objGuid, REFIID factoryGuid, void **factoryHandle)
// With the above, there was no function named 'DllGetClassObject' exported in the DLL. Shit.
// With the following monsterous red tape, 'DllGetClassObject' exported in the DLL,
//   but does it have the correct (Microsoft-compatible) calling convention ?
// Exported functions from a COM server (here: ASIO driver emasio.dll) found by tdump:
// >  000075D3    1 0000 DllCanUnloadNow
// >  00007E5A    2 0001 DllGetClassObject
// >  0000828F    3 0002 DllInstall
// >  00008282    4 0003 DllRegisterServer
// >  00007E21    5 0004 DllUnregisterServer
{
  HRESULT hr;

  // Check that the caller is passing our TMyAsioDriver GUID. That's the
  // only object our DLL implements
  if(   (MyIClassFactoryObj.lpVtbl != NULL )  // + additional safety check
     && IsEqualCLSID(objGuid, &CLSID_MyAsioDriver))
   {
     // Fill in the caller's handle with a pointer to our IClassFactory object.
     // We'll let our IClassFactory's QueryInterface do that, because it also
     // checks the IClassFactory GUID and does other book-keeping
     hr = classQueryInterface(&MyIClassFactoryObj, factoryGuid, factoryHandle);
   }
  else
   {
     // We don't understand this GUID. It's obviously not for our DLL.
     // Let the caller know this by clearing his handle and returning
     // CLASS_E_CLASSNOTAVAILABLE
     *factoryHandle = 0;
     hr = CLASS_E_CLASSNOTAVAILABLE;
   }

    return(hr);
}





/************************ DllCanUnloadNow() ***********************
* This is called by some OLE function in order to determine
* whether it is safe to unload our DLL from memory.
*
* RETURNS: S_OK if safe to unload, or S_FALSE if not.
*/

__declspec(dllexport) HRESULT PASCAL DllCanUnloadNow(void)
{
    // If someone has retrieved pointers to any of our objects, and
    // not yet Release()'ed them, then we return S_FALSE to indicate
    // not to unload this DLL. Also, if someone has us locked, return
    // S_FALSE
    return((OutstandingObjects | LockCount) ? S_FALSE : S_OK);
}





//---------------------------------------------------------------------------
// ex: __declspec(dllexport) BOOL WINAPI DllMain(HINSTANCE instance, DWORD fdwReason, LPVOID lpvReserved)
// Since there is already a DllMain in aio2winamp.c, use this 'dummy' instead,
//   which will be called from the 'real' DllMain() :
BOOL WINAPI ASIO_DllMain(HINSTANCE instance, DWORD fdwReason, LPVOID lpvReserved)
  // Called by the Windows OS when this DLL is loaded or unloaded.
  // > An optional entry point into a dynamic-link library (DLL).
  // > When the system starts or terminates a process or thread,
  // > it calls the entry-point function for each loaded DLL
  // > using the first thread of the process. The system also calls
  // > the entry-point function for a DLL when it is loaded or unloaded
  // > using the LoadLibrary and FreeLibrary functions.
  // - - -
  // To test (invoke) this function with the integrated debugger in Borland C++Builder,
  //  copy the following command line into 'Start'..'Parameter'..
  //  'Host Application': C:\cbproj\COM_in_plain_C\iexampleapp\IExample_Host.exe
  //  'Parameter':
  // Notes:
  //  * If this DLL was not registered as a COM object,
  //    windows will not even *try* to load it !
  //    If a Borland-debugger-breakpoint in a COM 'InProc'-server works at all,
  //    wasn't sure at the time of this writing.
  //    At least, when REGISTERING the COM object (through regsrv32.exe),
  //    DllMain() will be called, AND breakpoints in the Borland-debugger
  //    are active.
  //  * If, for some reason, this function isn't called (because of the
  //    usual annoyance with mangled names, or because there is another
  //    'main entry point' in the DLL, inserted by the foolish IDE
  //    in the form of 'DllEntryPoint'),
  //    the DLL will inevitably CRASH inside DllGetClassObject()
  //    because 'MyIClassFactoryObj' will not have a valid VTABLE then.
  //  * Fortunately Borland C++Builder allows setting a breakpoint here,
  //    and successfully stops when SpectrumLab (!) is used as the
  //    'ASIO host' and loads the DLL (temporarily) to find the properties
  //    of this ASIO driver (it doesn't rely on the windoze registry alone) .
  // - - -
{
  BOOL fResult = TRUE;
  switch (fdwReason)
   {
     case DLL_PROCESS_ATTACH:
      {
            // Clear static counts
            OutstandingObjects = LockCount = 0;

            // Initialize my IClassFactory with the pointer to its VTable
            MyIClassFactoryObj.lpVtbl = (IClassFactoryVtbl *)&IClassFactory_Vtbl;

            // We don't need to do any thread initialization
            DisableThreadLibraryCalls(instance);
      }
   }

  // > When the system calls the DllMain function with the DLL_PROCESS_ATTACH value,
  // > the function returns TRUE if it succeeds or FALSE if initialization fails.
  return fResult;  // TRUE or FALSE, not some magic number !
}


//---------------------------------------------------------------------------
// Registration / Deregistration of this DLL (as a COM component)
//---------------------------------------------------------------------------

// Registers our ASIO-Driver-DLL as a COM component.


// These first 3 strings will change if you create your own COM object
//  (Note: ASIO makes things a bit more complicated; see ASIO_Register.c)
static const TCHAR  OurDllName[] = _T(MY_DLL_NAME);
static const TCHAR  ObjectDescription[] = _T(MY_REGISTRY_ASIO_KEY); // "friendly" description, OPTIONAL
static const TCHAR  ClassKeyName[] = _T("Software\\Classes"); // for COM, ASIO needs additional keys...
static const TCHAR  CLSID_Str[] = _T("CLSID");
static const TCHAR  InprocServer32Name[] = _T("InprocServer32");
static const TCHAR  ThreadingModelKey[] = _T("ThreadingModel"); // this is the KEY-NAME (not the VALUE)
static const TCHAR  ThreadingModelVal[] = _T("Apartment"); // this is the KEY-VALUE for the threading-model: "both", "Apartment", etc


/************************ cleanup() ***********************
* Removes all of the registry entries we create in WinMain.
* This is called to do cleanup if there's an error
* installing our DLL as a COM component.
*/
static void cleanup(void)
{
    HKEY        rootKey;
    HKEY        hKey;
    HKEY        hKey2;
    TCHAR       buffer[39];

    stringFromCLSID(&buffer[0], (CLSID*)&CLSID_MyAsioDriver );
      // note: stringFromCLSID is now in ASIO_Register.c  !

    // Open "HKEY_LOCAL_MACHINE\Software\Classes"
    if (!RegOpenKeyEx(HKEY_LOCAL_MACHINE, &ClassKeyName[0], 0, KEY_WRITE, &rootKey))
    {
        // Delete our CLSID key and everything under it
        if (!RegOpenKeyEx(rootKey, &CLSID_Str[0], 0, KEY_ALL_ACCESS, &hKey))
        {
            if (!RegOpenKeyEx(hKey, &buffer[0], 0, KEY_ALL_ACCESS, &hKey2))
            {
                RegDeleteKey(hKey2, &InprocServer32Name[0]);

                RegCloseKey(hKey2);
                RegDeleteKey(hKey, &buffer[0]);
            }

            RegCloseKey(hKey);
        }

        RegCloseKey(rootKey);
    }
}





//---------------------------------------------------------------------------
__declspec(dllexport) HRESULT __stdcall DllRegisterServer(void)
  // Server registration, called on REGSVR32.EXE <dll-name> .
  // Registers a COM component (in the windows registry)  .
  // Based on "COM in plain C" By Jeff Glatt (there: RegIExample.c::WinMain) .
  //
  // - - -
  // To test (invoke) this function with the integrated debugger in Borland C++Builder,
  //  copy the following command line into 'Start'..'Parameter'..
  //  'Host Application': C:\WINDOWS\system32\regsvr32.exe
  //  'Parameter':        iexample_bcb6.dll
  // [quite amazing that the breakpoint REALLY WORKED this way,
  //  when the DLL was invoked through windows' RegSvr32.exe ]
  // - - -
{

#if(1)  // use the functions in ASIO_Register.c (the clean COM example was not sufficient):
  LONG    rc;
  char    errstr[128];

  rc = RegisterAsioDriver(
          (CLSID*)&CLSID_MyAsioDriver, // [in]  "class ID", the long hexadecimal thingy (GUID)
          MY_DLL_NAME,                 // [in]  the name of the compiled DLL (without path)
          MY_REGISTRY_ASIO_KEY,
          MY_ASIO_DRIVER_DESCRIPTION,
          "Apartment");

  if( rc!=0 )
   { wsprintf(errstr,"Register Server failed ! (%d)",rc);
     MessageBox(0, (LPCTSTR)errstr, (LPCTSTR)MY_DLL_NAME, MB_OK);
     return -1;
   }

  return S_OK;


#else  // simpler code to register a COM server (but not sufficient for ASIO):

  HRESULT     hresult = E_FAIL;  // not a HANDLE but an OLE/COM-related 'error code' by M$
  int         iresult;
  TCHAR       filename[MAX_PATH+1];
  HKEY        rootKey;
  HKEY        hKey;
  HKEY        hKey2;
  HKEY        hkExtra;
  TCHAR       buffer[39];
  DWORD       disposition;



  HMODULE hMod = GetModuleHandle(MY_DLL_NAME);
  // With the parameter NULL, GetModuleHandle returns a handle of the
  //  "file used to create the calling process" (microsoft-slang).
  // Who's that, "OUR" name or the current DLL-host's name ?
  // To avoid any 'assumptions' here, we pass in the name of the DLL(!)
  // without the full path. This is what Steinberg's simplistic demo does, too.
  if(hMod==NULL) // GetModuleHandle failed !
   { return E_FAIL;
   }
  if( 0== GetModuleFileName(hMod,filename,MAX_PATH) )
   { // .. If the function succeeds, the return value is the length,
     //    in characters, of the string copied to the buffer..
     return E_FAIL;
   }
  // At this point, filename contained something like
  //  "C:\cbproj\COM_in_plain_C\iexample_bcb6\iexample_bcb6.dll"
  // Not sure if and why this must be lower-case,
  //  but Steinberg's ASIO-example does this, too:
  CharLower((LPTSTR)filename);  // -> entire path and filename in LOWER CASE


  // Open "HKEY_LOCAL_MACHINE\Software\Classes"
  if (ERROR_SUCCESS==RegOpenKeyEx(HKEY_LOCAL_MACHINE, &ClassKeyName[0], 0, KEY_WRITE, &rootKey))
   {
     // Open "HKEY_LOCAL_MACHINE\Software\Classes\CLSID"
     if (ERROR_SUCCESS==RegOpenKeyEx(rootKey, &CLSID_Str[0], 0, KEY_ALL_ACCESS, &hKey))
      {
        // Create a subkey whose name is the ascii string that represents
        // our xyz-object's GUID
        stringFromCLSID(&buffer[0], (CLSID*)&CLSID_MyAsioDriver );
        // -> buffer contains something like "{987DB4BE-C357-4518-A7E3-3DDCAC13AB9F}" .
        // This Class-ID must appear in the windoze registry very soon, under
        //   HKEY_LOCAL_MACHINE.SOFTWARE.Classes.CLSID (amongst a myriade of other CLSIDs)
        if (ERROR_SUCCESS==RegCreateKeyEx(hKey, &buffer[0], 0, 0, REG_OPTION_NON_VOLATILE, KEY_WRITE, 0, &hKey2, &disposition))
         {
           // Set its default value to some "friendly" string that helps
           // a user identify what this COM DLL is for. Setting this value
           // is optional. You don't need to do it.
           RegSetValueEx(hKey2, 0, 0, REG_SZ, (const BYTE *)&ObjectDescription[0], sizeof(ObjectDescription));

           // Create an "InprocServer32" key whose default value is the path of this DLL
           if (ERROR_SUCCESS==RegCreateKeyEx(hKey2, &InprocServer32Name[0], 0, 0, REG_OPTION_NON_VOLATILE, KEY_WRITE, 0, &hkExtra, &disposition))
            {
              if (ERROR_SUCCESS==RegSetValueEx(hkExtra, 0, 0, REG_SZ, (const BYTE *)&filename[0], lstrlen(&filename[0]) + 1))
               {
                 // Create a "ThreadingModel" value set to the string "Apartment"
                 // because that's what other ASIO drivers use, too .
                 // For an ASIO-driver, the only valid threading-model seems to be "Apartment" !
                 if (ERROR_SUCCESS==RegSetValueEx(hkExtra, &ThreadingModelKey[0], 0, REG_SZ, (const BYTE *)&ThreadingModelVal[0], sizeof(ThreadingModelVal)))
                  {
                    hresult = S_OK;
                    // MessageBox(0, "Successfully registered the DLL as a COM component.", &ObjectDescription[0], MB_OK);
                  }
               }
              // Close all keys we opened/created.
              RegCloseKey(hkExtra);
            }
           RegCloseKey(hKey2);
         }
        RegCloseKey(hKey);
      } // end if ..RegOpenKeyEx(rootKey, &CLSID_Str[ ..
     RegCloseKey(rootKey);
   } // end if ..RegOpenKeyEx(HKEY_LOCAL_MACHINE ..

  // If an error, make sure we clean everything up
  if (hresult!=S_OK)
   {
     cleanup();
     // MessageBox(0, "Failed to register DLL as a COM component.", &ObjectDescription[0], MB_OK|MB_ICONEXCLAMATION);
   }

  return hresult;
  // After this, check the new entry with regedit. Search for the DLL name. Example:
  //
  // HKEY_LOCAL_MACHINE
  // |_SOFTWARE
  //   |_Classes
  //      |_CLSID
  //         |__ (thousands of other hexadecimal GUIDs for class-IDs)
  //         |__ {B2F22418-5FF7-4971-BD39-47650F89EB39}  :
  //              |                   Typ  = REG_SZ
  //              |                   Value= ASIO dummy driver
  //              |___ InprocServer32
  //                      Name:(Standard)     Typ=REG_SZ  Value=c:\cbproj\ASIO_in_plain_c\XYZ.dll
  //                      Name:ThreadingModel Typ=REG_SZ  Value=Apartment
#endif

}



//---------------------------------------------------------------------------
__declspec(dllexport) HRESULT __stdcall DllUnregisterServer(void)
  // Unregisters a COM component (removes it from the windows registry)  .
  // Based on "COM in plain C" By Jeff Glatt (there: UnregIExample.c::WinMain).
  // - - -
  // To test (invoke) this function with the integrated debugger in Borland C++Builder,
  //  copy the following command line into 'Start'..'Parameter'..
  //  'Host Application': C:\WINDOWS\system32\regsvr32.exe
  //  'Parameter':        -u iexample_bcb6.dll
  // - - -
{
#if(1)  // use the functions in ASIO_Register.c (the clean COM example was not sufficient):

  LONG    rc;
  char    errstr[128];

  rc = UnregisterAsioDriver(
         (CLSID*)&CLSID_MyAsioDriver, // [in]  "class ID", the long hexadecimal thingy (GUID)
          MY_DLL_NAME,                // [in]  the name of the compiled DLL (without path)
          MY_REGISTRY_ASIO_KEY );

  if (rc)
   { wsprintf(errstr,"Unregister Server failed ! (%d)",rc);
     MessageBox(0, errstr, MY_DLL_NAME, MB_OK);
     return -1;
   }

  return S_OK;

#else  // simpler code to unregister a COM server (but not sufficient for ASIO):
  HRESULT hresult = E_FAIL;  // not a HANDLE but an OLE/COM-related 'error code' by M$
  int     iresult;
  TCHAR filename[MAX_PATH];
  HKEY  rootKey;
  HKEY  hKey;
  HKEY  hKey2;
  TCHAR buffer[39];


  HMODULE hMod = GetModuleHandle(MY_DLL_NAME);
     // With the parameter NULL, GetModuleHandle returns a handle of the
     //  "file used to create the calling process" (microsoft-slang).
     // Who's that, "OUR" name or the current DLL-host's name ?
     // To avoid any 'assumptions' here, we pass in the name of the DLL(!)
     // without the full path. This is what Steinberg's simplistic demo does, too.
  if(hMod==NULL) // GetModuleHandle failed !
   { return E_FAIL;
   }
  if( 0== GetModuleFileName(hMod,filename,MAX_PATH) )
   { // .. If the function succeeds, the return value is the length,
     //    in characters, of the string copied to the buffer..
     return E_FAIL;
   }
  // Not sure if and why this must be lower-case,
  //  but Steinberg's ASIO-example does this, too:
  CharLower((LPTSTR)filename);

  stringFromCLSID(&buffer[0], (CLSID*)&CLSID_MyAsioDriver );

  // Open "HKEY_LOCAL_MACHINE\Software\Classes"
  if (ERROR_SUCCESS==RegOpenKeyEx(HKEY_LOCAL_MACHINE, &ClassKeyName[0], 0, KEY_WRITE, &rootKey))
   {
     // Delete our CLSID key and everything under it
     if (ERROR_SUCCESS==RegOpenKeyEx(rootKey, &CLSID_Str[0], 0, KEY_ALL_ACCESS, &hKey))
      {
        if (ERROR_SUCCESS==RegOpenKeyEx(hKey, &buffer[0], 0, KEY_ALL_ACCESS, &hKey2))
         {
           RegDeleteKey(hKey2, &InprocServer32Name[0]);
           RegCloseKey(hKey2);
           RegDeleteKey(hKey, &buffer[0]);
           hresult = S_OK;
         }
        RegCloseKey(hKey);
      }
     RegCloseKey(rootKey);
   }

  if( hresult == S_OK )
   {
     // MessageBox(0, "De-registered DLL as a COM component.", &ObjectDescription[0], MB_OK|MB_ICONEXCLAMATION);
   }

  return hresult;

#endif // (0,1)

}


#endif // SWI_ASIO_SUPPORTED
