//----------------------------------------------------------------------
// File:  c:\cbproj\AudioIO\aio2winamp.c  ("Audio-I/O to Winamp")
// Author: Wolfgang Buescher (DL4YHF)
// Date:   2011-07-17
// Purpose: Implementation of an "Audio-I/O"-DLL, which connects applications
//              producing and consuming uncompressed audio streams in real time
//              (AIO_ReadInputSamplePoints,  AIO_WriteOutputSamplePoints) .
//          In THIS case, shared memory objects are used for the data exchange.
//
//   At the same time, this DLL can be loaded as an "input"-plugin
//          for Nullsoft Winamp to receive audio from the "Audio-I/O"-utiliy .
//          (for that purpose, the file MUST be copied into the winamp
//           plugin folder, something like
//           "c:\Programme\Winamp\Plugins\in_AudioIO.dll"  )
// Project status:
//          Highly experimental, only tested under windows XP.
// Naming conventions:
//  - everything which has to do ONLY with the winamp output plugin is prefixed
//    with "WA_" (winamp, any plugin) .  See info on this in the Winamp SDK .
//
//  - everything which will be called from other "external" applications
//    (like SpecLab) is prefixed "AIO_" ( Audio-I/O), to emphasize that
//    "this part" doesn't depend on winamp .
//
// Build environment :
//   - Borland C++Builder V4.0
//     or
//   - DevCpp V4.9.9.2 with updated "Binutils"
//     or
//   - wxDev-C++ 6.10.2 (a DevCpp descendant which supports wxWidgets)
//     Note about the wxDev-C++ project:
//          The final output is the file
//           "c:\cproj\AudioIO\Default Profile\in_AudioIO.dll" .
//                            |_______________| (ugly folder name)
//
//
// Note: When compiling DLLs with certain compilers (like Borland C++Builder),
//       the names in the DLL get MANGLED so GetProcAddress() won't find them.
// Cure: Use a *.DEF-file, for example c:\cbproj\AudioIO\in_AudioIO_BCB4.def .
//       The DEF-file must be ADDED to the project, at least for Borland !
//
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Copyright, License, and Disclaimer (for the AudioIO-lib, excluding winamp plugin)
//      Copyright (c) 2008...2065, Wolfgang Bscher (DL4YHF)
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are met:
//   1. Redistributions of source code must retain the above copyright notice,
//      this list of conditions and the following disclaimer.
//   2. Redistributions in binary form must reproduce the above copyright notice,
//      this list of conditions and the following disclaimer in the
//      documentation and/or other materials provided with the distribution.
//   3. Neither the name of the author nor the name of other contributors
//      may be used to endorse or promote products derived from this software
//      without specific prior written permission.
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
//  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
//   OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
//   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
//   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
// STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
// EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//
// Revision history:
//  V0.03, 2013-11-10:  Fixed the 'buffer usage' bargraph display, which used
//                      to show 'zero' when not a single audio DESTINATION
//                      was attached to a certain stream (identified by name).
//                      Now, if there is NO DESTINATION, the buffer usage
//                      will increment until it reaches the top after a few
//                      hundred milliseconds, showing that 'the buffer is full'
//                      because there's nobody draining it.
//                      The 'buffer usage' in percent is calculated
//                      in AIO_WriteOutputSamplePoints(). Details there..
//
//  V0.02, 2011-10-25:  Added a 'dummy ASIO driver' inside this Audio-I/O-DLL,
//                      which allows any application which supports
//                      Steinberg's ASIO to receive its input from the
//                      Audio-I/O-DLL (at least, from in_AudioIO.dll) .
//                      All ASIO-specific modules are in the subfolder ASIO.
//                      To compile this DLL *WITH* ASIO support, you will need
//                      a few header files from Steinberg, and define the macro
//                      'SWI_SUPPORT_ASIO' somewhere in the project settings.
//          (BCB6 deutsch: "Projekt".."Optionen".."Verzeichnisse/Bedingungen")
//
//
//
//  V0.01, 2011-07-03:  Added new functions to support the 'T_ChunkInfo'
//                      structure in the DLL; see details about that
//                      in c:\cbproj\SoundUtl\ChunkInfo.h .
//                + Buffer (shared memory) for ONE WRITER and MULTIPLE READERS,
//                  which allow multiple instances of Spectrum Lab to take their
//                  input from the SAME SOURCE, which typically is the FIRST
//                  running instance, to preprocess the data only ONCE .
//                      The buffered data type is now 32-bit floating point,
//                      usually normalized to +/- 1.0, for simplicity and to
//                      avoid data conversions between 'int' and 'float' .
//  V0.00, 2008-03-27:  Project started as a WINAMP-"input"-plugin
//                      to send audio data from DL4YHF's "Spectrum Lab"
//                      into Winamp, to play with Icecast/Shoutcast streaming .
//---------------------------------------------------------------------------

#define L_ALLOW_TESTING 1  /* 1=yes, 0=no */
#define USE_DEBUG 1  // 0=no debugging ("release"),   1=create "debug" version

#include "aio_libs.h"  // includes all headers required for the IMPLEMENTATION
                       // of an 'audio-I/O-DLL' (not the DLL/host-interface ! )


//---------------------------------------------------------------------------
#include <shlobj.h>
#include "in2.h"   // contains typedef struct In_Module for winamp (c) Nullsoft

#include "Goodies.h" // stuff like "RotatedTextOut()" in ??\SoundUtl\Goodies.c

// ex: #include "SoundMaths.h" // special math functions: FFT for the spectrum display
// (for some reason, the inclusion of SoundMaths.c increased the size of the DLL
//  from 80 kByte to 169 kByte, when compiled with Borland C++Builder V4.
//  Extracting only the REQUIRED functions (Real FFT + Hanning Window)
//  from SoundMaths.c, and implementing them in extra modules,
//  reduced the size of the DLL from 169 kByte to 83 kByte ! Linker trouble ?)

#include "CalcRealFft.h"   // FFT with REAL input and COMPLEX output
#include "HanningWindow.h" // multiply an array of floats with a Hanning window
#ifndef _CHUNK_INFO_H_
# include "ChunkInfo.h" // T_ChunkInfo in c:\cbproj\SoundUtl\ChunkInfo.h
#endif
#include "AudioIO.h" // Prototypes, data types and error codes for "Audio I/O"

//---------------------------------------------------------------------------
// Debugging stuff:
#if(1)
  int DLL_iSourceLine = 0;
# define DOBINI() {DLL_iSourceLine=__LINE__;}
#else
# define DOBINI() {}
#endif


//---------------------------------------------------------------------------

#define INI_FILE_NAME    "AUDIO_IO_DLLS.INI"  /* -> c:\WINDOWS\AUDIO_IO_DLLS.INI */
#define INI_FILE_SECTION "AudioIO_to_Winamp"
#define WINAMP_PLUGIN_VERSION "v2.7"  // software version of this module FOR WINAMP'S POINT OF VIEW (!)
        // 2012-06-08 : Changed WINAMP_PLUGIN_VERSION to "v2.7"
        //              because Spectrum Lab is Version 2.7x now, too.

const char APPL_szWindowClassName[] = "in_AudioIO_DLL";
const char APPL_szWindowCaption[] = "Audio-I/O DLL Panel";
const char APPL_szIniFileName[]   = INI_FILE_NAME;
HINSTANCE  APPL_hInstance;  // here: DLL instance handle
char       APPL_szDllNameAndPath[_MAX_PATH+4]; // complete path + name of this DLL

// post this to the main window at end of file (after playback has stopped)
#define WM_WA_MPEG_EOF WM_USER + 2

// variables..
static char WA_szAppName[] = "WINAMP<-SpecLab Input Driver";
extern In_Module WA_sInModule;  // not really "extern" but "forward"
static int  g_iWinampAudioHandle = -1; // audio source (INDEX) used by winamp; <0 when "not open"
static char WA_sz255PlayingFileName[256];
static BOOL WA_fTerminateThead = FALSE; // flag to terminate the worker thread
static HANDLE WA_hInputThread  = INVALID_HANDLE_VALUE;
BOOL WA_fDump2File = FALSE;
#if(USE_DEBUG)
 BOOL WA_fDebug  = FALSE;
#endif

// Sample buffer for Winamp's processing chunks.  Found in some demo:
// > render 576 samples into buf.
// > note that if you adjust the size of sample_buffer, for say, 1024
// > sample blocks, it will still work, but some of the visualization
// > might not look as good as it could. Stick with 576 sample blocks
// > if you can, and have an additional auxiliary (overflow) buffer if
// > necessary..
// WB: ok, so let's use 576 samples per chunk.
//     Times TWO for "stereo", times TWO for the DSP
//      (which, as another demo says, may use double the size of the buffer) .
// So we end up with 576*2*2 "single samples", 16 bit each .
// The total buffer size is 576 samples * 2 channels * 2 for the DSP * 2 bytes per 16-bit
//     = 4608 bytes then.
#define C_WINAMP_MAX_SAMPLES_PER_CHUNK  576  /* tried: 576, (576*8) */
#define C_WA_SAMPLE_BUFFER_SIZE (C_WINAMP_MAX_SAMPLES_PER_CHUNK*2*2)
short WA_i16SampleBuffer[C_WA_SAMPLE_BUFFER_SIZE+1]; // this is the non-circular buffer for WINAMP
  // the VERY LAST 16-bit word in this buffer is used for debugging purposes !
  // It should NEVER be overwritten; if it does, it's a bug .
#define WA_BUF_MAGIC_VALUE 0x7ABC
#define L_DEBUG_AUDIO_THREAD  0   /* use "SanityCheck()" ? 0=no, 1=yes */
#if( L_DEBUG_AUDIO_THREAD )
 long SanityCheckFailedInLine = 0;
#endif


static BOOL fWaitingInMessageBox, fRegisteredClassName, fActingAsWriter;
static int  iThisInstancesDllInstanceIndex;
static int  iThisInstancesGuiUpdateCounter;
static int  iSelectedAudioSource = 0; // .. for the control panel
static int  iSelectedAudioReader = 0; // .. for the control panel
static HWND g_hwndControlPanel, g_hwndControlPanelsParent;
static tpCallbackFromDLL pCallbackForControlPanel;
static void *    pCallbackUserDataForControlPanel;
static int  g_TextHeight_px = 8;  // height of a character for the parameter display
// Old 'peak values', displayed in the peak meters (bargraph display) :
static float sfltOldPeakValue[AUDIO_IO_MAX_CHANNELS]={-1,-1,-1,-1};

//T_AudioMsgBuffer *W2S_pInputBuffer=NULL; // output buffer for sending audio messages
//HWND W2S_hwndSpecLabWindow=NULL;



//---------------------------------------------------------------------------
static T_AudioSourceInfo g_DummySourceInfo =
{ // Also used to avoid having to check for NULL-pointers, by using this object
  // for any kind of 'unknown' channel, when returning the address of an T_AudioSourceInfo object.
  "Test",  // char   sz40AudioStreamID[44]
  "Default, used if no audio source is available",  // char sz80Description[84]
  48000.0, // double dblSampleRate
  2,       //  int   nChannelsPerSamplePoint
  0        //  int   iAudioPlayerDisableOutput
}; // end g_DummySourceInfo



static T_ControlPanelInfo g_ControlPanelInfo =
{
  10,  10,     // window position for the control panel
  DISPLAY_MODE_SHOW_CONNECTIONS // display mode

}; // end g_ControlPanelInfo

static HANDLE hMapObject_SharedMemoryData = NULL; // handle to file mapping object (for shared memory)
static T_SharedMemoryData * g_pSharedMemoryData = NULL; // pointer to shared memory !
static T_AudioBuffer* g_pAudioBuffers[AUDIO_MAX_SOURCES] // 8(?) pointers to shared memory..
 = { NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL };
static HANDLE hMapObject_AudioBuffers[AUDIO_MAX_SOURCES] // 8(?) handles to shared memory..
 = { NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL };
//  For each application instance ("process") which
//  uses this DLL, there will be a different set, with different addresses,
//  of the variable declared above. Pointers (addresses) to memory
//  in one process are absolutely meaningless in another process ! !


//---------------------------------------------------------------------------
INT32 PreciseToNominalSampleRate( double dblPreciseSampleRate )
  // Returns a "nominal" sampling rate
  // (5512, 11025, 22050, 32000, 44100, 48000, 96000, or 192000 samples/second)
  // If the "precise" samping rate is none of these values,
  //  it will be returned (rounded to the nearest integer value) .
{
  if( dblPreciseSampleRate >= 5400 && dblPreciseSampleRate <= 5600 )
    return 5512;
  if( dblPreciseSampleRate >= 10900 && dblPreciseSampleRate <= 11100 )
    return 11025;
  if( dblPreciseSampleRate >= 22000 && dblPreciseSampleRate <= 22100 )
    return 22050;
  if( dblPreciseSampleRate >= 31950 && dblPreciseSampleRate <= 32050 )
    return 32000;
  if( dblPreciseSampleRate >= 44000 && dblPreciseSampleRate <= 44200 )
    return 44100;
  if( dblPreciseSampleRate >= 47800 && dblPreciseSampleRate <= 48200 )
    return 48000;
  if( dblPreciseSampleRate >= 95800 && dblPreciseSampleRate <= 96200 )
    return 96000;
  if( dblPreciseSampleRate >= 191500 && dblPreciseSampleRate <= 192500 )
    return 192000;
  return (INT32)(dblPreciseSampleRate+0.499);
} // end PreciseToNominalSampleRate()

//---------------------------------------------------------------------------
BOOL WritePrivateProfileInt(char *pszSectionName, char *pszKeyName, int iValue, char *pszIniFileName )
  // Why is there a "GetPrivateProfileInt()" in the windoze API
  //     but no equivalent "Set..." or "Write.." function ?  Tsk, tsk... ;o)
{
  BOOL fOk;
  char sz80Temp[84];
  wsprintf(sz80Temp, "%d", iValue );
  fOk = WritePrivateProfileString(pszSectionName, pszKeyName, sz80Temp, pszIniFileName);
  return fOk;
}

//---------------------------------------------------------------------------
static void ClearEventHistory(void)
{
  if(g_pSharedMemoryData != NULL )
   { g_pSharedMemoryData->iEventHistoryHeadIndex  = 0;
     g_pSharedMemoryData->iEventHistoryNumEntries = 0;
     ++g_pSharedMemoryData->iGuiUpdateCounter; // let all control panels update their GUIs
   }
} // end ClearEventHistory()

//---------------------------------------------------------------------------
static void AppendToEventHistory( char *pszMessage )
{
  int i;
  T_EventHistoryEntry *pEntry;

  if(g_pSharedMemoryData != NULL )
   { i = g_pSharedMemoryData->iEventHistoryHeadIndex; // Runs from 0 ... (NUM_EVENT_HISTORY_ENTRIES-1) .
     if( i<0 || i>=NUM_EVENT_HISTORY_ENTRIES )
      { i = 0;
      }
     pEntry = &g_pSharedMemoryData->EventHistory[i];
     // Note: The event-history should be protected by a critical section.
     // But since a critical section created in application (process) A cannot be used
     // in application B, we *hope* that entries into the event history
     // as rare as they should be (ideally, there are no entries/errors at all)
     // so we write new entries from different threads, and different processes:
     // > A Win32 CriticalSection is for intra-process synchronization
     // > (and is much faster as far as lock times),
     // > however it cannot be shared across processes.
     strncpy( pEntry->sz255Text, pszMessage, 255 );
     pEntry->sz255Text[255] = '\0';  // always keep "C" strings terminated !
     GetSystemTime( &pEntry->system_time );
     g_pSharedMemoryData->iEventHistoryHeadIndex = (i+1) & (NUM_EVENT_HISTORY_ENTRIES-1);
     if( g_pSharedMemoryData->iEventHistoryNumEntries < NUM_EVENT_HISTORY_ENTRIES )
      { ++g_pSharedMemoryData->iEventHistoryNumEntries;
      }
   }
} // end AppendToEventHistory()

//---------------------------------------------------------------------------
void ShowEventHistory( HDC hDC, int x1, int y1, int w, int h )
{
  int i,n,sl,tail_index,y;
  T_EventHistoryEntry *pEntry;
  char sz255[256];

  if( (g_pSharedMemoryData!=NULL) && (g_TextHeight_px>0) )
   {
     n = g_pSharedMemoryData->iEventHistoryNumEntries; // number of history buffer entries stored
     // Only show the MOST RECENT entries here (for simplicity) :
     i = h / g_TextHeight_px;
     if( n>i )
      {  n=i;
      }
     i = g_pSharedMemoryData->iEventHistoryHeadIndex; // Runs from 0 ... (NUM_EVENT_HISTORY_ENTRIES-1)
     tail_index = i-n;
     if( tail_index < 0 )
      {  tail_index += NUM_EVENT_HISTORY_ENTRIES;    // circular buffer wrapped
      }
     y = y1;
     for(i=0; i<n; ++i )
      {
        if(tail_index >= NUM_EVENT_HISTORY_ENTRIES)
         { tail_index = 0;
         }
        pEntry = &g_pSharedMemoryData->EventHistory[tail_index++];
        sprintf( sz255, "%02d:%02d:%02d.%02d ", // time from GetSystemTime()..
          (int)pEntry->system_time.wHour,
          (int)pEntry->system_time.wMinute,
          (int)pEntry->system_time.wSecond,
          (int)pEntry->system_time.wMilliseconds/10 );
        strncat( sz255, pEntry->sz255Text, 80 );
        sz255[80] = '\0';
        SimpleTextOut(hDC, sz255, 0, y );
        y += g_TextHeight_px;
      }
   }
} // end ShowEventHistory()


//---------------------------------------------------------------------------
static void UnmapAudioBufferFromMemory( int iAudioSource )
{
  void *pv;
  if( iAudioSource>=0 && iAudioSource<AUDIO_MAX_SOURCES )
   {
     pv = (void*)g_pAudioBuffers[iAudioSource] ;
     if( pv != NULL )
      { g_pAudioBuffers[iAudioSource] = NULL;  // mark this entry as "invalid"
        Sleep(50); // allow functions in other threads (which USED this buffer) to stop using it
        UnmapViewOfFile(pv); // Unmap shared memory from the process's address space.
      }

     if( hMapObject_AudioBuffers[iAudioSource] != NULL )
      { // Close the process's handle to the file-mapping object.
        CloseHandle(hMapObject_AudioBuffers[iAudioSource]);
        hMapObject_AudioBuffers[iAudioSource] = NULL;
      }
   }

} // end UnmapAudioBufferFromMemory()

//----------------------------------------------------------------------
static void SetFlagToUpdateControlPanels(void)
{
  if(g_pSharedMemoryData != NULL )
   { ++g_pSharedMemoryData->iGuiUpdateCounter; // let all control panels update their GUI
   }
}

//----------------------------------------------------------------------
int AudioBuf_GetChunkInfoIndexForSampleIndex(
         T_AudioBuffer *pAudioBuffer,  // [in] address of the audio buffer
         DWORD dwAudioBufArrayIndex,   // [in] index into pAudioBuffer->fltBuffer[]
         int *p_iOffsetInAudioSamples) // [out,optional] offset for adjustment (*)
   // Return value: Index 'i' into pAudioBuffer->ChunkInfoArray[i]
   //                          and pAudioBuffer->ChunkInfoLUT[i] .
   //        A NEGATIVE return value means 'unable to retrieve a T_ChunkInfo' .
   // (*) In some cases, there may not be a T_ChunkInfo available for the
   //     'exact' audio sample index. In that case, *pdblOffsetInAudioSamples
   //     contains the offset, measured as a NUMBER OF AUDIO SAMPLE POINTS,
   //     between dwAudioBufArrayIndex and the "nearest" point for which
   //     a T_ChunkInfo has been emitted by the audio stream producer .
   //  Example: Audio stream producer has emitted T_ChunkInfo entries
   //           for the sample in pAudioBuffer->fltBuffer[0], pAudioBuffer->ChunkInfoArray[0];
   //          and the next in pAudioBuffer->fltBuffer[1024], pAudioBuffer->ChunkInfoArray[1].
   //    Caller (audio stream reader) wants to know the T_ChunkInfo
   //           for the audio-sample-point in pAudioBuffer->fltBuffer[4]
   //           (assume ONE channel, i.e. ONE array-index per sample).
   //   AudioBuf_GetChunkInfoIndexForSampleIndex() will return
   //           i32ChunkInfoArrayIndex = 0,
   //           and *pdblOffsetInAudioSamples = 4.0 which means the caller
   //           must add FOUR TIMES the audio-sample-time to the timestamp
   //           in the info-chunk to have the best possible match.
   // Note: Similar principle used in aio2winamp.c and SpecLab\BigBuf.c ..
{
  int  i, iLUT, iBestChunkInfo, iPrecChunkInfo, iSuccChunkInfo, iChunkInfoArrayIndex=-1;
  long i32SampleIndexDiff, i32BestSampleIndexDiff;

  if( p_iOffsetInAudioSamples != NULL )
   { *p_iOffsetInAudioSamples = 0; // don't try to adjust anything
   }
  if( pAudioBuffer!=NULL )
   {
     // To retrieve the 'closest possible' chunk-info for a given audio-sample
     //  quickly ( based on the audio sample's index into fltBuffer[] ),
     //  the index into pAudioBuffer->fltBuffer[] is divided
     //   by AUDIO_SAMPLES_PER_INFO_BLOCK, which gives a 'coarse' estimate
     //  of an index into T_AudioBuffer.ChunkInfoLUT[iLUT] :
     iLUT = dwAudioBufArrayIndex / AUDIO_SAMPLES_PER_INFO_BLOCK;
     if( iLUT >= AUDIO_MAX_CHUNK_INFO_BLOCKS )  // oops..
      {  iLUT  = AUDIO_MAX_CHUNK_INFO_BLOCKS-1; // ..fixme ! (set breakpoint here)
      }
     // Beware, the audio-stream-writer may not have emitted a T_ChunkInfo
     // close enough for this index (in fact, he may not emit 'chunk infos' at all).
     //  But in most cases, T_AudioBuffer.ChunkInfoLUT[i].iChunkArrayIndex
     // tells us where to find the CLOSEST AVAILABLE T_ChunkInfo :
     iBestChunkInfo = pAudioBuffer->ChunkInfoLUT[iLUT].iChunkArrayIndex;
     if( iBestChunkInfo>=0 && iBestChunkInfo<AUDIO_MAX_CHUNK_INFO_BLOCKS )
      { i32BestSampleIndexDiff =
         (long)dwAudioBufArrayIndex - pAudioBuffer->ChunkInfoLUT[iBestChunkInfo].iAudioBufferIndex;
      } // end if < entry in the lookup-table seems to be valid >
     else
      { i32BestSampleIndexDiff = AUDIO_MAX_CHUNK_INFO_BLOCKS; // completely unusable
      }
     i = iLUT-1;   // check out the PRECEDING Chunk-Info-entry ...
     if(i<0) i=AUDIO_MAX_CHUNK_INFO_BLOCKS-1; // circular wrap for preceding entry
     iPrecChunkInfo = pAudioBuffer->ChunkInfoLUT[i].iChunkArrayIndex;
     if( iPrecChunkInfo>=0 && iPrecChunkInfo<AUDIO_MAX_CHUNK_INFO_BLOCKS )
      { i32SampleIndexDiff =
         (long)dwAudioBufArrayIndex - pAudioBuffer->ChunkInfoLUT[iPrecChunkInfo].iAudioBufferIndex;
        if( abs(i32SampleIndexDiff) < abs(i32BestSampleIndexDiff) )
         { iBestChunkInfo = iPrecChunkInfo;
           i32BestSampleIndexDiff = i32SampleIndexDiff;
         }
      } // end if < preceding entry in the lookup-table seems to be valid >
     i = iLUT+1;   // check out the SUCCEDING Chunk-Info-entry ...
     if(i>=AUDIO_MAX_CHUNK_INFO_BLOCKS) i=0; // circular wrap for successive entry
     iSuccChunkInfo = pAudioBuffer->ChunkInfoLUT[i].iChunkArrayIndex;
     if( iSuccChunkInfo>=0 && iSuccChunkInfo<AUDIO_MAX_CHUNK_INFO_BLOCKS )
      { i32SampleIndexDiff =
         (long)dwAudioBufArrayIndex - pAudioBuffer->ChunkInfoLUT[iSuccChunkInfo].iAudioBufferIndex;
        if( abs(i32SampleIndexDiff) < abs(i32BestSampleIndexDiff) )
         { iBestChunkInfo = iSuccChunkInfo;
           i32BestSampleIndexDiff = i32SampleIndexDiff;
         }
      } // end if < entry in the lookup-table seems to be valid >
     // At this point, pAudioBuffer->ChunkInfoArray[iBestChunkInfo] is "closest"
     // to the audio sample in pAudioBuffer->fltBuffer[dwAudioBufArrayIndex] .
     // BUT: Only use this info if it's less than half the buffer-capacity
     //      away from the requested sample index !
     if( abs(i32BestSampleIndexDiff) < (AUDIO_MAX_CHUNK_INFO_BLOCKS/2) )
      { iChunkInfoArrayIndex = iBestChunkInfo;
        if( pAudioBuffer->nChannelsPerSamplePoint > 1 )
           i32BestSampleIndexDiff /= pAudioBuffer->nChannelsPerSamplePoint;
        if( p_iOffsetInAudioSamples != NULL )
           *p_iOffsetInAudioSamples = i32BestSampleIndexDiff;
      }
   }
  return iChunkInfoArrayIndex;
} // end AudioBuf_GetChunkInfoIndexForSampleIndex()


//----------------------------------------------------------------------
T_ChunkInfo * AudioBuf_GetLatestChunkInfo(
         T_AudioBuffer *pAudioBuffer) // [in] address of the audio buffer
  // Returns the address of the most recent T_ChunkInfo structure
  //   which the 'audio stream producer' has hopefully sent
  //   along with a 'chunk' of audio samples .
  // The result may be NULL if there is no 'T_ChunkInfo' available yet !
  // Used by the graphic user interface since 2011-09 ( see OnWmPaint ).
{
  int i;
  if( pAudioBuffer!=NULL )
   { // Note: the entry with the highest index in pAudioBuffer->ChunkInfoArray[]
     //       is not the 'latest' (newest) audio-chunk-info !
     i = pAudioBuffer->iLatestChunkInfoArrayIndex;
     if( (i>=0) && (i<AUDIO_MAX_CHUNK_INFO_BLOCKS) )
      { return &pAudioBuffer->ChunkInfoArray[i];
      }
   }
  return NULL;
} // end AudioBuf_GetLatestChunkInfo()

//----------------------------------------------------------------------
BOOL AudioBuf_GetAdjustedChunkInfoForSampleIndex(
         T_AudioBuffer *pAudioBuffer, // [in] address of the audio buffer
         T_AudioSourceInfo *pAudioSourceInfo, // [in] additional info about the audio source
         DWORD dwAudioBufArrayIndex,  // [in] index into pAudioBuffer->fltBuffer[]
         T_ChunkInfo *pDstChunkInfo)  // [out] chunk info (caller's var-space) (*)
   // Note: Similar principle used in aio2winamp.c and SpecLab\BigBuf.c ..
  // (*) this "audio chunk info" structure MUST HAVE ALREADY BEEN INITIALIZED
  //     by the caller. If it's not initialized (i.e. "not valid upon entry"),
  //     we won't touch it here to avoid access violations with incompatible
  //     versions of the T_ChunkInfo struct. Details in ?\SoundUtl\ChunkInfo.h .
{
  int iChunkInfoArrayIndex, iOffsetInAudioSamples;
  double dblOffsetInSeconds;

  if( pAudioBuffer!=NULL && pDstChunkInfo!=NULL )
   { if( pDstChunkInfo->dwSizeOfStruct == sizeof(T_ChunkInfo) )
      { // ok, we're "compatible" with the caller's version of a T_ChunkInfo !
        iChunkInfoArrayIndex = AudioBuf_GetChunkInfoIndexForSampleIndex(
                pAudioBuffer, dwAudioBufArrayIndex, &iOffsetInAudioSamples );
        if(  (iChunkInfoArrayIndex>=0)
          && (pAudioBuffer->ChunkInfoArray[iChunkInfoArrayIndex].dwSizeOfStruct==sizeof(T_ChunkInfo))
          )
         { *pDstChunkInfo = pAudioBuffer->ChunkInfoArray[iChunkInfoArrayIndex];
           // Need to adjust a few members in the T_ChunkInfo,
           //  to compensate an offset of a few audio samples ?
           if( iOffsetInAudioSamples != 0 ) // yes, adjust.. see ChunkInfo.h ...
            { pDstChunkInfo->i64TotalSampleCounter += iOffsetInAudioSamples;
              if( pDstChunkInfo->dblPrecSamplingRate>0.0 )
               { dblOffsetInSeconds = (double)iOffsetInAudioSamples/pDstChunkInfo->dblPrecSamplingRate;
                 pDstChunkInfo->rtRecTime          += dblOffsetInSeconds;
                 pDstChunkInfo->ldblUnixDateAndTime += dblOffsetInSeconds;
                 // do NOT try to adjust the GPS data here !
               }
            }
           return TRUE;
         } // end if < AudioBuf_GetChunkInfoIndexForSampleIndex() successful >
        else // there is no valid T_ChunkInfo for the requested index,
         {   // but the caller expects one
             // -> prepare at least a "minimum T_ChunkInfo"
             //    as described in ..\SoundUtl\ChunkInfo.h :
           pDstChunkInfo->dwValidityFlags = 0; // we know 'almost nothing' yet..
           // pDstChunkInfo->dwNrOfSamplePoints = ?;        // must be set by caller
           // pDstChunkInfo->nChannelsPerSample = ?; // must be set by caller
           pDstChunkInfo->dblSampleValueRange = pAudioSourceInfo->dblSampleValueRange;
         }
      } // end if < version of the caller's T_ChunkInfo is compatible with this DLL >
   }
  return FALSE;
} // end AudioBuf_GetAdjustedChunkInfoForSampleIndex()

//----------------------------------------------------------------------
// DLL initialisation / termination code .
//   > DllMain is a placeholder for the library-defined function name.
//   > You must specify the actual name you use when you build your DLL.
//   > For more information, see the documentation
//   > included with your development tools.           ;o)
// In other words, every Compiler/IDE-writer cooked his own soup.
//----------------------------------------------------------------------
#ifdef __GNUC__
 BOOL APIENTRY DllMain ( HINSTANCE instance, DWORD fdwReason, LPVOID lpvReserved  )
 // "DllEntryPoint()" was never called when compiling this unit with DevCpp. WHY ?
#endif
#ifdef __BORLANDC__
__declspec(dllexport) BOOL WINAPI DllMain(HINSTANCE instance, DWORD fdwReason, LPVOID lpvReserved)
 // ex: int WINAPI DllEntryPoint(HINSTANCE hInst, unsigned long reason, void* reserved)
 // This function was called when the DLL was compiled with Borland C++Builder.
#endif
{
  int i, iAudioSource, iReaderIndex;
  T_AudioReader *pReader;
  T_AudioBuffer *pAudioBuffer;
  T_AudioSourceInfo *pSourceInfo;
  char sz40Section[44], sz255Data[256];
  // bool initproc=0;
  BOOL first_instance;
  APPL_hInstance = instance; // in a normal windoze app, this happens in WinMain()
  GetModuleFileName( instance, APPL_szDllNameAndPath, _MAX_PATH); // retrieve complete path + name of this DLL

  (void)lpvReserved; // unused parameter - so what ?

  switch(fdwReason)
   {
     case DLL_PROCESS_ATTACH /* 1 */ :
        // > The DLL is being loaded into the virtual address space
        // > of the current process as a result of the process starting up
        // > or as a result of a call to LoadLibrary.
        // > DLLs can use this opportunity to initialize any instance data (..)
        //  Use a memory-mapped "file" instead of SHARED MEMORY (not a "real file")
        hMapObject_SharedMemoryData = CreateFileMapping(
                 INVALID_HANDLE_VALUE, // don't use a "real file" (but what ??)
                 NULL,                 // default security attributes
                 PAGE_READWRITE,       // we want read/write access
                 0,                    // upper 32 bits of the size
                 sizeof(T_SharedMemoryData), // lower 32 bits of the size
                 "AudioIO_SourceTable"); // name of the mapped object
        if(!hMapObject_SharedMemoryData)
         { return FALSE;  // bad luck; couldn't create the "file"-mapping object.
                          // Another bad surprise from "Vista" or Windows 7 ?
         }
        // About CreateFileMapping :
        // > If the object exists before the function call, the function returns
        // > a handle to the existing object (with its current size,
        // > not the specified size), and GetLastError returns
        // > ERROR_ALREADY_EXISTS.
        // Here, we'll use this feature to find out if this is the
        //   FIRST instance (of the DLL being loaded into memory) or not:
        first_instance = (GetLastError() != ERROR_ALREADY_EXISTS);
        g_pSharedMemoryData = MapViewOfFile(hMapObject_SharedMemoryData, FILE_MAP_WRITE,0,0,0);
        if( g_pSharedMemoryData==NULL )
         { // bad luck; couldn't create a "view of the file" in *MEMORY*.
           // Yet another bad surprise from "Vista" or Windows 7 ?
           CloseHandle(hMapObject_SharedMemoryData);  // not sure if this is correct...
           return FALSE;
         }

        // If (and ONLY if) this is the first instance, load a few parameters
        // from a configuration file.  We use an ancient INI file for this
        // to avoid fooling around with the windoze registry .
        // The file will be in c:\WINDOWS\AUDIO_IO_DLLS.INI  .
        // >  Win32-based applications should store initialization information in the registry.
        // Just kidding, eh ? We DON'T MESS AROUND WITH THE STUPID REGISTRY here.
        if( first_instance )  // time to load the config !
         { g_pSharedMemoryData->iDllInstanceCounter = 1;
           g_pSharedMemoryData->iGuiUpdateCounter   = 0;
           g_pSharedMemoryData->iWinampCallFlags  = 0;
           memset( g_pSharedMemoryData->sz255WinampLastEvent, 0, 255 );
           // TEST: strcpy( g_pSharedMemoryData->sz255WinampLastEvent, "Winamp didn't call the DLL yet" );
           for( i=0; i<AUDIO_MAX_SOURCES ; ++i)
            { sprintf( sz40Section, "AudioSource%d", (int)i+1 );
              pSourceInfo = &g_pSharedMemoryData->source[i];
              sprintf( sz255Data, "Stream%d", (int)i+1 );
              GetPrivateProfileString( sz40Section,
                  "StreamID", sz255Data, pSourceInfo->sz40AudioStreamID, 40, APPL_szIniFileName);
              GetPrivateProfileString( sz40Section,
                  "Description", "", pSourceInfo->sz80Description, 80, APPL_szIniFileName);
              pSourceInfo->dblSampleRate = GetPrivateProfileInt( sz40Section,
                 "iSampleRate", (int)48000,  APPL_szIniFileName );
              pSourceInfo->nChannelsPerSamplePoint = GetPrivateProfileInt( sz40Section,
                 "nChannelsPerSamplePoint", (int)2,     APPL_szIniFileName );
              pSourceInfo->iAudioPlayerDisableOutput = GetPrivateProfileInt( sz40Section,
                 "iDisableOutput", (int)0,   APPL_szIniFileName );
              g_pSharedMemoryData->iDayLastUsed[i] = GetPrivateProfileInt( sz40Section,
                 "iDayLastUsed",   (int)0,   APPL_szIniFileName );
            }
           WA_i16SampleBuffer[C_WA_SAMPLE_BUFFER_SIZE] = WA_BUF_MAGIC_VALUE;
         } // end if( first_instance )
        else  // this is NOT the first instance of the DLL in memory:
         { ++g_pSharedMemoryData->iDllInstanceCounter;
           ++g_pSharedMemoryData->iGuiUpdateCounter; // let all control panels update their GUI
         }
        // Initialisation of a few internal variables FOR THE NEW INSTANCE(!) :
        strcpy(sz40Section,"AudioIOControlPanel");
        g_ControlPanelInfo.iX1 = ReadIntegerFromIniFile( (char*)APPL_szIniFileName, sz40Section, "x1", 10/*default*/ );
        g_ControlPanelInfo.iY1 = ReadIntegerFromIniFile( (char*)APPL_szIniFileName, sz40Section, "y1", 10/*default*/ );
        g_ControlPanelInfo.iDisplayMode = ReadIntegerFromIniFile( (char*)APPL_szIniFileName, sz40Section, "DisplayMode", 0/*default*/ );
        g_hwndControlPanel = g_hwndControlPanelsParent = NULL;
        pCallbackForControlPanel = NULL;
        pCallbackUserDataForControlPanel = NULL;
        fWaitingInMessageBox =  fRegisteredClassName = fActingAsWriter = FALSE;
        iThisInstancesDllInstanceIndex = g_pSharedMemoryData->iDllInstanceCounter;
        iThisInstancesGuiUpdateCounter = 0;
        break;

     case DLL_THREAD_ATTACH:
        //DisableThreadLibraryCalls(hinst);
        break;

     // The thread of the attached process terminates.
     case DLL_THREAD_DETACH:
        break;
     case DLL_PROCESS_DETACH /* 0 */ :
         // Clean up "instance data" ?
         if( g_hwndControlPanel != NULL )
          {  // Better don't call DestroyWindow() here, because:
             // > Windows are destroyed inside the message loop,
             // > not after exiting the message loop .
      // useless: CloseWindow( g_hwndControlPanel ); // STRANGE THINGS HAPPEN (*)
              SendMessage( g_hwndControlPanel, WM_CLOSE,0,0 ); // << see WndProc()..
                  // At this point, if all went well,
                  // g_hwndControlPanel should already be NULL - see WndProc !
              if( g_hwndControlPanel != NULL )
               { DestroyWindow( g_hwndControlPanel ); // shouldn't get here but WHO KNOWS !
                 g_hwndControlPanel = NULL;  // Is the window handle invalid "after destruction" ? Guess so..
               }
              // The call stack (to get to THIS point) was:
              //   TAudioIODLLControlForm::TestAudioIODLL() [in SpecLab]
              //    -> AIO_Host_DeleteInstanceData()
              //        -> AIO_Host_FreeDLL()
              //           -> C:\WINDOWS\system32\kernel32.dll
              //               -> "ntdll.dll"
              //                    -> "std" (wtf is that ?)
              //                        -> "__startupd" ("" ?)
              //                            -> DllEntryPoint(hInst=0x06760000, reason=0, reserved=NULL)
          } // end if < this instance had opened a CONTROL PANEL >

         if( fRegisteredClassName )
          {  // See AIO_ShowControlPanel() : our window class must be 'unregistered' (AFTER closing the window)
             // Because this is a DLL, the system will not do that automatically (unlike in an EXE) !
             fRegisteredClassName = FALSE;
             UnregisterClass( APPL_szWindowClassName, NULL );
          }

         if( g_pSharedMemoryData != NULL )
          {
           // Make sure all readers, and all writers of THIS dll-instance are closed:
           for( iAudioSource=0; iAudioSource<AUDIO_MAX_SOURCES ; ++iAudioSource)
            { pAudioBuffer = g_pAudioBuffers[iAudioSource];
              pSourceInfo = &g_pSharedMemoryData->source[iAudioSource];
              if( pAudioBuffer != NULL )
               { if( (pAudioBuffer->iWriterDllInstanceIndex == iThisInstancesDllInstanceIndex)
                  && (pAudioBuffer->dwWriterFlags & AUDIO_WRITER_OPENED ) )
                  { pAudioBuffer->dwWriterFlags &= ~AUDIO_WRITER_OPENED;
                    pSourceInfo->dwWriterFlags = pAudioBuffer->dwWriterFlags; // keep this COPY up-to-date
                    SetFlagToUpdateControlPanels();
                  }
                 for( iReaderIndex=0; iReaderIndex<AUDIO_MAX_READERS; ++iReaderIndex )
                  { pReader = &pAudioBuffer->reader[iReaderIndex];
                    if(  (pReader->iDllInstanceIndex == iThisInstancesDllInstanceIndex)
                      && (pReader->dwFlags & AUDIO_READER_OPENED ) )
                     { pReader->dwFlags  &= ~AUDIO_READER_OPENED;
                       // For DISPLAY PURPOSES, T_AudioBuffer.reader[]
                       // is duplicated in T_SharedMemoryData.T_AudioSourceInfo.reader[] :
                       g_pSharedMemoryData->source[iAudioSource].reader[iReaderIndex] = *pReader;
                       SetFlagToUpdateControlPanels();
                     }
                  } // end for < all readers of this audio-source >
               } // end if( pAudioBuffer != NULL )
            } // end for( iAudioSource .. )


           // Special actions to be taken when the LAST instance gets unloaded ("detached") ?
           --g_pSharedMemoryData->iDllInstanceCounter;
           if( g_pSharedMemoryData->iDllInstanceCounter==0 ) // no-one else is using the DLL now...
            {
              if( fWaitingInMessageBox )
               { // shit happens.. we're still waiting inside the 'About'-box !
                 MessageBeep(-1);   // helplessly beep - there's not much else we can do
                 while( fWaitingInMessageBox )
                  { Sleep(200);
                  }
               }

              // Save the configuration for the next session.
              //  Why ? An audio READER may need to know the sampling rate (etc)
              //        in advance, i.e. before the writer has opened / created a stream.
              // >  Win32-based applications should store initialization information in the registry.
              // No thanks - better keep hands off the overcrowded registry !
              strcpy(sz40Section,"About");
              sprintf(sz255Data,"Configuration file for DL4YHF's Audio-I/O-DLL, compiled %s",__DATE__ );
              WritePrivateProfileString( sz40Section, "i1", sz255Data, APPL_szIniFileName);
              WritePrivateProfileString( sz40Section, "i2", APPL_szDllNameAndPath, APPL_szIniFileName);
              WritePrivateProfileString( sz40Section, "i3", "", APPL_szIniFileName);
              WritePrivateProfileString( sz40Section, "d", "", APPL_szIniFileName);  // separator-dummy
              for( i=0; i<AUDIO_MAX_SOURCES ; ++i)
               { sprintf( sz40Section, "AudioSource%d", (int)i+1 );
                 pSourceInfo = &g_pSharedMemoryData->source[i];
                 WritePrivateProfileString( sz40Section,
                    "StreamID",    pSourceInfo->sz40AudioStreamID, APPL_szIniFileName);
                 WritePrivateProfileString( sz40Section,
                    "Description", pSourceInfo->sz80Description, APPL_szIniFileName);
                 WriteIntegerToIniFile( (char*)APPL_szIniFileName, sz40Section,
                    "iSampleRate",    (int)pSourceInfo->dblSampleRate );
                 WriteIntegerToIniFile( (char*)APPL_szIniFileName, sz40Section,
                    "nChannelsPerSamplePoint",   (int)pSourceInfo->nChannelsPerSamplePoint );
                 WriteIntegerToIniFile( (char*)APPL_szIniFileName, sz40Section,
                    "iDisableOutput", (int)pSourceInfo->iAudioPlayerDisableOutput );
                 WriteIntegerToIniFile( (char*)APPL_szIniFileName, sz40Section,
                    "iDayLastUsed", g_pSharedMemoryData->iDayLastUsed[i] );
                 WritePrivateProfileString( sz40Section, "d", "", APPL_szIniFileName);  // separator-dummy
               }
            } // end if < last "user" of the DLL says goodbye >

           // Unmap all audio buffers FOR THIS DLL INSTANCE (not "for all") :
           for(i=0; i<AUDIO_MAX_SOURCES; ++i )
            { // Delete g_pAudioBuffers[] and hMapObject_AudioBuffers[]
              // (only those which have been created by "THIS" INSTANCE of the DLL)
              UnmapAudioBufferFromMemory(i);
            }

           // Unmap shared memory from the process's address space.
           UnmapViewOfFile(g_pSharedMemoryData);

           // Close the process's handle to the file-mapping object.
           CloseHandle(hMapObject_SharedMemoryData);

#       if ( 0 &&  L_ALLOW_TESTING ) // TEST to find if we really get here (and WHEN) :
           for(i=0; i<3; ++i)
            { MessageBeep(-1);  // TEST: beep-beep-beep ("dongdongdong"): DLL finally unloaded
              Sleep(200);       // test result: got here when winamp terminated
            }
#       endif    // TEST

          } // end if( g_pSharedMemoryData != NULL )
         break;

     default:
         break;
   }

#ifdef SWI_ASIO_SUPPORTED  // does the Audio-I/O-DLL also mimick an ASIO driver ?
  ASIO_DllMain(instance, fdwReason, lpvReserved);  // -> ASIO_driver_in_c.c 
#endif // SWI_ASIO_SUPPORTED ?

  return 1;
} // end DllEntryPoint()

//---------------------------------------------------------------------------
static BOOL GetAudioBufferAndSourceInfoForWinamp(
               T_AudioBuffer **ppBuffer,      // [out] audio buffer to be used by Winamp
               T_AudioSourceInfo **ppSrcInfo) // [out] source-info to be used by Winamp
{
  unsigned int iAudioSource = (unsigned int)g_iWinampAudioHandle >> 8;

  if( g_iWinampAudioHandle<0 )   // didn't call WA_Play() yet, refuse to work...
   { return FALSE;
   }

  if( (g_pSharedMemoryData!=NULL) && (g_pAudioBuffers[iAudioSource]!= NULL) )
   { if( ppBuffer != NULL )
      { *ppBuffer  = g_pAudioBuffers[iAudioSource];
      }
     if( ppSrcInfo != NULL )
      { *ppSrcInfo = &g_pSharedMemoryData->source[iAudioSource];
      }
     return TRUE;
   }
  else
   { return FALSE;
   }
} // end GetAudioBufferAndSourceInfoForWinamp()

//---------------------------------------------------------------------------
static T_AudioSourceInfo *GetAudioSourceInfoForGUI(void) // NEVER returns NULL - pointer ALWAYS valid
{
  int i;
  // If 'iSelectedAudioSource' invalid, automatically select the first valid 'source' :
  if( iSelectedAudioSource<0 || iSelectedAudioSource>=AUDIO_MAX_SOURCES )
   {  iSelectedAudioSource=0;
   }
  if( g_pAudioBuffers[iSelectedAudioSource] == NULL )
   { for(i=0; i<AUDIO_MAX_SOURCES; ++i)
      { if( g_pAudioBuffers[i] != NULL )
         { iSelectedAudioSource = i;
           break;
         }
      }
   }

  if( g_pSharedMemoryData!=NULL &&  iSelectedAudioSource>=0 && iSelectedAudioSource<AUDIO_MAX_SOURCES  )
   { return &g_pSharedMemoryData->source[iSelectedAudioSource];
   }
  return &g_DummySourceInfo;  // caller doesn't need to care about invalid pointer here
} // end GetAudioSourceInfoForGUI()

//---------------------------------------------------------------------------
static T_AudioBuffer *GetAudioBufferForGUI(void) // may return NULL (!)
{
  if( iSelectedAudioSource>=0 && iSelectedAudioSource<AUDIO_MAX_SOURCES  )
   { return g_pAudioBuffers[iSelectedAudioSource];
   }
  return NULL;
} // end GetAudioBufferForGUI()

//---------------------------------------------------------------------------
static T_ControlPanelInfo *GetControlPanelInfoForGUI(void) // NEVER returns NULL - pointer always valid !
  // Returns the address of one entry in the table of 'control panel infos',
  //   g_pSharedMemoryData->cp_info[0..AUDIO_MAX_READERS/*!!*/].
  // Used to open the control panel at the same position on the desktop,
  // stored between two sessions.
{
  return &g_ControlPanelInfo;
} // end GetControlPanelInfoForGUI()


//---------------------------------------------------------------------------
static T_AudioSourceInfo *GetAudioSourceInfoByStreamID( // may return NULL if "not found" !
        char *pszAudioStreamID , // [in] 'Stream ID' for the Audio-I/O-DLL, identifies the AUDIO SOURCE.
                                 //      use pszAudioStreamID=NULL to find the first UNUSED entry.
        int  *piAudioSource) // [out,optional] INDEX of the audio source; 0..AUDIO_MAX_SOURCES-1
{
  int iAudioSource;
  T_AudioSourceInfo *pAudioSourceInfo;
  BOOL found;
  if( g_pSharedMemoryData != NULL )
   {
     for( iAudioSource=0; iAudioSource<AUDIO_MAX_SOURCES ; ++iAudioSource)
      { pAudioSourceInfo = &g_pSharedMemoryData->source[iAudioSource];
        found = FALSE;
        if( pszAudioStreamID != NULL )   // looking for this certain ID ?
         { if(strcmp(pszAudioStreamID,pAudioSourceInfo->sz40AudioStreamID)==0)
            { // Use THIS entry for the specified stream; it may contain
              // some info about the stream from a previous session !
              found = TRUE;
            }
         }
        else  // looking for an UNUSED entry in the array :
         { // > If sz40AudioStreamID[0] == '\0', this entry has NEVER been used yet.
           if( pAudioSourceInfo->sz40AudioStreamID[0] == '\0' )
            { // this entry is UNUSED (neither at the moment, nor does it contain
              // any info about a stream which may have been used in a previous session)
              found = TRUE;
            }
         }
        if( found )
         { if( piAudioSource != NULL )
            { *piAudioSource = iAudioSource;
            }
           return pAudioSourceInfo;
         }
      }
   }
  // Arrived here: didn't find a matching entry (or unused entry) -> return NULL
  if( piAudioSource != NULL )
   { *piAudioSource = 0;
   }
  return NULL;
} // end GetAudioSourceInfoByStreamID()

//---------------------------------------------------------------------------
static long GetCurrentUnixDayNumber( void )
  // Used in this DLL to decide which of the entries in T_SharedMemoryData.source[]
  // has not been used for the longest time, and should be used when running
  // out of table entries .
{
  SYSTEMTIME st;
  GetSystemTime(&st);   // should return UTC but this can be seriously doubted
  return YearMonthDayToUnixDays( st.wYear, st.wMonth/*1..12*/, st.wDay/*1..31*/ );
} // end GetCurrentUnixDayNumber()

//---------------------------------------------------------------------------
static T_AudioBuffer * GetAudioBufferByStreamID(
        char * pszAudioStreamID, // [in,optional] stream ID' for the Audio-I/O-DLL, identifies the AUDIO SOURCE
                                 //      may be NULL to "use the first existing stream"
        int   iCreateOptions,    // [in] AUDIO_DONT_CREATE_BUFFER, AUDIO_CREATE_BUFFER_IF_NOT_EXISTING, ..
#       define AUDIO_DONT_CREATE_BUFFER            0x0000
#       define AUDIO_CREATE_BUFFER_IF_NOT_EXISTING 0x0001
        T_AudioSourceInfo **ppAudioSourceInfo, // [out,optional] address of the 'audio source info' for this stream
        int  *piAudioSource) // [out,optional] 0..AUDIO_MAX_SOURCES-1, will be part of the 'handle' later
  // Caution ! The result may be NULL if the specified audio source
  //           is unknown,  and couldn't be created (optionally) .
{
  T_AudioSourceInfo *pAudioSourceInfo;
  T_AudioBuffer     *pAudioBuffer;
  BOOL fAudioBufferIsNew;
  int iAudioSource, iAudioReader, iNewSourceIndex;
  int iDayNumber, iOldestDayNumber;
  char sz80[84], sz127Msg[128];

  if( piAudioSource != NULL )
   { *piAudioSource = 0;   // don't leave anything to fate (even when "not successful")
   }

  if( g_pSharedMemoryData != NULL )
   {
     for( iAudioSource=0; iAudioSource<AUDIO_MAX_SOURCES ; ++iAudioSource)
      { pAudioBuffer = g_pAudioBuffers[iAudioSource];
        pAudioSourceInfo = &g_pSharedMemoryData->source[iAudioSource];
        if( pAudioBuffer != NULL )
         {
           if( ppAudioSourceInfo != NULL )
            { *ppAudioSourceInfo = pAudioSourceInfo;
            }
           if( piAudioSource != NULL )
            { *piAudioSource = iAudioSource;
            }
           if( pszAudioStreamID == NULL )
            { // caller doesn't care about stream IDs -> use the 1st existing buffer
              return pAudioBuffer;
            }
           if( strcmp(pszAudioStreamID, pAudioSourceInfo->sz40AudioStreamID) == 0 )
            { // stream IDs match, and the buffer already exists -> use THIS buffer
              return pAudioBuffer;
            }
         } // end if( g_pAudioBuffers[iAudioSource] != NULL )
      } // end for( iAudioSource ..

     // Arrived here: A buffer for the specified stream-ID doesn't exist yet, so CREATE it ?
     //   (when trying to "create" the memory-mapped "file", we may find that
     //    it has already been created by some other instance. In that case,
     //    "create" means "open the same FILE object a second time" .
     //    Details on that after the call to 'CreateFileMapping()' ! )
     if( iCreateOptions & AUDIO_CREATE_BUFFER_IF_NOT_EXISTING )
      {
        // Always use the same indices (i = 0..AUDIO_MAX_SOURCES)
        // into g_pSharedMemoryData->source[i],
        //      g_pAudioBuffers[i]
        //  and hMapObject_AudioBuffers[i]  !
        iNewSourceIndex = -1;   // still unknown..
        if( pszAudioStreamID != NULL )
         { // caller wants to be connected to a CERTAIN (named) audio stream:
           pAudioSourceInfo = GetAudioSourceInfoByStreamID( // may return NULL if "not found" !
                          pszAudioStreamID, &iNewSourceIndex );
           if( pAudioSourceInfo == NULL ) // the stream-ID is completely unknown -> need a new entry.
            { //
              // Use the OLDEST entry which hasn't been used yet,
              //  or (if ALL entries have already been used already)
              //  use that entry hasn't been used for the longest time,
              //  which means 'forget the info about that stream-ID .
              iOldestDayNumber = 99999999;
              iNewSourceIndex = 0;

              for( iAudioSource=0; iAudioSource<AUDIO_MAX_SOURCES ; ++iAudioSource)
               { iDayNumber = g_pSharedMemoryData->iDayLastUsed[iAudioSource];
                 if( iDayNumber <= 0 )
                  { iNewSourceIndex = iAudioSource;
                    break;   // entry has NEVER been used -> use it now
                  }
                 else
                  { if( iDayNumber < iOldestDayNumber )
                     { iOldestDayNumber = iDayNumber;
                       iNewSourceIndex  = iAudioSource;
                     }
                  }
               }
            } // end if( pAudioSourceInfo == NULL )
         }
        else // caller wants to be connected to the FIRST EXISTING audio stream
         {   // (for example, called form WA_Play() without stream name in the pseudo-URL)
             // Note: That source (memory mapped file) may have been created
             //       in A DIFFERENT INSTANCE of this DLL !
           for( iAudioSource=0; iAudioSource<AUDIO_MAX_SOURCES ; ++iAudioSource)
            { if( g_pSharedMemoryData->source[iAudioSource].dwWriterFlags & AUDIO_WRITER_OPENED )
               { // Bingo.. found an audio source which is currently OPEN,
                 // maybe in ANOTHER INSTANCE of this DLL (doesn't matter where):
                 iNewSourceIndex  = iAudioSource;
                 // Don't try to RENAME this stream further below !
                 pszAudioStreamID = g_pSharedMemoryData->source[iAudioSource].sz40AudioStreamID;
                 break;
               }
            }
         }
        if( (iNewSourceIndex>=0) && (iNewSourceIndex<AUDIO_MAX_SOURCES) )
         { // ok, the buffer will be created at THIS index, as a SHARED MEMORY object:
           pAudioSourceInfo = &g_pSharedMemoryData->source[iNewSourceIndex];

           // Create the 'memory mapped file' (which is hopefully not a REAL file):
           if( pszAudioStreamID!=NULL && pszAudioStreamID[0]!='\0' )
            { sprintf( sz80, "AudioIOBuffer_%s", pszAudioStreamID );
            }
           else // if no 'audio stream ID' is specified, use a NUMBERING:
            { sprintf( sz80, "AudioIOBuffer_NoName%d", iNewSourceIndex );
            }
           hMapObject_AudioBuffers[iNewSourceIndex] = CreateFileMapping(
                 INVALID_HANDLE_VALUE, // don't use a "real file" (but what ??)
                 NULL,                 // default security attributes
                 PAGE_READWRITE,       // we want read/write access
                 0,                    // upper 32 bits of the size
                 sizeof(T_AudioBuffer),  // lower 32 bits of the size
                 sz80 );  // name of the mapped object (this is NOT a filename)
           // About CreateFileMapping :
           // > If the object exists before the function call, the function returns
           // > a handle to the existing object (with its current size,
           // > not the specified size), and GetLastError returns
           // > ERROR_ALREADY_EXISTS.
           // > If the function fails, the return value is NULL.
           if( hMapObject_AudioBuffers[iNewSourceIndex] != NULL ) // CreateFileMapping(): success...
            { fAudioBufferIsNew = (GetLastError()!=ERROR_ALREADY_EXISTS); // this is NOT an error
                   // (but a quite stupid way to return an 'additional function output)
              // Map the 'file' (which isn't a real file) into this processes' address space:
              pAudioBuffer = MapViewOfFile(
                   hMapObject_AudioBuffers[iNewSourceIndex], FILE_MAP_WRITE,0,0,0);
              if( pAudioBuffer!=NULL )
               { // Successfully OPENED, or NEWLY CREATED, the shared memory object.
                 // Regardless whether it already existed before, "clean" the buffer:
                 if( pszAudioStreamID != NULL )
                  { // use this NEW stream ID (if specified), also in the persistant audio source table:
                    strncpy( pAudioSourceInfo->sz40AudioStreamID, pszAudioStreamID, 40 );
                  }
                 else  // provide a 'dummy name' for the stream, which also appears in the GUI:
                  { sprintf( pAudioSourceInfo->sz40AudioStreamID, "NoName%d", (int)iNewSourceIndex+1 );
                  }
                 pAudioSourceInfo->sz40AudioStreamID[40] = '\0'; // make sure string is terminated
                 g_pSharedMemoryData->iDayLastUsed[iNewSourceIndex] = GetCurrentUnixDayNumber();
                 // Note: Do not overwrite anything else in the 'audio source table',
                 //       because some application may need to know the 'old' settings !
                 // Block-copy the 'info' from the persistent table into the newly allocated buffer:
                 // ex: pAudioBuffer->source_info = g_pSharedMemoryData->source[iNewSourceIndex];
                 // (removed this duplicate/redundant info in TWO places)
                 // Only if this audio buffer is 'new' (and not already created by any READER),
                 //  clear the table of 'audio readers' :
                 if( fAudioBufferIsNew )
                  { // This audio-stream-buffer didn't exist before the call,
                    // so clear the table of 'readers', and other potential garbage:
                    for( iAudioReader=0; iAudioReader<AUDIO_MAX_READERS; ++iAudioReader)
                     { memset( &pAudioBuffer->reader[iAudioReader], 0, sizeof(T_AudioReader) );
                     }
                    pAudioBuffer->dwWriterFlags  = 0;  // not 'opened for writing' yet (etc)
                    pAudioSourceInfo->dwWriterFlags = pAudioBuffer->dwWriterFlags; // <- just a COPY !
                    pAudioBuffer->iLastErrorCode = 0;
                    pAudioBuffer->iResetRequest  = 1; // 'reset' the buffer somewhere (later)
                    pAudioBuffer->dwHeadIndex    = 0;
                    pAudioBuffer->nSamplePointsBuffered = 0;
                    pAudioBuffer->nChannelsPerSamplePoint = pAudioSourceInfo->nChannelsPerSamplePoint;
                    pAudioBuffer->nSamplePointsMax = AUDIO_BUFFER_SIZE / pAudioSourceInfo->nChannelsPerSamplePoint;
                    pAudioBuffer->iLatestChunkInfoArrayIndex = -1; // no T_ChunkInfo stored in ChunkInfoArray[] yet
                    sprintf( sz127Msg, "Created '%s'", sz80 );
                    AppendToEventHistory( sz127Msg );
                  }
                 else // the audio buffer already existed before,
                  {   // ... so DON'T KILL OTHER READERS !
                    sprintf( sz127Msg, "Re-opened '%s'", sz80 );
                    AppendToEventHistory( sz127Msg );
                  }
                 g_pAudioBuffers[iNewSourceIndex] = pAudioBuffer; // "ready for use" now
                 if( ppAudioSourceInfo != NULL )
                  { *ppAudioSourceInfo = pAudioSourceInfo;
                  }
                 if( piAudioSource != NULL )
                  { *piAudioSource = iNewSourceIndex;
                  }
                 return pAudioBuffer;
               }
              else
               { // Bad luck; couldn't create a "view of the file" in *MEMORY*.
                 // Close the file handle again, because there's something wrong with the file
                 CloseHandle(hMapObject_AudioBuffers[iNewSourceIndex]);
                 hMapObject_AudioBuffers[iNewSourceIndex] = NULL;
                 sprintf( sz127Msg, "Couldn't map '%s'", sz80 );
                 AppendToEventHistory( sz127Msg );
               }
            } // end if( hMapObject_AudioBuffers ...
           else
            { // bad luck; couldn't create the "file"-mapping object.
              sprintf( sz127Msg, "Couldn't create '%s'", sz80 );
              AppendToEventHistory( sz127Msg );
            }
         } // end if( (iNewSourceIndex>=0) && (iNewSourceIndex<AUDIO_MAX_SOURCES) )
      } // end if( iCreateOptions & AUDIO_CREATE_BUFFER_IF_NOT_EXISTING )
   } // end if( g_pSharedMemoryData != NULL )
  if( ppAudioSourceInfo != NULL )
   { *ppAudioSourceInfo = NULL;
   }
  return NULL;
} // end GetAudioBufferByStreamID()


//---------------------------------------------------------------------------
static T_AudioBuffer * GetAudioBufferByHandle( // may return NULL !
                          int iAudioIoHandle,  // [in] handle returned by .._OpenAudioOutput() / _OpenAudioInput()
                          int *piSourceIndex,  // [out,optional] index into T_SharedMemoryData.source[]
                          int *piReaderIndex,  // [out,optional] index into T_AudioBuffer.reader[]
        T_AudioSourceInfo **ppAudioSourceInfo) // [out,optional] address of the audio-source info
  // iAudioIoHandle identifies both, the audio WRITER (source) and the audio READER.
  // It doesn't have anything to do with a windows-handle, or a file-handle !
  // Here (inside aio2winamp),  the "audio source index" is in the upper bits,
  //    and the index into T_AudioBuffer. T_AudioReader reader[]
  //    is in the lower 8 bits (even though we don't need such a lot) .
  // GetAudioBufferByHandle() may return a NULL pointer if the handle
  //    is obviously invalid, or if the buffer doesn't exist anymore !
{
  unsigned int iAudioSource = (unsigned int)iAudioIoHandle >> 8;
  unsigned int iAudioReader = (unsigned int)iAudioIoHandle & 0x0FF;
  if( piSourceIndex != NULL )
   { *piSourceIndex = 0;
   }
  if( piReaderIndex != NULL )
   { *piReaderIndex = 0;    // always make sure this 'output' is valid
   }
  if( ppAudioSourceInfo!=NULL )
   { *ppAudioSourceInfo = NULL;
   }
  if( iAudioSource<AUDIO_MAX_SOURCES && iAudioReader<AUDIO_MAX_READERS )
   {
     if( ppAudioSourceInfo!=NULL && g_pSharedMemoryData!=NULL)
      { *ppAudioSourceInfo = &g_pSharedMemoryData->source[iAudioSource];
      }
     if( g_pAudioBuffers[iAudioSource] != NULL )
      {
        if( piSourceIndex != NULL )
         { *piSourceIndex = iAudioSource;
         }
        if( piReaderIndex != NULL )
         { *piReaderIndex = iAudioReader;
         }
        return g_pAudioBuffers[iAudioSource];
      }
   }
  return NULL;
} // end GetAudioBufferByHandle()

//---------------------------------------------------------------------------
static INT32 GetNumberOfFreeSamplePointsInOutputBuffer(
                     int iAudioSource,    // [in]
                     int *piNumActiveReaders ) // [out,optional]
  // Returns the number of 'unused' entries in T_AudioBuffer.fltBuffer[]
  // for the SLOWEST (but active) reader of this buffer.
  // If no reader is active, the return value indicates an 'empty' buffer.
{
  int iReaderIndex, nActiveReaders = 0;
  T_AudioBuffer *pAudioBuffer;
  T_AudioSourceInfo *pSource;
  T_AudioReader *pReader;
  INT32 i32HeadIndex, nUsedEntries, nFreeSamplePoints = 0;
  INT32 nMinimumOfFreeSamplePoints = (AUDIO_BUFFER_SIZE-1);


  if(  iAudioSource>=0 && iAudioSource<AUDIO_MAX_SOURCES
   &&  (g_pSharedMemoryData!=NULL)
   &&  ( (pAudioBuffer=g_pAudioBuffers[iAudioSource])!=NULL) )
   { pSource = &g_pSharedMemoryData->source[iAudioSource];

     //   What to return if there is no active reader ?
     //   Indicate an 'empty' buffer (to avoid blocking the writer),
     //   but (since 2013-11) let the buffer indicator bargraph show 100% .
     nFreeSamplePoints = AUDIO_BUFFER_SIZE-1;
     if( pSource->nChannelsPerSamplePoint > 1 )
      { nFreeSamplePoints /= pSource->nChannelsPerSamplePoint;
      }
     i32HeadIndex = (INT32)pAudioBuffer->dwHeadIndex;
     for( iReaderIndex=0; iReaderIndex<AUDIO_MAX_READERS; ++iReaderIndex )
      { pReader = &pAudioBuffer->reader[iReaderIndex];
        if( pReader->dwFlags & AUDIO_READER_OPENED )  // ok, this reader is 'active':
         { ++nActiveReaders;
           nUsedEntries = i32HeadIndex - (INT32)pAudioBuffer->reader[iReaderIndex].dwTailIndex;
           if( nUsedEntries < 0 )  // index wrapped around  (circular buffer wrap)
            {  nUsedEntries += AUDIO_BUFFER_SIZE;
            }
           nFreeSamplePoints = (AUDIO_BUFFER_SIZE-1) - nUsedEntries;
           if( pSource->nChannelsPerSamplePoint > 1 )
            { nFreeSamplePoints /= pSource->nChannelsPerSamplePoint;
            }
           if( nFreeSamplePoints < nMinimumOfFreeSamplePoints )
            { nMinimumOfFreeSamplePoints = nFreeSamplePoints;
            }
         }
      }
   }
  if( piNumActiveReaders != NULL )
   { *piNumActiveReaders = nActiveReaders;
   }
  return nFreeSamplePoints;
} // end GetNumberOfFreeSamplePointsInOutputBuffer()


//---------------------------------------------------------------------------


#if( L_DEBUG_AUDIO_THREAD )
void SanityCheck( int iSourceLineNr )
  // performs a simple, internal "sanity check" .
  // Errors can be displayed in the "About"-box .
{
  if( WA_i16SampleBuffer[C_WA_SAMPLE_BUFFER_SIZE] != WA_BUF_MAGIC_VALUE )
   { // OUCH ! Some nasty bug has bitten into the end of the Winamp-audio-chunk .
     // Who dunnit ?  Set a breakpoint here if you can, or watch the display
     // in this plugin's "About" box. At least, you will see the sourcecode
     // line number of the place where we *detected* the error
     // (but we still don't know the reason, of course...) .
     if( SanityCheckFailedInLine==0 ) // it's the first time we get here..
      {  SanityCheckFailedInLine = iSourceLineNr;
      }
     // 2008-04-04: Got here after calling the ODDCAST DSP plugin,
     //   but only when the plugin was compiled with DevCpp ? !
   }

} // end SanityCheck()
#endif // L_DEBUG_AUDIO_THREAD


char config_waveoutdir[MAX_PATH]="";
HANDLE hwavfile=NULL;
int WA_iSampleRate, WA_iNumChannels; // sample params for the WINAMP-side (!)
// ex: int WA_iBitsPerSample;  // since 2011, we use 16 bits/sample for the WINAMP-side
// Now, for simplicity, we use a fixed sample formats for the Winamp side as for the "Audio-I/O"-side !
volatile int writtentime, w_offset;

//---------------------------------------------------------------------------
// Worker threads for the audio player's INPUT
//   (for example, sends data from the shared buffer to winamp)
//---------------------------------------------------------------------------

// Just for reference, here is the original "test tone generator" .
//   To start this thread, select (in WINAMP's poor man's menu / "sinewave" icon):
//     "Play"..."URL"... and enter something like tone://1000
//     (to produce a 1000 Hz sine wave)
DWORD WINAPI __stdcall WA_ToneInputThread(void *b)
{
  int iNumBytesWeCanSendToWinamp, iNumSamplesWeCanSendToWinamp, iWinampBytesPerSample;
  double angle=0.0, d_angle=0.0;
  int i, iMaxSamplesPerChunk, iNumSamplesGotFromDSP;
  T_AudioSourceInfo *pAudioSourceInfo;
  T_AudioBuffer *pAudioBuffer;

  if( ! GetAudioBufferAndSourceInfoForWinamp(
               &pAudioBuffer,      // [out] audio buffer to be used by Winamp
               &pAudioSourceInfo)) // [out] source-info to be used by Winamp
   { return -1;
   }


  iWinampBytesPerSample = WA_iNumChannels * 2/*bytes per 16-bit integer*/;
  if( iWinampBytesPerSample < 1 )
      iWinampBytesPerSample = 1;  // oops.. avoid div by zero !
  iMaxSamplesPerChunk = C_WA_SAMPLE_BUFFER_SIZE / iWinampBytesPerSample;
  // Reserve half the buffer for the DSP (that's why we divide by 2 above)
  if( iMaxSamplesPerChunk > C_WINAMP_MAX_SAMPLES_PER_CHUNK/*ex:576*/ )
      iMaxSamplesPerChunk = C_WINAMP_MAX_SAMPLES_PER_CHUNK;  // > "Stick with 576 sample blocks if you can" (not good for Lame ?!)

  while (! *((BOOL*)b) )
   {
     iNumBytesWeCanSendToWinamp =  // 2008-03-28: Got the amazing number of 176400 bytes here !
        WA_sInModule.outMod->CanWrite(); // > returns number of bytes(!!) possible to write at a given time.
     iNumSamplesWeCanSendToWinamp = iNumBytesWeCanSendToWinamp / iWinampBytesPerSample;
     if( iNumSamplesWeCanSendToWinamp > iMaxSamplesPerChunk )
      {  iNumSamplesWeCanSendToWinamp = iMaxSamplesPerChunk;
      }
     // Caution : CanWrite returns number of BYTES(!) possible to write at a given time,
     //           not the number of SAMPLES(!) .
     // Only start processing if a reasonably large number of samples CAN be processed..
     //  (but, since some of the crazy winamp functions don't have a parameter
     //   telling it THE NUMBER OF SAMPLES TO BE PROCESSED,
     //   we end up using the same "number of samples" all the time (576) ?!
     if ( iNumSamplesWeCanSendToWinamp >= iMaxSamplesPerChunk )
      {
        int nSamplesLeft = iNumSamplesWeCanSendToWinamp;
        short *s = WA_i16SampleBuffer;

#if( L_DEBUG_AUDIO_THREAD )
        SanityCheck(__LINE__); // perform simple, internal "sanity check"
        // 2008-04-04: Sanity check failed HERE, when Oddcast/Lame plugin were active ?!
#endif // L_DEBUG_AUDIO_THREAD

        // Since 2011-07, the "test tone frequency" can be modified on-the-fly
        //  through the plugin's own control panel (input field labelled "ft/Hz")
        d_angle=3.14159*2.0*(double)pAudioSourceInfo->fltTestGeneratorFrequency/(double)44100.0;

        while (nSamplesLeft--)
         {
           double d = sin(angle)*4096.0; // ex: 32766.5 (but that hurt WB's ears, ouch..)
           for(i=0; i<WA_iNumChannels; ++i)
            { *s++  = (short)d;
            }
           angle += d_angle;
           if( angle>(3.14159*2.0) )
            {  angle -= (3.14159*2.0);
            }
         }
        iNumBytesWeCanSendToWinamp = iNumSamplesWeCanSendToWinamp * iWinampBytesPerSample;

#if( L_DEBUG_AUDIO_THREAD )
        SanityCheck(__LINE__); // perform simple, internal "sanity check"
#endif // L_DEBUG_AUDIO_THREAD


        // Time to call this other stuff in Winamp... but WHEN ? Was there an IF in this line once ?
         {
          int t = WA_sInModule.outMod->GetWrittenTime();
          WA_sInModule.SAAddPCMData((char *)WA_i16SampleBuffer,
                 WA_iNumChannels/*"nch"*/,  16/*ex: WA_iBitsPerSample, "bps"*/,  t/*imestamp*/);
#if( L_DEBUG_AUDIO_THREAD )
          SanityCheck(__LINE__); // perform simple, internal "sanity check"
#endif // L_DEBUG_AUDIO_THREAD
          WA_sInModule.VSAAddPCMData((char *)WA_i16SampleBuffer,
                 WA_iNumChannels/*"nch"*/,  16/*ex: WA_iBitsPerSample, "bps"*/,  t/*imestamp*/);
#if( L_DEBUG_AUDIO_THREAD )
          SanityCheck(__LINE__); // perform simple, internal "sanity check"
#endif // L_DEBUG_AUDIO_THREAD
         }
        if( WA_sInModule.dsp_isactive() )
         { iNumSamplesGotFromDSP = WA_sInModule.dsp_dosamples(WA_i16SampleBuffer,
             iNumSamplesWeCanSendToWinamp,  // NUMBER OF SAMPLES (not NUMBER OF BYTES !! )
                        16 ,                // bits per sample (what is "bps" - beats per second ?? grr )
                        WA_iNumChannels,    // "nch" must be the number of channels (please, use "SPEAKING NAMES")
                        WA_iSampleRate);    // "srate" = samples per second
           // >    dsp_dosamples() returns number of samples to output.
           // >    This can be as much as twice numsamples.
           // >    Be sure to allocate enough buffer for samples, then.
           if( iNumSamplesGotFromDSP > 0 )
            { iNumBytesWeCanSendToWinamp = iNumSamplesGotFromDSP * iWinampBytesPerSample;
            }
           else
            { iNumBytesWeCanSendToWinamp = 0;
            }
#if( L_DEBUG_AUDIO_THREAD )
           SanityCheck(__LINE__); // perform simple, internal "sanity check"
#endif // L_DEBUG_AUDIO_THREAD
         }
        if( WA_sInModule.UsesOutputPlug )
         { if (pAudioSourceInfo->iAudioPlayerDisableOutput && (iNumBytesWeCanSendToWinamp>0) )
            { // it seemed to be a bad idea to SKIP writing to the output.
              // So, instead, simply "null out" the block before sending it:
              memset( WA_i16SampleBuffer, 0, iNumBytesWeCanSendToWinamp );
            }
#if( L_DEBUG_AUDIO_THREAD )
           SanityCheck(__LINE__); // perform simple, internal "sanity check"
#endif // L_DEBUG_AUDIO_THREAD
           WA_sInModule.outMod->Write((char*)WA_i16SampleBuffer, iNumBytesWeCanSendToWinamp );
           // Much in contrast to dsp_dosamples(), this "outMod->Write"-thing
           // doesn't expect the NUMBER OF SAMPLES, but the NUMBER OF BYTES !
#if( L_DEBUG_AUDIO_THREAD )
           SanityCheck(__LINE__); // perform simple, internal "sanity check"
#endif // L_DEBUG_AUDIO_THREAD
         }
      } // end if < winamp says we "can write" enough samples to it >
     else
      { Sleep(10);
      }
   } // end < thread loop >
  return 0;
} // end WA_ToneInputThread()


//---------------------------------------------------------------------------
DWORD WINAPI __stdcall WA_ExternalInputThread(void *b)
  // Another worker thread, this one is for the "audi://" - mode
  //    ( "audi !" ,  latin for "listen !" .  In this mode,
  //      Winamp shall receive audio samples from an external application
  //      through a circular buffer in shared memory)
  //   To start this thread, select (in WINAMP's poor man's menu / "sinewave" icon):
  //     "Play"..."URL"... and enter something like audi://
  //   Then launch the application which can produce the audio stream.
  // To debug this thread with Borland C++Builder WHILE WINAMP IS RUNNING,
  //    read the author's notes in README_DEBUGGING_WINAMP.TXT .
{
  int iNumBytesWeCanSendToWinamp, nSamplePointsForWinamp, iWinampBytesPerSample;
  int iMaxSamplesPerChunk, iNumSamplesGotFromDSP;
  int iChannelOfSample;
  BOOL fMustWait;
  int  iSharedBufNumChannels; // ex: int iSharedBufBytesPerSample;
  long i32SharedBufHeadIndex;
  long i32SharedBufTailIndex;
  // ex: long i32SharedBufUsedBytes;
  long i32SharedBufUsedSamples;
  int   iSamplePoint;
  float f, *pfltSrc;
  short *pi16Dst;
  T_AudioBuffer     *pAudioBuffer;
  T_AudioSourceInfo *pAudioSourceInfo;


  iWinampBytesPerSample = WA_iNumChannels * 2/*bytes_per_int16*/;
  if( iWinampBytesPerSample < 1 )
      iWinampBytesPerSample = 1;  // oops.. avoid div by zero !
  iMaxSamplesPerChunk = C_WA_SAMPLE_BUFFER_SIZE / iWinampBytesPerSample;
  // Reserve half the buffer for the DSP (that's why we divide by 2 above)
  if( iMaxSamplesPerChunk > C_WINAMP_MAX_SAMPLES_PER_CHUNK/*ex:576*/ )
      iMaxSamplesPerChunk = C_WINAMP_MAX_SAMPLES_PER_CHUNK;  // > "Stick with 576 sample blocks if you can" (not good for Lame ?!)

  while (! *((BOOL*)b) )
   {
     // If someone/something killed the shared audio buffer, exit this thread
     if( ! GetAudioBufferAndSourceInfoForWinamp( &pAudioBuffer, &pAudioSourceInfo ) ) // may fail !
      { break;
      }

     fMustWait = TRUE;
     iNumBytesWeCanSendToWinamp =  // 2008-03-28: Got the amazing number of 176400 bytes here !
        WA_sInModule.outMod->CanWrite(); // > returns number of bytes(!) possible to write at a given time.
     nSamplePointsForWinamp = iNumBytesWeCanSendToWinamp / iWinampBytesPerSample;
     // Winamp tells us its output-module can accept another chunk of data.
     // Note: WA_i16SampleBuffer is two times larger than it needs to be
     //       in a "normal" application (but Winamp isn't "normal"),
     //       so at this point, we don't have to care about WA_sInModule.dsp_isactive() !

     if( nSamplePointsForWinamp > iMaxSamplesPerChunk )
      {  nSamplePointsForWinamp = iMaxSamplesPerChunk;
      }
     // Caution : CanWrite returns number of BYTES(!) possible to write at a given time,
     //           not the number of SAMPLES(!) .
     // Only start processing if a reasonably large number of samples CAN be processed..
     //  (but, since some of the crazy winamp functions don't have a parameter
     //   telling it THE NUMBER OF SAMPLES TO BE PROCESSED,
     //   we end up using the same "number of samples" all the time (576) ?!
     if ( nSamplePointsForWinamp >= iMaxSamplesPerChunk )
      {
#if( L_DEBUG_AUDIO_THREAD )
        SanityCheck(__LINE__); // perform simple, internal "sanity check"
#endif // L_DEBUG_AUDIO_THREAD
        // Convert the NUMBER OF SAMPLES back into the NUMBER OF BYTES,
        // because the " " "  may have been limited now :
        iNumBytesWeCanSendToWinamp = nSamplePointsForWinamp * iWinampBytesPerSample;
          // ( iNumBytesWeCanSendToWinamp is typically 2 or 4 times 576,
          //   see the strange notes in the winamp plugin SDK ! ! )

        // But do we have enough samples in the SHARED AUDIO BUFFER ?
        i32SharedBufHeadIndex = (long)(pAudioBuffer->dwHeadIndex & (AUDIO_BUFFER_SIZE-1) );
        i32SharedBufTailIndex = (long)(pAudioBuffer->reader[0].dwTailIndex & (AUDIO_BUFFER_SIZE-1) );
        iSharedBufNumChannels = pAudioSourceInfo->nChannelsPerSamplePoint;
        if( iSharedBufNumChannels >= 1 )
         {
           i32SharedBufUsedSamples = i32SharedBufHeadIndex - i32SharedBufTailIndex;
           if( i32SharedBufUsedSamples < 0 )  // index wrapped around..
            {  i32SharedBufUsedSamples += AUDIO_BUFFER_SIZE;
            }
           i32SharedBufUsedSamples /= iSharedBufNumChannels;
           if( i32SharedBufUsedSamples >= nSamplePointsForWinamp )
            { // Arrived here, we know we can send enough SAMPLE POINTS to Winamp.
              // Even if we COULD send more samples (because the shared buffer
              // is full enough), we don't, due to the notes in the Winamp plugin SDK.
              fMustWait = FALSE;  // no need to "Sleep" further below

              // Source (shared buffer) : 32 bit float, normalized to +/- 1.0
              // Destination (winamp)   : 16 bit signed integer .
              pfltSrc = pAudioBuffer->fltBuffer + i32SharedBufTailIndex;
              pi16Dst = WA_i16SampleBuffer;
              for( iSamplePoint=0; iSamplePoint<nSamplePointsForWinamp; ++iSamplePoint )
               {
                 for( iChannelOfSample=0; iChannelOfSample<WA_iNumChannels; ++iChannelOfSample )
                  {
                    if( iChannelOfSample<iSharedBufNumChannels )
                     { f = pfltSrc[iChannelOfSample];
                       if( f < -1.0 )  f = -1.0;
                       if( f >  1.0 )  f =  1.0;
                       *pi16Dst++ = (short)(f * 32767.0);
                     }
                    else // less number of channels in the shared buffer (source) :
                     { *pi16Dst++ = 0;  // fill the unused channels with 'silence'
                     }
                  } // end for (iChannelOfSample ..
                 pfltSrc += iSharedBufNumChannels;
                 i32SharedBufTailIndex += iSharedBufNumChannels;
                 if( i32SharedBufTailIndex >= AUDIO_BUFFER_SIZE ) // circular wrap
                  {  i32SharedBufTailIndex = 0;
                     pfltSrc = pAudioBuffer->fltBuffer;
                  }
               } // end for( iSamplePoint ...

#if( L_DEBUG_AUDIO_THREAD )
              SanityCheck(__LINE__); // perform simple, internal "sanity check"
#endif // L_DEBUG_AUDIO_THREAD

              // Write the "new" audio buffer tail-index back to the structure in shared memory:
              pAudioBuffer->reader[0].dwTailIndex = i32SharedBufTailIndex & (AUDIO_BUFFER_SIZE-1);

              // Time to call this other stuff in Winamp... but WHEN ? Was there an IF in this line once ?
               {
                 int t = WA_sInModule.outMod->GetWrittenTime();
                 WA_sInModule.SAAddPCMData((char *)WA_i16SampleBuffer,
                          WA_iNumChannels/*"nch"*/,  16/* ex: WA_iBitsPerSample "bps"*/,  t/*imestamp*/);
                 // Note: It seems impossible to send the NUMBER OF SAMPLES
                 //       or the NUMBER OF BYTES to process to SAAddPCMData() !
                 // However, it's not clear if THIS is the reason for the strange
                 //  Winamp crash on Mark Dennison's PC (trying to send MP3 via Oddcast).
#if( L_DEBUG_AUDIO_THREAD )
                 SanityCheck(__LINE__); // perform simple, internal "sanity check"
#endif
                 WA_sInModule.VSAAddPCMData((char *)WA_i16SampleBuffer,
                          WA_iNumChannels/*"nch"*/,  16/*BitsPerSample*/,  t/*imestamp*/);
#if( L_DEBUG_AUDIO_THREAD )
                 SanityCheck(__LINE__); // perform simple, internal "sanity check"
#endif
               }
              if( WA_sInModule.dsp_isactive() )
               { // > dsp_isactive() returns 1 if active (which means that the number
                 // > of samples returned by dsp_dosamples could be greater than
                 // > went in.. Use it to estimate if you'll have enough room in the
                 // > output buffer .
                 iNumSamplesGotFromDSP = WA_sInModule.dsp_dosamples(WA_i16SampleBuffer,
                        nSamplePointsForWinamp,  // NUMBER OF SAMPLES (not NUMBER OF BYTES !! )
                        16,                 // bits per sample (what is "bps" - beats per second ?? grr )
                        WA_iNumChannels,    // "nch" must be the number of channels (please, use "SPEAKING NAMES")
                        WA_iSampleRate);    // "srate" = samples per second
                 // >    dsp_dosamples() returns number of samples to output.
                 // >    This can be as much as twice numsamples.
                 // >    Be sure to allocate enough buffer for samples, then.
                 iNumBytesWeCanSendToWinamp = iNumSamplesGotFromDSP * iWinampBytesPerSample;
#if( L_DEBUG_AUDIO_THREAD )
                 SanityCheck(__LINE__); // perform simple, internal "sanity check"
#endif
               }
              if( WA_sInModule.UsesOutputPlug)
               { if (pAudioSourceInfo->iAudioPlayerDisableOutput && (iNumBytesWeCanSendToWinamp>0) )
                  { // it seemed to be a bad idea to SKIP writing to the output.
                    // So, instead, simply "null out" the block before sending it:
                    memset( WA_i16SampleBuffer, 0, iNumBytesWeCanSendToWinamp );
                  }
                 WA_sInModule.outMod->Write((char*)WA_i16SampleBuffer,(int)iNumBytesWeCanSendToWinamp );
#if( L_DEBUG_AUDIO_THREAD )
                 SanityCheck(__LINE__); // perform simple, internal "sanity check"
#endif
               }
            } // end if( i32SharedBufUsedSamples >= nSamplePointsForWinamp )
         } // end if( iSharedBufBytesPerSample >= 1 )
      } // end if < winamp says we "can write" enough samples to it >

     // If we didn't "wait somewhere", we should wait here.
     //  Otherwise this thread would turn into a nasty CPU hog !
     if( fMustWait )
      { Sleep(10);
      }
   } // end < thread loop >
  return 0;
} // end WA_ExternalInputThread()


//---------------------------------------------------------------------------
// Functions called by winamp through the vector table in the In_Module-structure :
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
void WA_Config(HWND hwndParent)
  // Called from WINAMP through In_Module::Config() to show the config dialog .
  // WINAMP-SPECIFIC !  The "Audio-I/O-side" is in AIO_ShowControlPanel() .
{
#if ( 0 )
  if( ! fWaitingInMessageBox )
   {
     fWaitingInMessageBox = TRUE;
     MessageBox(hwndParent,
        "To receive from AUDIO_I/O, open 'audi://  .\n"\
        "To produce a test tone, open 'tone://2600' (for 2600hz).",
        "Configuration",MB_OK);
     fWaitingInMessageBox = FALSE;
   }
#else
  // This only works ONCE, but never a second time, when trying to open
  //      the following configuration window THROUGH WINAMP'S PLUGIN SCREEN:
  AIO_ShowControlPanel( // NON-MODAL (not blocking) !
        hwndParent,  // [in] handle of the caller's main window
        NULL,        // [out,optional] handle of the DLL's control panel
        AIO_DISPLAY_SHOW_WINDOW, // [in] iDisplayOptions
        NULL,        // [in,optional] pCallbackUserDataForControlPanel
        NULL );      // [in,optional] address of a callback function, may be NULL
#endif
} // end WA_Config()


//---------------------------------------------------------------------------
void WA_About(HWND hwndParent)
  // "about"-dialog, called by winamp through In_Module::About() .
  // For simplicity, IN THIS CASE, also callable through AIO_ShowAboutBox() .
{
  char szText[2048];
  char *cp = szText;

  if( ! fWaitingInMessageBox )
   { fWaitingInMessageBox = TRUE;
     strcpy( cp,
       "Audio-I/O-DLL with WINAMP-input-plugin for SpecLab " WINAMP_PLUGIN_VERSION " .\n"
       "Written by Wolfgang Buescher (DL4YHF).\n\nCompiled: " __DATE__ );
     cp += strlen(cp);
     sprintf(cp, "\nLocation: %s", APPL_szDllNameAndPath );
     cp += strlen(cp);
     sprintf(cp, "\nWebsite:  www.qsl.net/dl4yhf/AudioIO/index.html" );

#  if( L_DEBUG_AUDIO_THREAD )
     cp += strlen(cp);
     if( SanityCheckFailedInLine != 0 )
      { wsprintf( cp, "\n\nSanity check failed in line %d , please terminate !",
                          (int)SanityCheckFailedInLine);
      }
     else
      { strcpy( cp,   "\n\nSanity check : ok." );
      }
     cp += strlen(cp);
#  endif // L_DEBUG_AUDIO_THREAD

     MessageBox(hwndParent, szText, "About this Audio-I/O DLL",MB_OK);
     fWaitingInMessageBox = FALSE;
   }
} // end WA_About()


//---------------------------------------------------------------------------
void WA_Init()
  // Called from WINAMP(!) through In_Module::Init() at 'program init' .
{
  T_AudioSourceInfo *pAudioSourceInfo;
  T_AudioBuffer     *pAudioBuffer;

  if( GetAudioBufferAndSourceInfoForWinamp(
               &pAudioBuffer,      // [out] audio buffer to be used by Winamp
               &pAudioSourceInfo)) // [out] source-info to be used by Winamp
   {
     // For simplicity, let Winamp use the same sample rate, sample format,
     //  and "number of channels" as for the shared audio buffer :
     WA_iSampleRate   = (int)pAudioSourceInfo->dblSampleRate;   // samples per second; possibly "calibrated"
     WA_iNumChannels  = pAudioSourceInfo->nChannelsPerSamplePoint;         // 0=not set yet, 1=mono, 2=stereo
     // ex: WA_iBitsPerSample = g_pSharedAudioBuffer->iBitsPerSample;  // 16 bits per sample, ALWAYS
     pAudioSourceInfo->iAudioPlayerInitialized = 1;  // "now winamp is there, too"
   }
}

//---------------------------------------------------------------------------
void WA_Quit()
  // Called from WINAMP through In_Module::Quit() at 'program quit' .
{
  T_AudioSourceInfo *pAudioSourceInfo;
  T_AudioBuffer     *pAudioBuffer;

  if( GetAudioBufferAndSourceInfoForWinamp(
               &pAudioBuffer,      // [out] audio buffer to be used by Winamp
               &pAudioSourceInfo)) // [out] source-info to be used by Winamp
   {
     pAudioSourceInfo->iAudioPlayerInitialized = 0;  // "winamp isn't with us any longer"
   }
  if( g_pSharedMemoryData != NULL )
   { strcpy( g_pSharedMemoryData->sz255WinampLastEvent, "Winamp called 'Quit'" );
     g_pSharedMemoryData->iWinampCallFlags |= WA_CALLED_QUIT;
     g_pSharedMemoryData->iWinampCallFlags &= ~( // allow showing the following 'winamp calls' again:
       WA_CALLED_GET_IN_MOD | WA_CALLED_IS_OUR_FILE | WA_CALLED_GET_FILE_INFO
     | WA_CALLED_PLAY | WA_CALLED_STOP | WA_CALLED_PLAY_ERR | WA_CALLED_PAUSE );
     AppendToEventHistory( g_pSharedMemoryData->sz255WinampLastEvent );
     ++g_pSharedMemoryData->iGuiUpdateCounter; // let all control panels update their GUIs
   }
}

//---------------------------------------------------------------------------
int WA_IsOurFile(char *fn)
  // Called from WINAMP through In_Module::IsOurFile() .
{
  char buf[8];
  memcpy(buf,fn,7);
  buf[7]=0;

  if( g_pSharedMemoryData != NULL )
   { sprintf( g_pSharedMemoryData->sz255WinampLastEvent,
         "Winamp: IsOurFile(%s)", (char*)((fn!=NULL)?fn:"NULL") );
     g_pSharedMemoryData->iWinampCallFlags |= WA_CALLED_IS_OUR_FILE;
     ++g_pSharedMemoryData->iGuiUpdateCounter; // let all control panels update their GUIs
     // Note: It's interesting to see HOW MANY QUESTIONS winamp will ask !
   }

  if( lstrcmp(buf,"tone://") == 0 )
   { // called in TEST MODE (produce a single sine wave) ..
     return 1;
   }
  else // not "tone://" but what else ?
  if( lstrcmp(buf,"audi://") == 0 )  // "listen for audio from an external application"
   { return 1;
   }
  return 0;
} // end WA_IsOurFile()

//---------------------------------------------------------------------------
int WA_Play(
      char *fn) // [in] "filename" (or, in this case, pseudo-URL
                //                  which should contain the stream-ID,
                //                  for example fn = audi://VLF )
  // Called from WINAMP through In_Module::Play() .
  // Notes:
  // -  To debug this thread with Borland C++Builder WHILE WINAMP IS RUNNING,
  //     read the author's notes in README_DEBUGGING_WINAMP.TXT .
{
  long i32NominalSampleRate;
  long i32BitRate;
  char buf[8], *cpStreamID;
  int n, maxlatency;
  int iNumChannelsForWinamp;
  double dblSampleRate;
  DWORD dwThreadID;
  T_AudioSourceInfo *pAudioSourceInfo;
  T_AudioBuffer     *pAudioBuffer;
  int  iAudioSource;
  T_ChunkInfo chunk_info;

  if( g_pSharedMemoryData == NULL )  // something is VERY wrong here...
   { return 0;   // this is Winamp's way to return an ERROR !
   }

  memcpy(buf,fn,7);
  buf[7]=0;           // -> buf:  "audi://" or "tone://" [first 7 chars of pseudo-URL)
  n = strlen(fn);
  cpStreamID = NULL;
  if( n>7 )
   { cpStreamID = strrchr(fn,'/');
     if(cpStreamID!=NULL) ++cpStreamID;
   }

  if( g_iWinampAudioHandle>=0 )
   { AIO_CloseAudioInput( g_iWinampAudioHandle );
   }

  g_iWinampAudioHandle = AIO_OpenAudioInput( // here: open an audio source for WINAMP...
        cpStreamID,   // [in] 'Stream ID' for the Audio-I/O-DLL, identifies the AUDIO SOURCE
                      // (optional, use NULL to use "the first available source")
        "Winamp",     // [in] short but hopefully UNIQUE name for the 'reader'
        -1.0,         // [in(!)] "wanted" sample rate, here: let the source decide
        &dblSampleRate,    // [out(!)] "realized" sample rate, samples per second,
        2,            // [in] number of channels PER SAMPLE POINT
        200,          // [in] iTimeout_ms, max timeout in milliseconds for THIS call
        AIO_OPEN_OPTION_NORMAL, // [in] bit combination of AIO_OPEN_OPTION_... flags
        &chunk_info); // [out,optional] chunk info, see c:\cbproj\SoundUtl\ChunkInfo.h
  if( g_iWinampAudioHandle < 0 )
   { if( (g_pSharedMemoryData->iWinampCallFlags & WA_CALLED_PLAY_ERR)==0 )
         { g_pSharedMemoryData->iWinampCallFlags |=  WA_CALLED_PLAY_ERR;
           sprintf( g_pSharedMemoryData->sz255WinampLastEvent, "WA_Play(%s): couldn't open input",fn);
           AppendToEventHistory( g_pSharedMemoryData->sz255WinampLastEvent );
           ++g_pSharedMemoryData->iGuiUpdateCounter; // let all control panels update their GUIs
         }
     return 0;
   }
  pAudioBuffer = GetAudioBufferByHandle( // may return NULL if something's wrong !
             g_iWinampAudioHandle, // [in] handle returned by .._OpenAudioOutput() / _OpenAudioInput()
                    &iAudioSource, // [out,optional] index into T_SharedMemoryData.source[]
                             NULL, // [out,optional] index into T_AudioBuffer.reader[]
               &pAudioSourceInfo); // [out,optional] address of audio source info
  if( pAudioBuffer==NULL )
   { if( (g_pSharedMemoryData->iWinampCallFlags & WA_CALLED_PLAY_ERR)==0 )
         { g_pSharedMemoryData->iWinampCallFlags |=  WA_CALLED_PLAY_ERR;
           sprintf( g_pSharedMemoryData->sz255WinampLastEvent, "WA_Play(%s): internal error",fn);
           AppendToEventHistory( g_pSharedMemoryData->sz255WinampLastEvent );
           ++g_pSharedMemoryData->iGuiUpdateCounter; // let all control panels update their GUIs
         }
        return 0;
   }

#if( L_DEBUG_AUDIO_THREAD )
  SanityCheck(__LINE__); // perform simple, internal "sanity check"
#endif // L_DEBUG_AUDIO_THREAD

  strncpy(WA_sz255PlayingFileName,fn,255);

  pAudioSourceInfo->iAudioPlayerInputPaused = 0;  // here: set by WINAMP when "starting to play"
  memset(WA_i16SampleBuffer,0,C_WA_SAMPLE_BUFFER_SIZE * sizeof(WA_i16SampleBuffer[0]));

  // We won't confuse Winamp with our "calibrated" sampling rate, because
  // it will most likely be irritated if we tell it the true sampling rate
  // is 44123.567 samples per second instead of 44100 .
  i32NominalSampleRate = PreciseToNominalSampleRate( pAudioSourceInfo->dblSampleRate );
  WA_iSampleRate   = i32NominalSampleRate; // samples per second; possibly "calibrated"
  iNumChannelsForWinamp = WA_iNumChannels = pAudioSourceInfo->nChannelsPerSamplePoint;
  // ex: WA_iBitsPerSample = pAudioSourceInfo->iBitsPerSample;  // 16 bits per sample in most cases
  // Note: Even though the "AudioIO"-side of this DLL may change the number of channels,
  //       we won't change WA_iNumChannels while playing the current track / stream !

  maxlatency = WA_sInModule.outMod->Open( i32NominalSampleRate, iNumChannelsForWinamp,
                         16/*BitsPerSample*/,
                         -1/*bufferlenms*/, -1/*prebufferms*/);
  // > " bufferlenms and prebufferms must be in ms. 0 to use defaults ".
  // ( why do they use -1 instead of zero in the sample plugin, then ? )
  if (maxlatency < 0)
   {
     return 1;
   }

  i32BitRate = i32NominalSampleRate * 16/*iBitsPerSample*/ * iNumChannelsForWinamp;
  // dividing by 1000 for the first parameter of setinfo makes it
  // display 'H'... for hundred.. i.e. 14H Kbps.
  WA_sInModule.SetInfo( i32BitRate/1000/*bitrate*/,
                        i32NominalSampleRate/1000/*samprate*/,
                        iNumChannelsForWinamp/*"stereo"*/, 1/*"synched"*/
                      );

  WA_sInModule.SAVSAInit(maxlatency,i32NominalSampleRate);
  WA_sInModule.VSASetInfo(i32NominalSampleRate,iNumChannelsForWinamp);
      // "call this in Play() to tell the vis plug-ins the current output params"

  WA_sInModule.outMod->SetVolume(-666); // set the output plug-ins default volume

  WA_fTerminateThead = FALSE;

  #define L_WORKER_TRHEAD_STACK_SIZE (256*1024)

  if( lstrcmp(buf,"tone://") == 0 )
   {  // called in TEST MODE (produce a single sine wave) ..
      pAudioSourceInfo->fltTestGeneratorFrequency = atoi( fn+7 );
      pAudioSourceInfo->iAudioPlayerMode = AIO_PLAYER_MODE_TONE;
      WA_hInputThread = (HANDLE) CreateThread(NULL,
                          L_WORKER_TRHEAD_STACK_SIZE,
                          (LPTHREAD_START_ROUTINE) WA_ToneInputThread,
                          (void *) &WA_fTerminateThead,
                          CREATE_SUSPENDED, // DWORD dwCreationFlags
                          &dwThreadID);
   }
  else // not "tone://" but what else ?
  if( lstrcmp(buf,"audi://") == 0 )
   {
      pAudioSourceInfo->iAudioPlayerMode = AIO_PLAYER_MODE_AUDI;
      WA_hInputThread = (HANDLE) CreateThread(NULL,
                          L_WORKER_TRHEAD_STACK_SIZE,
                          (LPTHREAD_START_ROUTINE)WA_ExternalInputThread,
                          (void *) &WA_fTerminateThead,
                          CREATE_SUSPENDED, // DWORD dwCreationFlags
                          &dwThreadID);

   }
  if( WA_hInputThread != NULL )
   {
     SetThreadPriority( WA_hInputThread, THREAD_PRIORITY_BELOW_NORMAL );
     ResumeThread( WA_hInputThread ); // ex: m_pFIFOThread->ResumeThread();
     if( g_pSharedMemoryData != NULL )
      { if( (g_pSharedMemoryData->iWinampCallFlags & WA_CALLED_PLAY )==0 )
         { g_pSharedMemoryData->iWinampCallFlags |=  WA_CALLED_PLAY;
           sprintf( g_pSharedMemoryData->sz255WinampLastEvent, "WA_Play(%s): running",fn);
           AppendToEventHistory( g_pSharedMemoryData->sz255WinampLastEvent );
           ++g_pSharedMemoryData->iGuiUpdateCounter; // let all control panels update their GUIs
         }
      }
   }

#if( L_DEBUG_AUDIO_THREAD )
  SanityCheck(__LINE__); // perform simple, internal "sanity check"
#endif // L_DEBUG_AUDIO_THREAD

  return 0;
}

//---------------------------------------------------------------------------
void WA_Pause()
  // Called from WINAMP through In_Module::Pause() .
{
  T_AudioSourceInfo *pAudioSourceInfo;
  T_AudioBuffer     *pAudioBuffer;

  if( GetAudioBufferAndSourceInfoForWinamp(
               &pAudioBuffer,      // [out] audio buffer to be used by Winamp
               &pAudioSourceInfo)) // [out] source-info to be used by Winamp
   {
     pAudioSourceInfo->iAudioPlayerInputPaused = 1;
     WA_sInModule.outMod->Pause(1);
     if( g_pSharedMemoryData != NULL )
      { if( (g_pSharedMemoryData->iWinampCallFlags & WA_CALLED_PAUSE)==0 )
         { g_pSharedMemoryData->iWinampCallFlags |=  WA_CALLED_PAUSE;
           strcpy( g_pSharedMemoryData->sz255WinampLastEvent, "Winamp: Paused !");
           AppendToEventHistory( g_pSharedMemoryData->sz255WinampLastEvent );
           ++g_pSharedMemoryData->iGuiUpdateCounter; // let all control panels update their GUIs
         }
      }
   }
}

//---------------------------------------------------------------------------
void WA_UnPause()
  // Called from WINAMP through In_Module::UnPause() .
{
  T_AudioSourceInfo *pAudioSourceInfo;
  T_AudioBuffer     *pAudioBuffer;

  if( GetAudioBufferAndSourceInfoForWinamp(
               &pAudioBuffer,      // [out] audio buffer to be used by Winamp
               &pAudioSourceInfo)) // [out] source-info to be used by Winamp
   { pAudioSourceInfo->iAudioPlayerInputPaused = 0;
     WA_sInModule.outMod->Pause(0);
     if( g_pSharedMemoryData != NULL )
      { if( (g_pSharedMemoryData->iWinampCallFlags & WA_CALLED_PAUSE)!=0 )
         { g_pSharedMemoryData->iWinampCallFlags &= ~WA_CALLED_PAUSE;
           strcpy( g_pSharedMemoryData->sz255WinampLastEvent, "Winamp: UnPaused !");
           AppendToEventHistory( g_pSharedMemoryData->sz255WinampLastEvent );
           ++g_pSharedMemoryData->iGuiUpdateCounter; // let all control panels update their GUIs
         }
      }
   }
}

//---------------------------------------------------------------------------
int WA_IsPaused()
  // Called from WINAMP through In_Module::IsPause() .
{
  T_AudioSourceInfo *pAudioSourceInfo;
  T_AudioBuffer     *pAudioBuffer;
  if( GetAudioBufferAndSourceInfoForWinamp(
               &pAudioBuffer,      // [out] audio buffer to be used by Winamp
               &pAudioSourceInfo)) // [out] source-info to be used by Winamp
   { return pAudioSourceInfo->iAudioPlayerInputPaused;
   }
  else
   { return 0;
   }
}

//---------------------------------------------------------------------------
void WA_Stop()
  // Called from WINAMP through In_Module::Stop() to "unload" the stream .
{
	if ( WA_hInputThread != INVALID_HANDLE_VALUE )
	{
		WA_fTerminateThead = TRUE;
		if (WaitForSingleObject( WA_hInputThread,INFINITE) == WAIT_TIMEOUT)
		{
			MessageBox(WA_sInModule.hMainWindow,"error asking thread to die!\n","error killing decode thread",0);
			TerminateThread(WA_hInputThread,0);
		}
		CloseHandle(WA_hInputThread);
		WA_hInputThread = INVALID_HANDLE_VALUE;
	}

	WA_sInModule.outMod->Close();
	WA_sInModule.SAVSADeInit();

  if( g_iWinampAudioHandle>=0 )
   { AIO_CloseAudioInput( g_iWinampAudioHandle );
     g_iWinampAudioHandle = -1;
   }


  if( g_pSharedMemoryData != NULL )
   { if( (g_pSharedMemoryData->iWinampCallFlags & WA_CALLED_STOP)==0 )
         { g_pSharedMemoryData->iWinampCallFlags |=  WA_CALLED_STOP;
           strcpy( g_pSharedMemoryData->sz255WinampLastEvent, "Winamp: Stopped !");
           AppendToEventHistory( g_pSharedMemoryData->sz255WinampLastEvent );
           ++g_pSharedMemoryData->iGuiUpdateCounter; // let all control panels update their GUIs
         }
   }
}

//---------------------------------------------------------------------------
int WA_GetLength()
  // Called from WINAMP through In_Module::GetLength()
  //   to "get the length" (of what, the CURRENT TRACK?...)   in milliseconds.
{ return -1000;
}

//---------------------------------------------------------------------------
int WA_GetOutputTime()
  // Called from WINAMP through In_Module::GetOutputTime()  .
  //  > Returns current output time in ms.
  //  > Usually returns outMod->GetOutputTime() .    So do we....
{ return WA_sInModule.outMod->GetOutputTime();
}

//---------------------------------------------------------------------------
void WA_SetOutputTime(int time_in_ms)
  // Called from WINAMP through In_Module::SetOutputTime()  .
  //  > Seeks to point in stream (in milliseconds) .
  // We don't support that in this plugin... how could we ?
{ (void)time_in_ms;
}

//---------------------------------------------------------------------------
void WA_SetVolume(int volume)
  // Called from WINAMP through In_Module::SetVolume()  .
  //  > from 0 to 255.. usually just call outMod->SetVolume .
{ WA_sInModule.outMod->SetVolume(volume);
}

//---------------------------------------------------------------------------
void WA_SetPan(int pan)
  // Called from WINAMP through In_Module::SetPan()  .
  //  > from -127 to 127.. usually just call outMod->SetPan .
  // (this "pan"-thingy is the "balance" between left and right channel,
  //  or what ? ? )
{ WA_sInModule.outMod->SetPan(pan);
}

//---------------------------------------------------------------------------
int WA_ShowInfoBox(char *fn, HWND hwnd)
  // Called from WINAMP through In_Module::ShowInfoBox()  .
  // No further details given .
{
   fn=fn; hwnd=hwnd;
	return 0;
}

//---------------------------------------------------------------------------
void WA_GetFileInfo(char *filename, char *title, int *length_in_ms)
  // Called from WINAMP through In_Module::GetFileInfo()  .
{
  T_AudioSourceInfo *pAudioSourceInfo;
  T_AudioBuffer     *pAudioBuffer;

  if( g_pSharedMemoryData != NULL )
   { if( (g_pSharedMemoryData->iWinampCallFlags & WA_CALLED_GET_FILE_INFO)==0 )
      { g_pSharedMemoryData->iWinampCallFlags |=  WA_CALLED_GET_FILE_INFO;
        sprintf( g_pSharedMemoryData->sz255WinampLastEvent, "Winamp: GetFileInfo(%s,%s).",
            (filename!=NULL)?filename:"NULL",
            (title!=NULL)   ? title  :"NULL" );
        AppendToEventHistory( g_pSharedMemoryData->sz255WinampLastEvent );
        ++g_pSharedMemoryData->iGuiUpdateCounter; // let all control panels update their GUIs
      }
   } // end if( g_pSharedMemoryData != NULL )

  if (!filename || !*filename)
   { // "if filename == NULL, current playing file is used"
     filename=WA_sz255PlayingFileName;
   }

  if (title)         // Winamp wants to know the "TITLE" we're playing..
   {
     if( GetAudioBufferAndSourceInfoForWinamp(
               &pAudioBuffer,      // [out] audio buffer to be used by Winamp
               &pAudioSourceInfo)) // [out] source-info to be used by Winamp
      {
        switch( pAudioSourceInfo->iAudioPlayerMode ) // what are we doing ?
         { case AIO_PLAYER_MODE_TONE:   // we're just sending a TEST TONE :
              // ex: wsprintf(title,"%.1f Hz Tone",(float)pAudioSourceInfo->fltTestGeneratorFrequency );
              // BEWARE ! wsprintf, unlike sprintf, does NOT seem to support floating point !!
              sprintf(title,"%.1f Hz Tone",(float)pAudioSourceInfo->fltTestGeneratorFrequency );
              break;
           case AIO_PLAYER_MODE_AUDI:   // we're listening to the input from the AudioIO-DLL :
              strcpy(title,"input from AudioIO");
              break;
           default :
              wsprintf(title,"Audio Player mode #%d ?",(int)pAudioSourceInfo->iAudioPlayerMode );
              break;
         }
      }
     else
      { strcpy(title,"AudioIO not ready to play");
      }
   }
  if (length_in_ms)   // Winamp wants to know the length of the played file in milliseconds..
   {
     *length_in_ms=-1000;  // .. we don't know, because it's a STREAM .
   }
} // end WA_GetFileInfo()

//---------------------------------------------------------------------------
void WA_EQSet(int on, char data[10], int preamp)
  // Called from WINAMP through In_Module::EQSet()  .
  // No further details given .
{
  on=on; data=data; preamp=preamp;
}

//---------------------------------------------------------------------------
// The only DLL function called by winamp.  NO UNDERSCORE, NO NAME MANGLING !
//---------------------------------------------------------------------------
__declspec( dllexport ) In_Module * winampGetInModule2()
{
  if( g_pSharedMemoryData!=NULL )
   { if( (g_pSharedMemoryData->iWinampCallFlags & WA_CALLED_GET_IN_MOD)==0 )
      { g_pSharedMemoryData->iWinampCallFlags |=  WA_CALLED_GET_IN_MOD;
        strcpy( g_pSharedMemoryData->sz255WinampLastEvent, "Winamp called 'GetInModule2'");
        AppendToEventHistory( g_pSharedMemoryData->sz255WinampLastEvent );
        ++g_pSharedMemoryData->iGuiUpdateCounter; // let all control panels update their GUIs
      }
   }
  return &WA_sInModule;
}

//---------------------------------------------------------------------------
In_Module WA_sInModule = {   // structure for winamp input plugin, see "in2.h"
	IN_VER,                   // module type (IN_VER)
	"Input from Audio-IO " WINAMP_PLUGIN_VERSION, // description of module, with version string
	0, // hmainwindow , winamp's main window (filled in by winamp)
	0, // hdllinstance, DLL instance handle (Also filled in by winamp)

	"\0", // char *FileExtensions; something like "mp3\0Layer 3 MPEG\0mp2\0Layer 2 MPEG\0mpg\0Layer 1 MPEG\0"
         // May be altered from Config, so the user can select what they want

	0,    // int is_seekable : 0=this stream is not seekable
	1,    // int UsesOutputPlug; does this plug-in use the output plug-ins? (musn't ever change, ever :)


	WA_Config,  // void (*Config)(HWND hwndParent); configuration dialog
	WA_About,   // void (*About)(HWND hwndParent);  about dialog
	WA_Init,    // void (*Init)();  called at program init
	WA_Quit,    // void (*Quit)();  called at program quit

   WA_GetFileInfo, // void (*GetFileInfo)(char *file, char *title, int *length_in_ms);
                   //       if file == NULL, current playing is used
	WA_ShowInfoBox, // int (*InfoBox)(char *file, HWND hwndParent);

	WA_IsOurFile,   // int (*IsOurFile)(char *fn); called before extension checks,
                   // to allow detection of mms://,  etc

	// playback stuff ...
	WA_Play,     // int (*Play)(char *fn); return zero on success, -1 on file-not-found,
                // some other value on other (stopping winamp) error
	WA_Pause,    // void (*Pause)();   pause stream
	WA_UnPause,  // void (*UnPause)(); unpause stream
	WA_IsPaused, // int (*IsPaused)(); ispaused? return 1 if paused, 0 if not
	WA_Stop,     // void (*Stop)();    stop (unload) stream

	// time stuff
	WA_GetLength, // int (*GetLength)();			// get length in ms
	WA_GetOutputTime, // int (*GetOutputTime)();		// returns current output time in ms. (usually returns outMod->GetOutputTime()
	WA_SetOutputTime, // void (*SetOutputTime)(int time_in_ms);	// seeks to point in stream (in ms). Usually you signal yoru thread to seek, which seeks and calls outMod->Flush()..

	// volume stuff
	WA_SetVolume, // void (*SetVolume)(int volume);	// from 0 to 255.. usually just call outMod->SetVolume
	WA_SetPan,    // void (*SetPan)(int pan);	// from -127 to 127.. usually just call outMod->SetPan

	// in-window builtin vis stuff
	NULL, // void (*SAVSAInit)(int maxlatency_in_ms, int srate);
         // call once in Play(). maxlatency_in_ms should be the value returned from outMod->Open()
         // call after opening audio device with max latency in ms and samplerate
	NULL, // void (*SAVSADeInit)(); call in Stop()


	// simple vis supplying mode
	NULL, // void (*SAAddPCMData)(void *PCMData, int nch, int bps, int timestamp);
         // sets the spec data directly from PCM data
         // quick and easy way to get vis working :)
         // needs at least 576 samples :)

	// advanced vis supplying mode, only use if you're cool. Use SAAddPCMData for most stuff.
   NULL, // int (*SAGetMode)(); gets csa (the current type (4=ws,2=osc,1=spec))
         // use when calling SAAdd()
   NULL, // void (*SAAdd)(void *data, int timestamp, int csa);
         // sets the spec data, filled in by winamp


   // vis stuff (plug-in)
   // simple vis supplying mode
   NULL, // void (*VSAAddPCMData)(void *PCMData, int nch, int bps, int timestamp);
         // sets the vis data directly from PCM data
         // quick and easy way to get vis working :)
         // needs at least 576 samples :)

	// advanced vis supplying mode, only use if you're cool. Use VSAAddPCMData for most stuff.
   NULL, // int (*VSAGetMode)(int *specNch, int *waveNch);
         // use to figure out what to give to VSAAdd
   NULL, // void (*VSAAdd)(void *data, int timestamp);
         // filled in by winamp, called by plug-in


	// call this in Play() to tell the vis plug-ins the current output params.
   NULL, // void (*VSASetInfo)(int nch, int srate);


	// dsp plug-in processing:
	// (filled in by winamp, called by input plug)

	// returns 1 if active (which means that the number of samples returned by dsp_dosamples
	// could be greater than went in.. Use it to estimate if you'll have enough room in the
	// output buffer
   NULL, // int (*dsp_isactive)();

   NULL, // int (*dsp_dosamples)(short int *samples, int numsamples, int bps, int nch, int srate);
         // returns number of samples to output. This can be as much as twice numsamples.
         // be sure to allocate enough buffer for samples, then.


	// eq stuff
   WA_EQSet, // void (*EQSet)(int on, char data[10], int preamp);
             // 0-64 each, 31 is +0, 0 is +12, 63 is -12. Do nothing to ignore.

	// info setting (filled in by winamp)
   NULL, // WA_SetInfo, // void (*SetInfo)(int bitrate, int srate, int stereo, int synched);
               // if -1, changes ignored? :)    (WHAT is minus one ? it returns VOID!?! )

	NULL  // Out_Module *outMod; filled in by winamp, optionally used :)

}; // end In_Module WA_sInModule   ( = WINAMP-specific part )



//---------------------------------------------------------------------------
//  Debugging stuff
//---------------------------------------------------------------------------
#if(USE_DEBUG)
void Debug(char *msg)
{
 if( WA_fDebug )
  {
    MessageBox( NULL, // handle of owner window
                 msg, // address of text in message box
      "Winamp2Speclab plugin", // address of title of message box
      MB_OK );        // style of message box
  }
} // end Debug
#endif


//---------------------------------------------------------------------------
//  Configuration subroutines
//---------------------------------------------------------------------------

#if(0)
#pragma argsused
static int CALLBACK WINAPI BrowseCallbackProc( HWND hwnd, UINT uMsg, LPARAM lParam, LPARAM lpData)
  // Callback function used with the SHBrowseForFolder function.
  // The browse dialog box calls this function to notify it about events.
{
  if (uMsg == BFFM_INITIALIZED)
   {
     SetWindowText(hwnd,"Select Directory");
     SendMessage(hwnd,BFFM_SETSELECTION,(WPARAM)1,(LPARAM)config_waveoutdir);
   }
  return 0;
}
#endif // (0,1)

#pragma argsused
BOOL WINAPI _DllMainCRTStartup(HANDLE hInst, ULONG ul_reason_for_call, LPVOID lpReserved)
{
  if( ul_reason_for_call == DLL_PROCESS_ATTACH )
     DisableThreadLibraryCalls( (HMODULE ) hInst);   // is this cast correct ?!
  return TRUE;
}


//---------------------------------------------------------------------------
//  Routines called by Winamp through the vectors in struct Out_Module :
//---------------------------------------------------------------------------


char *scanstr_back(char *str, char *toscan, char *defval)
{
  char *s=str+strlen(str)-1;
  if (strlen(str) < 1) return defval;
  if (strlen(toscan) < 1) return defval;
  while (1)
   {
     char *t=toscan;
     while (*t)
         if (*t++ == *s) return s;
     t=CharPrev(str,s);
     if (t==s) return defval;
     s=t;
   }
}

#if(0) // old stuff (when WB used windows messages to kick "the other side" alive)
//---------------------------------------------------------------------------
void W2S_ConnectDestinationWindow(void)
{ // Find the window handle of SpectrumLab (which is hopefully running already):
  W2S_hwndSpecLabWindow = FindWindow( "TSpectrumLab", NULL );
#if(USE_DEBUG)
  if( W2S_hwndSpecLabWindow!=NULL )
      Debug("Found SL's window handle");
#endif
}
#endif // (0,1)

#if(0) // old stuff (when this sourcecode was a winamp-OUTPUT-plugin)
#pragma argsused
//---------------------------------------------------------------------------
int WA_Open(int samplerate, int numchannels, int bitspersamp, int bufferlenms, int prebufferms)
  // Opens an audio output (device or file).
  //       Called by winamp (when this file was an OUTPUT plugin..)
  //   Returns >=0 on success, <0 on failure .
  //   NOTE: bufferlenms and prebufferms are ignored in most if not all output plug-ins.
  //         ... so don't expect the max latency returned to be what you asked for.
  //   returns max latency in ms (0 for diskwriters, etc)
  //   bufferlenms and prebufferms must be in ms. 0 to use defaults.
  //   prebufferms must be <= bufferlenms
{
  char *t,*p;
  char temp2[MAX_PATH],temp[MAX_PATH];

  // Prepare DL4YHF's "audio output buffer" which will be used to send the samples:
#if(USE_DEBUG)
  Debug("Init audio message buffer");
#endif
  AudioMsgBuf_Init( &W2S_pInputBuffer, samplerate, numchannels );
  W2S_pInputBuffer->info.i32StreamOptions = AUDIOMSG_OPT_BLOCK_SENDER;
  W2S_pInputBuffer->info.i32Volume = g_pSharedAudioBuffer->iAudioPlayerVolume;  // output volume, 0...255
  W2S_pInputBuffer->info.i32Balance= g_pSharedAudioBuffer->iAudioPlayerBalance; // output balance, -128..0..128
  W2S_ConnectDestinationWindow();  // detect destination window handle (-> W2S_hwndSpecLabWindow)

  if( WA_fDump2File )
   { // Original demo to open raw file output ...
     // may be useful for debugging so WoBu left it in

     GetWindowText(WA_sInModule.hMainWindow,temp2,sizeof(temp2));
     // t=temp2;    // BCB V4: "t is assigned a value that is never used". Indeed.
     t=scanstr_back(temp2,"-",NULL);
     if(t)
        t[-1]=0;

     if(temp2[0] && temp2[1] == '.')
      {
        char *p1,*p2;
        p1=temp;
        p2=temp2;
        while (*p2) *p1++=*p2++;
        *p1=0;
        p1 = temp2+1;
        p2 = temp;
        while (*p2)
         {
           *p1++ = *p2++;
         }
        *p1=0;
        temp2[0] = '0';
      }
     p=temp2;
     while (*p != '.' && *p) p++;
     if (*p == '.')
      {
        *p = '-';
        p=CharNext(p);
      }
     while (*p)
      {
        if( *p == '.' || *p == '/' || *p == '\\' || *p == '*' ||
            *p == '?' || *p == ':' || *p == '+' || *p == '\"' ||
            *p == '\'' || *p == '|' || *p == '<' || *p == '>') *p = '_';
         {  p=CharNext(p);
         }
      }

     p=config_waveoutdir;
     if (p[0]) while (p[1]) p++;

     if (!config_waveoutdir[0] || config_waveoutdir[0] == ' ')
          WA_Config(WA_sInModule.hMainWindow);
     if (!config_waveoutdir[0])
          wsprintf(temp,"%s.raw",temp2);
     else if (p[0]=='\\')
          wsprintf(temp,"%s%s.raw",config_waveoutdir,temp2);
     else
          wsprintf(temp,"%s\\%s.raw",config_waveoutdir,temp2);
     hwavfile = CreateFile(temp,GENERIC_READ|GENERIC_WRITE,0,NULL,CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL,NULL);
     if (!hwavfile)
        return -1;  // guess this means "error"

   } // end if( WA_fDump2File )

  w_offset = writtentime = 0;
  WA_iNumChannels = numchannels;
  WA_iSampleRate  = samplerate;
  // ex: WA_iBitsPerSample= bitspersamp;


  return 0;  // guess this means "no error"
}
#endif // (0,1)

#if(0) // old stuff (when this sourcecode was a winamp-OUTPUT-plugin)
void WA_Close(void)
{
  if( hwavfile )
    { CloseHandle(hwavfile);
      hwavfile = NULL;
    }
}
#endif // (0,1)

#if(0) // old stuff (when this sourcecode was a winamp-OUTPUT-plugin)
#pragma argsused
int WA_Write(char *buf, int len)
   // Writes audio samples to the output (file or whatever).
   // Params: Len == bytes to write (<= 8192 always).
   //         buf is straight audio data.
   // Return: 0 on success.
   //         1 returns not able to write (yet). Non-blocking, always.
{
  // ex: int t;   // what is "t", and why is it "int" ?!?
  DWORD dwBytesWritten;
  static BOOL first_call=TRUE;
  int   i32Result;

  writtentime += len;

  if( hwavfile )  // dump the samples to a file (for debugging purposes ?)
   {
#if(USE_DEBUG)
     if( first_call )
         Debug("Writing to dump file");
#endif

     // WriteFile(hwavfile,buf,len,&t,NULL); // BC: "cannot convert int* to unsigned long*"
     WriteFile(hwavfile,buf,len,&dwBytesWritten,NULL); // BC: "cannot convert int* to unsigned long*"
   }


  // Send the audio samples to SpecLab ?
  if(W2S_hwndSpecLabWindow==NULL)  // try again to find the destination window (by its class name):
     W2S_ConnectDestinationWindow();  // detect destination window handle (-> W2S_hwndSpecLabWindow)

  if(W2S_hwndSpecLabWindow!=NULL && W2S_pInputBuffer!=NULL)
   { // ok to fill the output buffer, and to try sending the data to "the other application" :
     if(WA_iBitsPerSample==16)
      {
#if(USE_DEBUG)
        if( first_call )
            Debug("Filling output buffer");
#endif
        W2S_pInputBuffer->dwBufIndexIn = W2S_pInputBuffer->dwBufIndexOut = 0;
        // Note: 'Len' = number of bytes to write.
        AudioMsgBuf_FillFromInt16(
               W2S_pInputBuffer,   // destination (buffer for sending)
               (SHORT*)buf,         // source (Winamp's data block)
               len/sizeof(SHORT));  // number of 16-bit-int's to send
      }
#if(USE_DEBUG)
     if( first_call )
         Debug("Sending output buffer");
#endif
     // Send the audio sample block; here: using WM_COPYDATA ....
     i32Result = AudioMsgBuf_SendAudio_WMSG( W2S_pInputBuffer, // source buffer
                  WA_sInModule.hMainWindow, // my window handle
                  W2S_hwndSpecLabWindow );      // his window handle
     if( i32Result == AUDIOMSG_RESULT_CANT_SEND )
      { // Most likely, the destination window doesn't exist. Forget the handle,
        // so the main application will try to find the destination window again .
        W2S_hwndSpecLabWindow = NULL;  // forget this handle - it seems to be invalid
      }

#if(USE_DEBUG)
     if( first_call )
         Debug("Sent output buffer");
#endif
   } // end if(W2S_hwndSpecLabWindow!=NULL && W2S_pInputBuffer!=NULL)

  Sleep(0);
  return 0;  // success
} // end WA_Write()
#endif // (0,1)

//----------------------------------------------------------------------
// From Winamp SDK, formerly in "config.c" ....
//----------------------------------------------------------------------

static char INI_FILE[MAX_PATH];
extern char config_waveoutdir[MAX_PATH];

static void _r_s(char *name,char *data, int mlen)
{
  char buf[2048];
  strcpy(buf,data);
  GetPrivateProfileString(WA_szAppName,name,buf,data,mlen,INI_FILE);
}
#define RS(x) (_r_s(#x,x,sizeof(x)))
#define WS(x) (WritePrivateProfileString(WA_szAppName,#x,x,INI_FILE))

static void config_init(void)
{
  char *p=INI_FILE;
  GetModuleFileName(NULL,INI_FILE,sizeof(INI_FILE));
  while (*p) p++;
  while (p >= INI_FILE && *p != '.') p--;
  strcpy(p+1,"ini");
}

void config_read()
{
  config_init();
  RS(config_waveoutdir);
}

void config_write()
{
  WS(config_waveoutdir);
}
// <<<<<<<<<<<< end of the Winamp-specific stuff <<<<<<<<<<<<<<<<<<<<<<<<<<<<

//---------------------------------------------------------------------------
// Interface to the "External Source"
//    (which is SpecLab, or other programs which can load this DLL)
//---------------------------------------------------------------------------

//----------------------------------------------------------------------
// DLL functions called by the "external audio source" .
//   NO UNDERSCORE, NO NAME MANGLING ! (read the long story above)
//----------------------------------------------------------------------


//----------------------------------------------------------------------
// Mandatory functions.
//     Prototypes must precisely match the DLL-function-pointers
//     in ?\AudioIO.c !  (also, use INT32 here, not "int"...
//     otherwise we end up as a complete mess should "int" be 64 bit 
//     one day.  Remember the days when "int" was always 16 bit ? Haha..
//----------------------------------------------------------------------

//----------------------------------------------------------------------
__declspec( dllexport ) const char *AIO_ErrorCodeToString( INT32 iAIOErrorCode )
  // Converts a (negative) Audio-I/O error code into a human readable string.
  // Since some error codes (numbers below minus 1000) may be DLL-specific,
  // this function is also located inside the DLL (and not in AudioIO.c) .
{
    switch( iAIOErrorCode )
      { // Caution: When adding new error strings here, make sure to modify:
        //  - C:\cbproj\AudioIO\AudioIO.c    (Audio-I/O-"Host" for applications)
        //  - c:\cbproj\AudioIO\aio2winamp.c (reference implementation of a DLL)
        case AIO_NO_ERROR              :  return "no error";                      // 0
        case AIO_GENERAL_ERROR         :  return "general error";                 // -1
        case AIO_ERROR_DLL_NOT_LOADED  :  return "DLL not loaded";                // -2
        case AIO_ERROR_DLL_NOT_COMPATIBLE: return "DLL not compatible";           // -3
        case AIO_ERROR_NOT_IMPLEMENTED :  return "function not implemented";      // -4
        case AIO_ERROR_INVALID_PARAMETER: return "invalid parameter";             // -5
        case AIO_ERROR_OUT_OF_RESOURCES:  return "out of resources";              // -6
        case AIO_ERROR_FORMAT_NOT_SUPPORTED: return "format not supported";       // -10
        case AIO_ERROR_SAMPRATE_NOT_SUPPORTED: return "samplerate not supported"; // -11
        case AIO_ERROR_FILE_NOT_FOUND  :  return "file not found";                // -12
        case AIO_ERROR_INPUT_NOT_OPEN  :  return "input not open";                // -14
        case AIO_ERROR_OUTPUT_NOT_OPEN :  return "output not open";               // -15
        case AIO_INVALID_HANDLE        :  return "invalid handle";                // -17
        case AIO_ERROR_CANCELLED       :  return "cancelled by user";             // -18
        case AIO_ERROR_CANNOT_WRITE_YET:  return "cannot write yet";              // -20
        case AIO_ERROR_CANNOT_READ_YET :  return "cannot read yet";               // -21
        default:
           if( iAIOErrorCode >= 0 )
            { return "positive results are no errors";
            }
           else
            { return "unknown error code";
            }
      } // end switch
} // end AIO_ErrorCodeToString()


//----------------------------------------------------------------------
__declspec( dllexport ) INT32 AIO_GetOutputParams(
        char *pszAudioStreamID, // [in] 'Stream ID' for the Audio-I/O-DLL, identifies the AUDIO SOURCE
           // (here: optional, use NULL to query info about "the first available source")
        double *pdblSampleRate, // [out] samples per second, "precise", optional
        INT32 *piNumChannels)   // [out] 1=mono, 2=stereo, 4=quadrophonic or 2*I/Q
   // Called by the application to query the current settings
   //  for a certain "outgoing" audio connection - BEFORE OPENING IT .
   // (*) For an explanation of the "audio stream ID", see
   //     AIO_Host_OpenAudioOutput() in C:\cbproj\AudioIO\AudioIO.c !
{
  int iAudioSource;
  T_AudioSourceInfo *pAudioSourceInfo;
  T_AudioBuffer *pAudioBuffer = GetAudioBufferByStreamID( pszAudioStreamID,
          AUDIO_DONT_CREATE_BUFFER, &pAudioSourceInfo, &iAudioSource);
  if( pAudioBuffer!=NULL && pAudioSourceInfo!=NULL )
   {
     if(pdblSampleRate)
      { *pdblSampleRate = pAudioSourceInfo->dblSampleRate;
      }
     if(piNumChannels )
      { *piNumChannels = pAudioSourceInfo->nChannelsPerSamplePoint;
      }
     // if(piBitsPerSample)
     //  { *piBitsPerSample = 16; // ex: g_pSharedAudioBuffer->iBitsPerSample;
     //  }

     return AIO_NO_ERROR;
   }
  else
   { return AIO_INVALID_HANDLE;  // here: problem with the shared memory-mapped file
   }
} // end AIO_GetOutputParams()

//----------------------------------------------------------------------
__declspec( dllexport ) INT32 AIO_GetWinampParams(
       double *pdblSampleRate, // [out] samples per second used by Winamp
       INT32 *piNumChannels,     // [out] 1=mono, 2=stereo   used by Winamp
  //ex:INT32 *piBitsPerSample,   // [out] usually 16 (bits per sample)
       INT32 *piAudioPlayerInitialized, // [out] 0=winamp not connected to the DLL,  1=winamp connected(*)
       INT32 *piAudioPlayerInputPaused) // [out] 0=winamp's input is paused, 1=winamp's input is running
  // (*) "connected" means, Winamp is there, has loaded the plugin DLL,
  //     and has called the "Init()"-function of the Winamp-plugin.
  //     As soon as Winamp calls the "Quit"-function, the flag gets cleared.
  // All pointers in the argument list are OPTIONAL.
  //  If you don't need some of the above values, just pass a NULL pointer .
{
  T_AudioSourceInfo *pAudioSourceInfo;
  T_AudioBuffer *pAudioBuffer;

  if( GetAudioBufferAndSourceInfoForWinamp(
               &pAudioBuffer,      // [out] audio buffer to be used by Winamp
               &pAudioSourceInfo)) // [out] source-info to be used by Winamp
   {
     if(pdblSampleRate)
      { *pdblSampleRate = pAudioSourceInfo->dblSampleRate;
      }
     if(piNumChannels )
      { *piNumChannels  = pAudioSourceInfo->nChannelsPerSamplePoint;
      }
  // if(piBitsPerSample)
  //  { *piBitsPerSample = 16; // ex: g_pSharedAudioBuffer->iBitsPerSample;
  //  }
     if(piAudioPlayerInitialized)
      { *piAudioPlayerInitialized = pAudioSourceInfo->iAudioPlayerInitialized;
      }
     if(piAudioPlayerInputPaused)
      { *piAudioPlayerInputPaused = pAudioSourceInfo->iAudioPlayerInputPaused;
      }
     return AIO_NO_ERROR;
   }
  else
   { return AIO_INVALID_HANDLE;  // here: problem with the shared memory-mapped file
   }

} // end AIO_GetWinampParams()

//----------------------------------------------------------------------
__declspec( dllexport ) INT32 AIO_OpenAudioOutput(
        char * pszAudioStreamID,  // [in] 'Stream ID' for the Audio-I/O-DLL, identifies the AUDIO SOURCE (here: mandatory)
        char * pszDescription,    // [in] descriptive name to identify the SOURCE.
           // This name may appear on a kind of "patch panel"
           // on the configuration screen of the Audio-I/O DLL .
           // For example, SpecLab will identify itself as "SpectrumLab1"
           // for the first running instance, "SpectrumLab2" for the 2nd, etc .
       double dblSampleRate,   // [in] samples per second, "precise"
       INT32  nChannelsPerSamplePoint)    // [in] 1=mono, 2=stereo, 4=quadrophonic
     //INT32  iBitsPerSample ) // [in] usually 16 (bits per sample)
  // Called from external audio source to open the DLL for "output"
  //   ("output" must seen from that application's point of view) .
  // If this was successfull, it may begin to periodically call
  //    AIO_WriteAudio() to send chunks of audio blocks .
  // Returns >= 0 on success,  < 0 on failure (like Winamp's output-plugins) .
  // Details (only in c:\cbproj\AudioIO\aio2winamp.c::AIO_OpenAudioOutput() ) :
  //  Usually, the audio writer and the audio reader are two different programs
  //  (or, at least, different instances of the same program running).
  //  We cannot assume anything about what happens first (for the same stream):
  //  (A) AIO_OpenAudioOutput() called before AIO_OpenAudioInput() .
  //      In this case, application 'W' (writer) which called AIO_OpenAudioOutput()
  //      will immediately start filling the buffer even if there is no reader yet.
  //  (B) AIO_OpenAudioInput() called before AIO_OpenAudioOutput() .
  //      In this case, application 'R' (reader) which called AIO_OpenAudioInput()
  //      may be blocked for a few hundred milliseconds (in AIO_OpenAudioInput()),
  //      but return if no other application starts writing into the buffer,
  //      and AIO_ReadInputSamplePoints() will return "no data" until
  //      application 'W' has been initialized and starts filling the buffer.
  //
{
  T_AudioSourceInfo * pAudioSourceInfo;
  T_AudioBuffer     * pAudioBuffer;
  int iAudioSource = 0;

  // ex: if( dblSampleRate < 5512.0 || dblSampleRate > 192000.0 )
  //      { return AIO_ERROR_SAMPRATE_NOT_SUPPORTED;
  //      }
  //     if( iBitsPerSample != 16 )
  //      { return AIO_ERROR_FORMAT_NOT_SUPPORTED;
  //      }
  //  removed; we shouldn't make any assumption about the SR capabilities here !

  if( nChannelsPerSamplePoint < 1   ||  nChannelsPerSamplePoint > AUDIO_IO_MAX_CHANNELS )
   { return AIO_ERROR_FORMAT_NOT_SUPPORTED;
   }

  pAudioBuffer = GetAudioBufferByStreamID( pszAudioStreamID,
                    AUDIO_CREATE_BUFFER_IF_NOT_EXISTING, // [in] iCreateOptions
                    &pAudioSourceInfo,    // [out] address of audio-source info
                    &iAudioSource);  // [out] index into sources[], will be part of the 'handle'
  if( pAudioBuffer!=NULL && pAudioSourceInfo!=NULL )
   { pAudioSourceInfo->dblSampleRate = dblSampleRate;
     if( pszAudioStreamID == NULL )
      {  pszAudioStreamID = "_unnamed_";  // name MUST NOT BE EMPTY (to mark the entry as "allocated")
      }
     strncpy( pAudioSourceInfo->sz40AudioStreamID, pszAudioStreamID, 40 );
     pAudioSourceInfo->sz40AudioStreamID[40] = '\0';
     if( pszDescription != NULL )
      { strncpy( pAudioSourceInfo->sz80Description, pszDescription, 40 );
        pAudioSourceInfo->sz80Description[80] = '\0';
      }
     else
      { pAudioSourceInfo->sz80Description[0] = '\0';
      }

     // Need to re-initialize the buffer ?
     if( pAudioSourceInfo->nChannelsPerSamplePoint != nChannelsPerSamplePoint )
      { pAudioSourceInfo->nChannelsPerSamplePoint  = nChannelsPerSamplePoint;
        pAudioBuffer->iResetRequest = 1;
      }
     pAudioBuffer->iWriterDllInstanceIndex = iThisInstancesDllInstanceIndex;
     pAudioBuffer->dwWriterFlags |= AUDIO_WRITER_OPENED;
     pAudioSourceInfo->dwWriterFlags = pAudioBuffer->dwWriterFlags; // <- just a COPY !
     SetFlagToUpdateControlPanels(); // let all visible control panels update themselves
     // When successfull, the return value is a NON-NEGATIVE HANDLE, compatible with GetAudioBufferByHandle():
     return iAudioSource<<8;
   }
  else
   { return AIO_INVALID_HANDLE;  // here: problem with the shared memory-mapped file
   }

} // end AIO_OpenAudioOutput()


//----------------------------------------------------------------------
__declspec( dllexport ) INT32 AIO_CanWriteOutput(
        INT32 iAudioIoHandle )   // [in] audio-handle, see ..OpenAudioOutput()
  // Returns number of SAMPLE POINTS possible to write at a given time.
  // Never will decrease unless you call Write (or Close)..
  // like in Winamp's output plugin .
{
  INT32 i32HeadIndex, i32TailIndex, i32UsedBufEntries, i32FreeBufEntries;
  int  iSourceIndex,iReaderIndex;
  T_AudioSourceInfo *pAudioSourceInfo;
  T_AudioBuffer *pAudioBuffer = GetAudioBufferByHandle( // may return NULL if something's wrong !
                    iAudioIoHandle, // [in] handle returned by .._OpenAudioOutput() / _OpenAudioInput()
                     &iSourceIndex, // [out,optional] index into T_SharedMemoryData.source[]
                     &iReaderIndex, // [out,optional] index into T_AudioBuffer.reader[]
                &pAudioSourceInfo); // [out,optional] address of the T_AudioSourceInfo struct
  if( pAudioBuffer==NULL || pAudioSourceInfo==NULL )
   { return AIO_INVALID_HANDLE;  // here: problem with the shared memory-mapped file
   }
  if( (pAudioBuffer->dwWriterFlags & AUDIO_WRITER_OPENED) == 0 )
   { return AIO_ERROR_OUTPUT_NOT_OPEN;
   }


  i32HeadIndex = (INT32)(pAudioBuffer->dwHeadIndex & (AUDIO_BUFFER_SIZE-1) );
  i32TailIndex = (INT32)(pAudioBuffer->reader[iReaderIndex].dwTailIndex & (AUDIO_BUFFER_SIZE-1) );
  i32UsedBufEntries = i32HeadIndex - i32TailIndex;


  if( i32UsedBufEntries < 0 )  // index wrapped around..
   {  i32UsedBufEntries += AUDIO_BUFFER_SIZE;
   }
  i32FreeBufEntries = (AUDIO_BUFFER_SIZE-1) - i32UsedBufEntries;
  // Note: this special type of circular buffer cannot be filled to 100.0 %,
  //       because dwHeadIndex==dwTailIndex means "the buffer is completely empty"
  //       (not "completely full") .

  if( pAudioBuffer->nChannelsPerSamplePoint > 1 )
     i32FreeBufEntries /= pAudioBuffer->nChannelsPerSamplePoint;


  return i32FreeBufEntries;
} // end AIO_CanWriteOutput()


//----------------------------------------------------------------------
__declspec( dllexport ) INT32 AIO_WriteOutputSamplePoints(
        INT32 iMyHandle,          // [in] handle to the input "device"
        float *pfltSource,        // [in] audio samples, as 32-bit FLOATING POINT numbers, grouped as "sample points",
                                  //      USUALLY (*) normalized to -1.0 ... +1.0 for full ADC swing
        INT32 iNumSamplePoints,   // [in] number of SAMPLE POINTS(!) to write or send
        INT32 nChannelsPerSample, // [in] number of samples PER SAMPLE POINT
        INT32 iTimeout_ms ,       // [in] max timeout in milliseconds, or AIO_TIMEOUT_NON_BLOCKING
        T_ChunkInfo *pInChunkInfo, // [in,optional] chunk info, see c:\cbproj\SoundUtl\ChunkInfo.h, MAY BE NULL(!)
        INT32 *piHaveWaited_ms )  // [out,optional] "have been waiting here for XX milliseconds"
  // Replacement for AIO_WriteAudioOutput(),
  //  supports the 'T_ChunkInfo' struct defined in c:\cbproj\SoundUtl\ChunkInfo.h
  //  (with precise timestamp, calibrated sample rate, GPS date+time+position, etc),
  //  and supports multi-channel audio (grouped with MULTIPLE CHANNELS per "sample point") .
  //
  // Return : >= 0 on success (indicates the number sample points which can be
  //                           placed in the buffer if we'd immediately
  //                           call AIO_WriteAudioOutput() again ) .
  //          < 0  on error :
  //    AIO_ERROR_CANNOT_WRITE_YET : cannot write YET (please wait..)
  //    AIO_ERROR_OUTPUT_NOT_OPEN  : cannot write, and won't be able to write
  //                           (not opened, or "the other side" has shut down)
  //
  // Implementation details :
  //   - There is only ONE WRITER for a certain audio source,
  //     but 0, 1, or multiple READERS.
  //   - There is NO conversion of the sampling rate !
  //     Thus, the sampling rate is dictated by the WRITER, not by any of the
  //     READERS.
  //   - The buffer (shared memory) holds approximately one SECOND of audio,
  //     thus the readers have plenty of time to read and process the stream.
  //   - (*) If the input is NOT normalized to -1.0 ... +1.0 ,
  //     the 'T_ChunkInfo' structure must be passed in,
  //     with pInChunkInfo->dblSampleValueRange set to the sample value range.
  //
{
  DWORD dwHeadIndex, dwFirstHeadIndex, dwHeadIndexIncr;
  int   nFreeSamplePointsInOutputBuffer;
  int   i,nReaders,iAudioSource,iReaderIndex, iChannel, nChannels;
  int   iLUT, iLUT1, iLUT2;  // indices into the 'chunk info LookUpTable'
  int   iWaited_ms = 0;
  float flt, *pfltDest;
  float fltPosPeak[AUDIO_IO_MAX_CHANNELS], fltNegPeak[AUDIO_IO_MAX_CHANNELS];
  T_AudioSourceInfo *pAudioSourceInfo;
  T_AudioBuffer *pAudioBuffer = GetAudioBufferByHandle( // may return NULL if something's wrong !
             iMyHandle, // [in] handle returned by .._OpenAudioOutput() / _OpenAudioInput()
         &iAudioSource, // [out,optional] index into T_SharedMemoryData.source[]
         &iReaderIndex, // [out,optional] index into T_AudioBuffer.reader[]
    &pAudioSourceInfo); // [out,optional] address of the T_AudioSourceInfo struct
  if( pAudioBuffer==NULL || pAudioSourceInfo==NULL )
   { return AIO_INVALID_HANDLE;  // here: problem with the shared memory-mapped file
   }

  if( piHaveWaited_ms != NULL )
   { *piHaveWaited_ms = 0;
   }

  nChannels  = nChannelsPerSample;  // ...in THE CALLER's source buffer...
  if( nChannels > pAudioSourceInfo->nChannelsPerSamplePoint )
   {  nChannels = pAudioSourceInfo->nChannelsPerSamplePoint;   // limit to the number of channels in the DESTINATION-buffer
   }
  if(nChannels<1 || nChannels>AUDIO_IO_MAX_CHANNELS || nChannelsPerSample<1)
   { return AIO_ERROR_INVALID_PARAMETER;
   }
  for( iChannel=0; iChannel<AUDIO_IO_MAX_CHANNELS; ++iChannel )
   { fltPosPeak[iChannel] = fltNegPeak[iChannel] = 0.0;
   }
  // The number of channels is VARIABLE; see T_AudioBuffer.nChannelsPerSamplePoint .
  // The 'number of channels per SAMPLE POINT' shouldn't change
  // (it is set only ONCE in AIO_OpenAudioOutput() ) !
  // FUTURE PLAN: Let *DIFFERENT WRITERS* fill *DIFFERENT CHANNELS* in the buffer,
  //              thus T_AudioBuffer.nChannelsPerSamplePoint may be larger than
  //              the function argument 'nChannelsPerSample' !
  // we MAY have to repeat this loop (if this shall be a "blocking" call)...
  while(1)
   {
     dwHeadIndex = pAudioBuffer->dwHeadIndex & (AUDIO_BUFFER_SIZE-1);
     nFreeSamplePointsInOutputBuffer = GetNumberOfFreeSamplePointsInOutputBuffer( iAudioSource, &nReaders );
     // Note: The circular buffer used here cannot be filled to 100.0 %,
     //       because dwHeadIndex==dwTailIndex means "the buffer is completely empty"
     //       (not "completely full") .  But it's thread-safe by design .
     if( nFreeSamplePointsInOutputBuffer < iNumSamplePoints )
      { // There are LESS FREE ENTRIES in the output-buffer than we need,
        // most likely because one of the readers couldn't keep up the pace.
        if( iTimeout_ms <= 0 )
         { // non-blocking call (or max waiting-time exceeded) ? return !
           return AIO_ERROR_CANNOT_WRITE_YET;
         }
        else // wait a few dozen milliseconds, then try again
         { Sleep(50);
           iTimeout_ms -= 50;
           iWaited_ms  += 50;
           if( piHaveWaited_ms != NULL )
            { *piHaveWaited_ms = iWaited_ms;
            }
         }
      } // end if < not enough "free" buffer space yet >
     else // enough free entries in the audio buffer now :
      { break;  // exit from waiting-loop
      }
   } // end while < waiting until there's enough space in the output-buffer >

  // Sleep(20); // TEST (!)

  // Arrived here: there are enough "free buffer entries" to write the samples
  //   into the shared audio buffer .
  nFreeSamplePointsInOutputBuffer -= iNumSamplePoints; // used as result code (later)
  if( nFreeSamplePointsInOutputBuffer < 0 )
   {  nFreeSamplePointsInOutputBuffer = 0;  // bug ?
   }
  dwFirstHeadIndex = dwHeadIndex;
  dwHeadIndexIncr = (DWORD)pAudioSourceInfo->nChannelsPerSamplePoint;
  for(i=0; i<iNumSamplePoints; ++i )
   { pfltDest = pAudioBuffer->fltBuffer + dwHeadIndex;
     for( iChannel=0; iChannel<nChannels; ++iChannel )
      { pfltDest[iChannel] = flt = pfltSource[iChannel];
        if( flt>fltPosPeak[iChannel] )
         { fltPosPeak[iChannel] = flt;
         }
        if( flt<fltNegPeak[iChannel] )
         { fltNegPeak[iChannel] = flt;
         }
      }
     pfltSource  += nChannelsPerSample;
     dwHeadIndex += dwHeadIndexIncr;
     if( dwHeadIndex >= AUDIO_BUFFER_SIZE )
      { dwHeadIndex = 0;
        pfltDest = pAudioBuffer->fltBuffer;
      }
   }
  pAudioBuffer->dwHeadIndex = dwHeadIndex;
  i = pAudioBuffer->nSamplePointsBuffered + iNumSamplePoints;
  if( i>pAudioBuffer->nSamplePointsMax )
   {  i=pAudioBuffer->nSamplePointsMax;
   }
  pAudioBuffer->nSamplePointsBuffered = i;
  if( pInChunkInfo!=NULL )   // copy some info from the most recent T_ChunkInfo ?
   { // Note that the first 256 bytes of the T_ChunkInfo structure were "always there",
     //  so there's no need to examine pInChunkInfo->dwSizeOfStruct for the following:
     if( pInChunkInfo->dblSampleValueRange>0 )
      { pAudioSourceInfo->dblSampleValueRange = pInChunkInfo->dblSampleValueRange;
      }
     // Queue up the T_ChunkInfo entries in an extra buffer .
     // For details, see AudioBuf_GetChunkInfoIndexForSampleIndex():
     iLUT = iLUT1 = dwFirstHeadIndex / AUDIO_SAMPLES_PER_INFO_BLOCK; // first index into the 'chunk info LookUpTable'
     pAudioBuffer->ChunkInfoArray[iLUT1] = *pInChunkInfo;       // BLOCK-COPY of 256 bytes
     pAudioBuffer->ChunkInfoArray[iLUT1].nChannelsPerSample = pAudioBuffer->nChannelsPerSamplePoint;
     pAudioBuffer->iLatestChunkInfoArrayIndex = iLUT1; // extra service for AudioBuf_GetLatestChunkInfo()
     if( dwHeadIndex > 0 )
      { iLUT2 = (dwHeadIndex-1) / AUDIO_SAMPLES_PER_INFO_BLOCK; // last index into the 'chunk info LookUpTable'
      }
     else
      { iLUT2 = 0;
      }
     // Note that pAudioBuffer->ChunkInfoLUT[iLUT1 .. iLUT2] is a CIRCULAR (wrapping) buffer !
     // The index range of iLUT is 0..<AUDIO_MAX_CHUNK_INFO_BLOCKS-1> (=255) .
     // Examples:   iLUT1 = 0;    iLUT2 = 0    ->  only set ChunkInfoLUT[0]
     //             iLUT1 = 255;  iLUT2 = 0    ->  set ChunkInfoLUT[255] + [0]
     //             iLUT1 = 254;  iLUT2 = 3    ->  set ChunkInfoLUT[254,255,0,1,2,3]
     i = 0;  // just for safety: loop limiter.
     do
      { pAudioBuffer->ChunkInfoLUT[iLUT].iAudioBufferIndex = dwFirstHeadIndex; // index into fltBuffer[]
        pAudioBuffer->ChunkInfoLUT[iLUT].iChunkArrayIndex  = iLUT1; // index into ChunkInfoArray[]
        ++iLUT;
        if( iLUT>=AUDIO_MAX_CHUNK_INFO_BLOCKS )
         {  iLUT = 0;
         }
      } while( (iLUT != iLUT2) && ((i++)<AUDIO_MAX_CHUNK_INFO_BLOCKS) );

   } // end if( pInChunkInfo!=NULL )

  // Set the new peak-values (accessable through the shared memory block)
  for( iChannel=0; iChannel<AUDIO_IO_MAX_CHANNELS; ++iChannel )
   { flt = fltPosPeak[iChannel];
     if( fltNegPeak[iChannel] < -flt )
      { flt = -fltNegPeak[iChannel];
      }
     if( pAudioSourceInfo->dblSampleValueRange>0 )  // normalize the peak-value range to 0..1 ?
      { flt /= pAudioSourceInfo->dblSampleValueRange;
      }
     pAudioSourceInfo->fltAudioPeakValue[iChannel] = flt;
   }


  // Calculate the 'buffer usage' in percent (for display in the control panel) .
  // 2013-11-10: When not a single 'reader' was attached to an audio source,
  //             the buffer usage indicator (bargraph) remained at ZERO,
  //             because nFreeSamplePointsInOutputBuffer didn't go up .
  //             Indeed it's correct to return with
  //             nFreeSamplePointsInOutputBuffer = 'the entire buffer',
  //   because (quoted from GetNumberOfFreeSamplePointsInOutputBuffer() :
  //    > Returns the number of 'unused' entries in T_AudioBuffer.fltBuffer[]
  //    > for the SLOWEST (but active) reader of this buffer.
  //    > If no reader is active, the return value indicates an 'empty' buffer.
  //   But HERE, for the 'buffer usage bargraph' (bu), things are different:
  if( pAudioBuffer->nSamplePointsMax <= 0 )
   { // buffer cannot store anything at all -> usage is ZERO !
     pAudioSourceInfo->iCalculatedBufferUsage_0to255 = 0;
   }
  else if( nReaders < 1 )
   { // Added 2013-11-10: If there's NO ACTIVE READER AT ALL, show the
     // 'number of sample points buffered' which does NOT depend on the presence of readers:
     i = pAudioBuffer->nSamplePointsBuffered;
     pAudioSourceInfo->iCalculatedBufferUsage_0to255 = (255*i) / pAudioBuffer->nSamplePointsMax;
   }
  else  // similar if one or more readers are 'attached' :
   { i = pAudioBuffer->nSamplePointsMax - nFreeSamplePointsInOutputBuffer;
     pAudioSourceInfo->iCalculatedBufferUsage_0to255 = (255*i) / pAudioBuffer->nSamplePointsMax;
   }




  // Return value for "success": zero or positive .
  //   The number indicates HOW MANY SAMPLE POINTS (with N channels PER POINT)
  //   can be placed in the buffer if we'd immediately
  //   call the audio-write-function again .
  // This allows the "External Source" to fill the buffer as fast as possible .
  return nFreeSamplePointsInOutputBuffer;
} // end AIO_WriteOutputSamplePoints()



//----------------------------------------------------------------------
__declspec( dllexport ) void AIO_CloseAudioOutput(
     INT32 iAudioIoHandle ) // [in] audio-handle, from ..OpenAudioOutput()
  // Polite applications call this function when they have nothing more to send,
  // or when they shut down.
{
  int iAudioSource;
  T_AudioSourceInfo *pSourceInfo;
  T_AudioBuffer *pAudioBuffer = GetAudioBufferByHandle( // may return NULL if something's wrong !
                                  iAudioIoHandle/*in*/,
                                  &iAudioSource, // [out,optional] index into T_SharedMemoryData.source[]
                                  NULL/*iReaderIndex not needed*/,
                                  &pSourceInfo/*out: address of audio-source-info*/ );
  if( pAudioBuffer!=NULL )
   { // Mark this audio-buffer as 'no longer open for writing',
     //  but DO NOT DELETE THE BUFFER YET (it may still be read by others).
     //  The shared memory (actually a memory mapped, shared FILE)
     //  will only be deleted when no one uses it anymore;
     //  see details near case DLL_PROCESS_DETACH in the "DLL-entry-point" .
     pAudioBuffer->dwWriterFlags &= ~AUDIO_WRITER_OPENED;
     pSourceInfo->dwWriterFlags = pAudioBuffer->dwWriterFlags; // <- just a COPY !
     SetFlagToUpdateControlPanels(); // let all visible control panels update themselves
   }

} // end AIO_CloseAudioOutput()




//----------------------------------------------------------------------
__declspec( dllexport ) INT32 AIO_OpenAudioInput(
        char *pszAudioStreamID,   // [in] 'Stream ID' for the Audio-I/O-DLL, identifies the AUDIO SOURCE
           // (here: optional, use NULL to use "the first available source")
        char *pszAudioReaderID,   // [in] short but hopefully UNIQUE name which identifies
           // the audio sink ("audio consumer"). This name may appear
           // on a kind of "patch panel" on the configuration screen of the Audio-I/O DLL .
           // For example, SpecLab will identify itself as "SpectrumLab1"
           // for the first running instance, "SpectrumLab2" for the 2nd, etc .
        double  dblWantedSampleRate,  // [in(!)] "wanted" sample rate, often IGNORED (reasons below..),
                                      //   may be negative if the reader accepts 'any' sample rate
        double *pdblOutSampleRate,    // [out(!)] "realized" sample rate, samples per second,
           // Usually dictated by the FEEDING SOURCE;
           //   the sampling rate is NOT resampled by the DLL !
           //   dblWantedSampleRate can be used, however, for hardware drivers
           //   (like soundcard or SDR drivers) if the hardware supports
           //   different 'native ' sample rates.
        INT32 nChannelsPerSample,     // [in] number of channels PER SAMPLE POINT
        INT32 iTimeout_ms ,           // [in] max timeout in milliseconds for THIS call
           // (may have to wait for a short time, for example when...
           //   - the I/O lib needs to communicate with an external device (to open it)
           //   - the I/O lib needs to wait until the 'writer' (=the other side)
           //     has called AIO_Host_OpenAudioOutput() and set the sampling rate, etc
        INT32 iOpenOptions,          // [in] bit combination of AIO_OPEN_OPTION_... flags
        T_ChunkInfo *pOutChunkInfo)  // [out,optional] chunk info, see c:\cbproj\SoundUtl\ChunkInfo.h, MAY BE NULL (!)
  // Return value:
  //     >= 0 on success,  in this case it's a "handle" value
  //                        which must be passed to AIO_WriteAudioOutput later.
  //      < 0 on failure (like Winamp's output-plugins) .
  //
  // Specification: in AudioIO.h::AIO_Host_OpenAudioInput() .
  // Details, especially about the calling sequence
  //           of AIO_Host_OpenAudioOutput() and AIO_Host_OpenAudioInput() :
  //          ONLY in c:\cbproj\AudioIO\aio2winamp.c::AIO_OpenAudioOutput() !
{
  int iAudioSource, iReaderIndex, iRetryNr;
  BOOL fTryAgain;
  T_AudioSourceInfo * pAudioSourceInfo;
  T_AudioReader * pReader;
  T_AudioBuffer * pAudioBuffer = GetAudioBufferByStreamID(
                    pszAudioStreamID,   // [in,optional] stream ID, NULL="use the first existing stream"
                    AUDIO_CREATE_BUFFER_IF_NOT_EXISTING, // [in] iCreateOptions
                    &pAudioSourceInfo , // [out,optional] address of the audio-source-info
                    &iAudioSource); // [out]0..AUDIO_MAX_SOURCES-1, will become part of the 'handle'
  char *cp,sz40AudioReaderID[48],sz7Array[8];

  if( pszAudioReaderID != NULL ) // if the caller provides a NAME, save it here:
   { strncpy( sz40AudioReaderID, pszAudioReaderID, 40 );
     sz40AudioReaderID[40] = '\0';
   }
  else
   { sz40AudioReaderID[0] = '\0';
   }
  if(sz40AudioReaderID[0]==0) // provide a unique name for this caller .. but how ?
   { sprintf( sz40AudioReaderID, "NoName%d",(int)iThisInstancesDllInstanceIndex );
   }


  (void)dblWantedSampleRate;  // unused parameter
  (void)iTimeout_ms;
  (void)iOpenOptions;

  if( (pAudioBuffer==NULL) || (g_pSharedMemoryData==NULL) || (pAudioSourceInfo==NULL) )
   { return AIO_INVALID_HANDLE;  // here: problem with the shared memory-mapped file
   }

  if( pszAudioStreamID!=NULL ) // the audio reader wants to 'connect' a certain source:
   {
   } // end if( pszAudioStreamID!=NULL )

  // Check if the caller already had THE SAME reader opened before.
  //   (if so, the caller may have failed to call AIO_CloseAudioInput;
  //    in that case this DLL won't allow him to eat up more resources)
  for(iRetryNr=2; iRetryNr<=9; ++iRetryNr ) // only try pseudo-array-indices 2..9
   { fTryAgain = FALSE;
     for( iReaderIndex=0;  iReaderIndex<AUDIO_MAX_READERS;  ++iReaderIndex )
      { pReader = &pAudioBuffer->reader[iReaderIndex];
        if( pReader->dwFlags & AUDIO_READER_OPENED )
         { // Here is an 'already opened' reader ... is it the ID used by the caller ?
           if( strcmp( pReader->sz40ReaderID, sz40AudioReaderID ) == 0 )
            { // The 'reader ID' (string) is the same, but is it really THE SAME CALLER ?
              if( pReader->iDllInstanceIndex == iThisInstancesDllInstanceIndex )
               { // Most likely, this is THE SAME caller, using the same "reader ID",
                 // who already opened this stream for input but failed to close it
                 // -> CLOSE IT AUTOMATICALLY before re-opening it further below.
                 pAudioBuffer->reader[iReaderIndex].dwFlags = 0;
               }
              else // same "reader ID" but a different DLL-instance-index :
               { // Most likely, there are two or more instances of THE SAME program
                 // running (which don't know about this, otherwise they'd use
                 // different audio-reader-IDs to avoid this problem).
                 // -> ALLOW TO OPEN A SECOND READER with (almost) the same reader-ID,
                 //    with an array-like suffix, for example "SpectrumLab[2]" :
                 cp = strrchr( sz40AudioReaderID, '[' );
                 if( cp!=NULL ) *cp='\0';
                 sprintf(sz7Array,"[%d]",(int)iRetryNr );
                 strcat( sz40AudioReaderID, sz7Array );
                 fTryAgain = TRUE;
               }
            } // end if < pReader->sz40ReaderID == sz40AudioReaderID >
         } // end if < .. AUDIO_READER_OPENED >
      } // end for iReaderIndex..
     if( ! fTryAgain )
      { break;
      }
   } // end for iRetryNr..

  // Allocate the first unused entry in the table of 'readers' for this stream:
  for( iReaderIndex=0;  iReaderIndex<AUDIO_MAX_READERS;  ++iReaderIndex )
   { pReader = &pAudioBuffer->reader[iReaderIndex];
     if( (pReader->dwFlags & AUDIO_READER_OPENED ) == 0 )
      { // Found an 'unused' entry in the table of audio readers -> use it
        pReader->dwFlags = AUDIO_READER_OPENED; // entry now occupied
        pReader->pvUserDataForCallback = NULL;
        pReader->pCallbackFunction = NULL; // callback not set yet
        pReader->dwTailIndex = pAudioBuffer->dwHeadIndex; // purge buffer for THIS reader
        strcpy( pReader->sz40ReaderID, sz40AudioReaderID );
        pReader->nChannelsPerSample = nChannelsPerSample;
        pReader->iDllInstanceIndex = iThisInstancesDllInstanceIndex;
        if( pdblOutSampleRate != NULL ) // caller already wants to know the sampling rate:
         { *pdblOutSampleRate = pAudioSourceInfo->dblSampleRate;
         }
        if( (pOutChunkInfo != NULL )  // caller already needs the first "chunk info"
         && (pOutChunkInfo->dwSizeOfStruct >= CHUNK_INFO_STRUCT_SIZE_V1) )
         { // (contains details about the stream, timestamp, sample rate, GPS data, etc) .
           // CAUTION: To avoid version conflicts, the CALLER(!) must already
           //          have initialized the T_ChunkInfo structure,
           //          because it belongs to 'his' address space.
           //          For example, the caller may use a SMALLER T_ChunkInfo
           //          because it was compiled with an ancient version/copy
           //          of the file c:\cbproj\SoundUtl\ChunkInfo.h .
           //          -> Use ChunkInfo_CopyFromTo() here, instead of a simple
           //             'block copy' !
           AudioBuf_GetAdjustedChunkInfoForSampleIndex( pAudioBuffer, pAudioSourceInfo,
                      pReader->dwTailIndex, // [in]  index into fltBuffer[]
                      pOutChunkInfo);       // [out] destination chunk info
           // pOutChunkInfo->dwNrOfSamplePoints = ?; // not applicable yet
           pOutChunkInfo->nChannelsPerSample = nChannelsPerSample;
         } // end if( pOutChunkInfo != NULL )

        // For DISPLAY PURPOSES, T_AudioBuffer* g_pSharedMemoryData->reader[]
        // is duplicated in T_SharedMemoryData.T_AudioSourceInfo.reader[] :
        g_pSharedMemoryData->source[iAudioSource].reader[iReaderIndex] = *pReader;

        // let all visible control panels update themselves a.s.a.p :
        SetFlagToUpdateControlPanels();
        // -> Success; return the non-negative 'handle' value
        //    which uniquely identifies the SOURCE and the READER ,
        //    compatible with GetAudioBufferByHandle() :
        return (iAudioSource<<8) | iReaderIndex;
      }
   } // end for iReaderIndex..

  return AIO_INVALID_HANDLE; // arrived here: error, "ran out of handles"

} // end AIO_OpenAudioInput()



//----------------------------------------------------------------------
__declspec( dllexport ) INT32 AIO_ReadInputSamplePoints( // returns the number of sample-POINTS, or a negative error code
        INT32 iAudioIoHandle,     // [in] audio-handle, from ..OpenAudioInput()
        float *pfltDest,          // [out] audio samples, as 32-bit FLOATING POINT numbers, grouped as "sample points"
        INT32 iNumSamplePoints,   // [in] maximum number of SAMPLE POINTS(!) to read
        INT32 nChannelsPerSample, // [in] number of channels PER SAMPLE POINT
        INT32 iTimeout_ms ,       // [in] max timeout in milliseconds, or AIO_TIMEOUT_NON_BLOCKING
        T_ChunkInfo *pOutChunkInfo, // [out,optional] chunk info, see c:\cbproj\SoundUtl\ChunkInfo.h, MAY BE NULL (!)
        INT32 *piHaveWaited_ms)     // [out,optional] "have been waiting here for XX milliseconds"
{
  long i32HeadIndex, i32TailIndex, i32UsedBufEntries, nSamplesInBuffer;
  int  iAudioSource, iReaderIndex, nChannelsPerSampleInBuffer;  // 1=mono, 2=stereo, 4=quadrophonic, etc
  int  iChannel, nChannels;  // for the copying-loop (all channels in ONE SAMPLE)
  int  iWaited_ms = 0;
  float *pfltSource;
  INT32 nSamplePointsRead = 0;    // -> return value
  T_AudioSourceInfo *pAudioSourceInfo;
  T_AudioReader *pAudioReader;
  T_AudioBuffer *pAudioBuffer = GetAudioBufferByHandle( // may return NULL if something's wrong !
                   iAudioIoHandle, // [in] handle returned by .._OpenAudioOutput() / _OpenAudioInput()
                    &iAudioSource, // [out,optional] index into T_SharedMemoryData.source[]
                    &iReaderIndex, // [out,optional] index into T_AudioBuffer.reader[]
               &pAudioSourceInfo); // [out,optional] address of audio source info
  if( piHaveWaited_ms != NULL )
   { *piHaveWaited_ms = 0;
   }
  if( pAudioBuffer==NULL || pAudioSourceInfo==NULL )
   { return AIO_INVALID_HANDLE;  // here: problem with the shared memory-mapped file
   }
  pAudioReader = &pAudioBuffer->reader[iReaderIndex];
  if( (pAudioReader->dwFlags & AUDIO_READER_OPENED) == 0 )
   { return AIO_ERROR_INPUT_NOT_OPEN;
   }
  nChannelsPerSampleInBuffer = pAudioSourceInfo->nChannelsPerSamplePoint;
  if( nChannelsPerSampleInBuffer<1 )
   { return AIO_ERROR_CANNOT_READ_YET;
   }
  nChannels = nChannelsPerSampleInBuffer;
  if( nChannelsPerSample < nChannels )
   {  nChannels = nChannelsPerSample;   // min(source,dest) for the copying-loop below
   }

  // May have to repeat the following (if the caller wants to "wait" for data)
  while(1)
   {
     i32HeadIndex = (INT32)(pAudioBuffer->dwHeadIndex & (AUDIO_BUFFER_SIZE-1) );  // here READ-ONLY
     i32TailIndex = (INT32)(pAudioReader->dwTailIndex & (AUDIO_BUFFER_SIZE-1) );
     i32UsedBufEntries = i32HeadIndex - i32TailIndex;
     if( i32UsedBufEntries < 0 )  // index wrapped around..
      {  i32UsedBufEntries += AUDIO_BUFFER_SIZE;
      }
     nSamplesInBuffer = i32UsedBufEntries / nChannelsPerSampleInBuffer;

     if( nSamplesInBuffer >= iNumSamplePoints )
      { // There are enough samples waiting for this reader: Read them and return.
        pfltSource = pAudioBuffer->fltBuffer + i32TailIndex;
        while( nSamplePointsRead < iNumSamplePoints )
         {
           for( iChannel=0; iChannel<nChannels; ++iChannel )
            { pfltDest[iChannel] = pfltSource[iChannel];
            }
           pfltDest     += nChannelsPerSample;  // .. for the CALLER'S destination buffer
           pfltSource   += nChannelsPerSampleInBuffer; // .. for the DLL's shared buffer
           i32TailIndex += nChannelsPerSampleInBuffer;
           if( i32TailIndex >= AUDIO_BUFFER_SIZE )
            { i32TailIndex = 0;
              pfltSource = pAudioBuffer->fltBuffer;
            }
           ++nSamplePointsRead;
         } // end while < nSamplePointsRead >
        if( (pOutChunkInfo!=NULL )  // caller is interested in the T_ChunkInfo :
         && (pOutChunkInfo->dwSizeOfStruct >= CHUNK_INFO_STRUCT_SIZE_V1) )
         { // (contains details about the stream, timestamp, sample rate, GPS data, etc) .
           // CAUTION: To avoid version conflicts, the CALLER(!) must already
           //          have initialized the T_ChunkInfo structure,
           //          because it belongs to 'his' address space.
           //          For example, the caller may use a SMALLER T_ChunkInfo
           //          because it was compiled with an ancient version/copy
           //          of the file c:\cbproj\SoundUtl\ChunkInfo.h .
           //          -> Use ChunkInfo_CopyFromTo() here, instead of a simple
           //             'block copy' !
           AudioBuf_GetAdjustedChunkInfoForSampleIndex( pAudioBuffer, pAudioSourceInfo,
                      pAudioReader->dwTailIndex, // [in]  index into fltBuffer[]
                      pOutChunkInfo);            // [out] destination chunk info
           pOutChunkInfo->dwNrOfSamplePoints = nSamplePointsRead; // !!
           // The caller may read LESS channels than in the buffer, thus :
           pOutChunkInfo->nChannelsPerSample = nChannelsPerSample;
         } // end if( pOutChunkInfo != NULL )
        pAudioReader->dwTailIndex = (DWORD)i32TailIndex;
        break;
      }
     else //  nSamplesInBuffer < iNumSamplePoints :
      { // There are LESS SAMPLE POINTS in the buffer than the caller wants to read.
        // This is quite normal, the reader calls the DLL 'as often as he can',
        // and we'll often need to "wait somewhere" until samples are available.
        // Does the caller want to wait HERE ?
      }

     if( iTimeout_ms <= 0 )
      { // non-blocking call (or max waiting-time exceeded) ? return with "no samples" !
        // No-No: return AIO_ERROR_CANNOT_WRITE_YET;  // this is NOT an error !
        break;
      }
     else // wait a few dozen milliseconds, then try again
      { Sleep(50);
        iTimeout_ms -= 50;
        iWaited_ms  += 50;
      } // end if < not enough "free" buffer space yet >
   } // end while < waiting until there's enough space in the output-buffer >

  if( piHaveWaited_ms != NULL )
   { *piHaveWaited_ms = iWaited_ms;
   }

  return nSamplePointsRead; // -> number of sample POINTS (with N channels PER POINT) actually read
} // end AIO_ReadInputSamplePoints()


//----------------------------------------------------------------------
__declspec( dllexport )  void AIO_CloseAudioInput(
     INT32 iAudioIoHandle )   // [in] audio-reader-handle, see ..OpenAudioInput()
  // Polite applications call this function when they shut down.
{
  BOOL fModified = FALSE;
  int iAudioSource, iReaderIndex;
  T_AudioSourceInfo *pAudioSourceInfo;
  T_AudioBuffer *pAudioBuffer = GetAudioBufferByHandle( // may return NULL if something's wrong !
                    iAudioIoHandle, // [in] handle returned by .._OpenAudioOutput() / _OpenAudioInput()
                    &iAudioSource, // [out,optional] index into T_SharedMemoryData.source[]
                    &iReaderIndex, // [out,optional] index into T_AudioBuffer.reader[]
                    &pAudioSourceInfo); // [out,optional] address of audio source info

  if( pAudioBuffer!=NULL  )
   { fModified = ( (pAudioBuffer->reader[iReaderIndex].dwFlags & AUDIO_READER_OPENED) != 0);
     // clear flag AUDIO_READER_OPENED to declare this instance of an 'audio reader' as CLOSED
     pAudioBuffer->reader[iReaderIndex].dwFlags &= ~AUDIO_READER_OPENED;
     // For DISPLAY PURPOSES, T_AudioBuffer.reader[]
     // is duplicated in T_SharedMemoryData.T_AudioSourceInfo.reader[] :
     if( g_pSharedMemoryData != NULL )
      { g_pSharedMemoryData->source[iAudioSource].reader[iReaderIndex]
           = pAudioBuffer->reader[iReaderIndex];
      }
     if(fModified)
      { SetFlagToUpdateControlPanels(); // let all visible control panels update themselves
      }
   }
} // end AIO_CloseAudioInput()


//----------------------------------------------------------------------
__declspec( dllexport ) INT32 AIO_ShowAboutBox(
     HWND hwndParent )    // [in] handle to the parent window
  // Since 2011-06-14, also callable by SpecLab : Select
  //   "Options".."Audio Settings", set the input device to "Audio I/O",
  //   click on "Ctrl", select "in_AudioIO.dll" (as Input- or Output-DLL),
  //   then click the "?" at the right side of the DLL-name-field .
{
   WA_About(hwndParent);
   return AIO_NO_ERROR;

     // CAUTION: The "About"-box, implemented in the DLL, is usually
     //    a normal 'blocking' windoze message box [ MessageBox() ] .
     //    But when this 'about' box was opened from a menu in the DLL's
     //    non-modal control panel, the 'about' box remained open while
     //    the application unloaded the DLL  ->   BANG !
     //    There is no easy way to close such a dialog programmatically
     //    (automatically) when unloading the DLL.


} // end AIO_ShowAboutBox()


//----------------------------------------------------------------------
// Implementation of the 'control window' .
//  Based on DL4YHF's "Sound Input Utility" (c:\cbproj\SoundUtl\SndInMain.cpp).
//  Uses NO 'Visual Component Library', no 'MFC', no fancy library,
//  only plain Win32 API programming to keep the size of the DLL small
//  and to avoid the need for additional, bulky, libraries (like the VCL) .
//----------------------------------------------------------------------


//---- Internal variables for the "user interface", unfortunately GLOBAL -----
int  MainWnd_iUpdating= 0; // set while some user controls must not react on changes
int  MainWnd_iConnectMeterToOutput=0;
int  MainWnd_iPrevDisplayMode = -1;
HWND MainWnd_hDisplayModeCombo;
HWND MainWnd_hEditNcoFreq;
HWND MainWnd_hEditTestFreq;  // "test frequency" (to superimpose a TEST TONE to the output)
#define C_HORZ_EDITS   2     // count of edit fields, horizontally aligned

 // Menu items for "system menu" alias "window menu" alias "control menu"
 // (the little popup which opens when you click the symbol in the upper left corner)
#define IDM_SHOW_ABOUT_BOX   1
#define IDM_CLEAR_UNUSED     3
#define IDM_SHOW_CONNECTIONS 4
#define IDM_SHOW_CHUNK_INFO  5
#define IDM_SHOW_WAVEFORM    6
#define IDM_SHOW_SPECTRUM    7
#define IDM_SHOW_EVT_HISTORY  8
#define IDM_CLEAR_EVT_HISTORY 9

#if( SWI_ASIO_SUPPORTED ) // defined in SWITCHES.H (project-specific options)
# define IDM_ASIO_ENABLED    10
# define IDM_ASIO_SEL_STREAM 11
# define IDM_ASIO_REGISTER   12
# define IDM_ASIO_UNREGISTER 13
#endif // SWI_ASIO_SUPPORTED ?

#define ID_COMBO_DISP_MODE  20


#define MY_BACKGROUND_COLOR COLOR_BTNFACE

RECT MainWnd_rParamDisplay;  // client region for the parameter display

//---------------------------------------------------------------------------
void UpdateMyMenu(void)
{
  HWND  hwndMain = g_hwndControlPanel;
  HMENU hMenu = GetSystemMenu( hwndMain, FALSE/*revert*/ );
  T_ControlPanelInfo *pCPInfo = GetControlPanelInfoForGUI(); // ptr always valid

  CheckMenuItemByIDAndCondition( hMenu, IDM_SHOW_CONNECTIONS, //condition:
            pCPInfo->iDisplayMode == DISPLAY_MODE_SHOW_CONNECTIONS );
  CheckMenuItemByIDAndCondition( hMenu, IDM_SHOW_CHUNK_INFO,
            pCPInfo->iDisplayMode == DISPLAY_MODE_SHOW_CHUNK_INFO  );
  CheckMenuItemByIDAndCondition( hMenu, IDM_SHOW_WAVEFORM  ,
            pCPInfo->iDisplayMode == DISPLAY_MODE_SHOW_WAVEFORM    );
  CheckMenuItemByIDAndCondition( hMenu, IDM_SHOW_SPECTRUM  ,
            pCPInfo->iDisplayMode == DISPLAY_MODE_SHOW_SPECTRUM    );
  CheckMenuItemByIDAndCondition( hMenu, IDM_SHOW_EVT_HISTORY  ,
            pCPInfo->iDisplayMode == DISPLAY_MODE_SHOW_EVT_HISTORY );

  // PrintIntoMenuItem( // first 3 params like CheckMenuItem :
  //               hMenu, IDM_CONNECT_METER_TO, MF_BYCOMMAND,
  //                   // parameter 4,5 and following like "wsprintf" :
  //               "Meter connected to %s",
  //               MainWnd_iConnectMeterToOutput?"output":"input" );
}





//**************************************************************************
// Handle Menu- and other events
//**************************************************************************



void OnNcoEdit(WORD wNotifyCode) // something happened with the NCO frequency edit field
{ // wNotifyCode has been passed from the edit control, through a WM_COMMAND message.
 char sz80Temp[81];  char *cp;

 if(MainWnd_iUpdating==0)
  {
   switch(wNotifyCode)
    {
     case EN_CHANGE: // The user has modified text in an edit control
         GetWindowText(MainWnd_hEditNcoFreq,sz80Temp,16);
         cp = sz80Temp;
         (void)cp;
#if(0)
         SOUND_fltCenterFrequency = strtod( cp, &cp/*char **endptr*/  );
         DecimatingAudioBuffer.SetNcoFrequency(SOUND_fltCenterFrequency);
         SOUND_fltCenterFrequency = DecimatingAudioBuffer.GetNcoFrequency();
         SndIn_fltOldCenterFrequency = SOUND_fltCenterFrequency;
#endif
         break;
    } // end switch(wNotifyCode)
  } // end if <not "updating">
} // end OnNcoEdit()

void OnTestFreqEdit(WORD wNotifyCode) // something happened with the TEST frequency edit field
{
 char sz80Temp[81];  char *cp;
 float fltNewFreq;
 T_AudioSourceInfo *pSourceInfo = GetAudioSourceInfoForGUI(); // NEVER returns NULL - pointer ALWAYS valid

 if(MainWnd_iUpdating==0)
  {
   switch(wNotifyCode)
    {
     case EN_CHANGE: // The user has modified text in an edit control
        GetWindowText(MainWnd_hEditTestFreq,sz80Temp,16);
        cp = sz80Temp;
        fltNewFreq = strtod( cp, &cp/*char **endptr*/  );
        pSourceInfo->fltTestGeneratorFrequency = fltNewFreq;
        if( pSourceInfo->fltTestGeneratorFrequency > 0 )
         {  pSourceInfo->iAudioPlayerMode = AIO_PLAYER_MODE_TONE;
         }
        break;
    } // end switch(wNotifyCode)
  } // end if <not "updating">
} // end OnTestFreqEdit()


void OnWmTimer(HWND hWnd, WORD wTimerID ) // called on WM_TIMER
{
  int   iChn,nChannels;
  float fltPeakValue;
  BOOL  fParams4ControlPanelModified = FALSE;
  T_AudioSourceInfo *pSourceInfo = GetAudioSourceInfoForGUI(); // NEVER returns NULL - pointer ALWAYS valid
  T_ControlPanelInfo *pCPInfo = GetControlPanelInfoForGUI(); // always valid
  char sz80[84];

  (void)wTimerID;

  // Check if the control panel should be redrawn...
  nChannels = pSourceInfo->nChannelsPerSamplePoint;
  if( nChannels>AUDIO_IO_MAX_CHANNELS )
   {  nChannels=AUDIO_IO_MAX_CHANNELS;
   }
  for(iChn=0; iChn<nChannels; ++iChn)
   { // there may be one or two audio channels..
     fltPeakValue = pSourceInfo->fltAudioPeakValue[iChn];
     if( sfltOldPeakValue[iChn] != fltPeakValue )
      { sfltOldPeakValue[iChn] = fltPeakValue;
        fParams4ControlPanelModified = TRUE;
      }
   }
  if( pSourceInfo->iCalculatedBufferUsage_0to255 != pSourceInfo->iDisplayedBufferUsage_0to255 )
   {
     pSourceInfo->iDisplayedBufferUsage_0to255 = pSourceInfo->iCalculatedBufferUsage_0to255;
     fParams4ControlPanelModified = TRUE;
   }
  if( g_pSharedMemoryData != NULL )
   { if( g_pSharedMemoryData->iGuiUpdateCounter != iThisInstancesGuiUpdateCounter )
      { // Someone suggested to update the display via SetFlagToUpdateControlPanels():
        iThisInstancesGuiUpdateCounter = g_pSharedMemoryData->iGuiUpdateCounter;
        fParams4ControlPanelModified = TRUE;
      }
   }
  if( MainWnd_iPrevDisplayMode != pCPInfo->iDisplayMode )
   { MainWnd_iPrevDisplayMode = pCPInfo->iDisplayMode;
     fParams4ControlPanelModified = TRUE;
     // Change the 'caption' (window title) ?
     switch( MainWnd_iPrevDisplayMode )
      {
        case DISPLAY_MODE_SHOW_WAVEFORM:
           sprintf(sz80, "Waveform: %s", pSourceInfo->sz40AudioStreamID );
           SetWindowText( hWnd, sz80 );
           break;
        case DISPLAY_MODE_SHOW_SPECTRUM:
           sprintf(sz80, "Spectrum: %s", pSourceInfo->sz40AudioStreamID );
           SetWindowText( hWnd, sz80 );
           break;
        case DISPLAY_MODE_SHOW_EVT_HISTORY:
           strcpy(sz80, "Audio I/O Event History" );
           SetWindowText( hWnd, sz80 );
           break;
        default:
           SetWindowText( hWnd, APPL_szWindowCaption );
           break;
      }
   }

  if( fParams4ControlPanelModified )
   {
     // update 'parameter display' :
     // > RedrawWindow updates the specified rectangle or region in a window's client area. (Win32 API)
     RedrawWindow( hWnd, NULL, NULL, RDW_INVALIDATE );
   } // end if (MainWnd_fRunning)

} // end OnWmTimer()


//**************************************************************************
// Red tape for the windows 'GUI'  (formerly a windows application)
//**************************************************************************

 // dimensions of a few 'visible controls', found by trial-and-error
 // ex: #define C_COMBO_HEIGHT 20 /* combobox for display mode selection (height without the listbox) */
 #define C_EDIT_HEIGHT    20
 #define C_STATIC_TEXT_WIDTH 40 /* enough for strings like "fc/Hz" */
 #define C_BUTTON_HEIGHT  18
 #define C_VERT_SEPARATION 4
 #define C_HORZ_SEPARATION 4


//---------------------------------------------------------------------------
void ShowWaveform( HDC hDC, int x1, int y1, int w, int h )
{ T_AudioSourceInfo *pSourceInfo = GetAudioSourceInfoForGUI(); // NEVER returns NULL - pointer ALWAYS valid
  T_AudioBuffer   * pAudioBuffer = GetAudioBufferForGUI(); // may return NULL (!)
  T_ChunkInfo     * pChunkInfo   = AudioBuf_GetLatestChunkInfo( pAudioBuffer );
  long i32SampleIndex, i32TrigIndex;
  int  i,x,y,nChannels;
  float f1,f2,delta,best_delta,range,factor,offset;
  BOOL ok = FALSE;
  // Ugly windows GDI stuff:
  HPEN hPen, hOldPen;


  if( pAudioBuffer!=NULL && pChunkInfo!=NULL && pSourceInfo!=NULL)
   { nChannels = pSourceInfo->nChannelsPerSamplePoint;
     range     = pSourceInfo->dblSampleValueRange;
     if( (range>0.0) && (nChannels>0) && (pAudioBuffer->fltBuffer!=NULL) )
      { ok = TRUE;
      }
   }


  if( ! ok )
   { SimpleTextOut(hDC, "Signal not available", 0, y1 );
   }
  else
   {
     // Erase the curve area, using a bit block transfer (fast!)
     BitBlt( hDC, x1, y1, w, h,
           NULL,       // source bitmap device context
           0, 0,       // start x,y on source bitmap
           WHITENESS); // blit method

     // Draw a very crude grid (vertical with 10 dB / div) :
     hPen = CreatePen(PS_SOLID, 1, RGB(0xCC,0xCC,0xCC));
     hOldPen = SelectObject(hDC, hPen);
     for(i=0; i<=10; ++i )  // amplitude grid ...
      { y = (h-1) * i / 10;
        MoveToEx(hDC, x1,   y, NULL);
        LineTo(  hDC, x1+w, y);
      }
     for(i=0; i<=10; ++i )  // frequency grid ... 10 divisions, 11 lines
      { x = (w-1) * i / 10;
        MoveToEx(hDC, x, y1, NULL);
        LineTo(  hDC, x, y1+h );
      }
     SelectObject(hDC, hOldPen);
     DeleteObject(hPen);      // <<< never forget this (when using the ugly windoze GDI) to avoid leaks
     hPen = CreatePen(PS_SOLID, 1, RGB(0,0,0) );
     SelectObject(hDC, hPen);

     // Always show the MOST RECENT samples, and one sample per pixel.
     // (step a bit further 'back in time' to have some headroom for the trigger)
     i32SampleIndex = (long)pAudioBuffer->dwHeadIndex - (long)(2*w*nChannels);
     while( i32SampleIndex < 0 )
      {  i32SampleIndex += AUDIO_BUFFER_SIZE;
      }
     // Beginning at i32SampleIndex, for a maximum of <w> samples,
     // look for the next rising edge / zero crossing :
     f1 = pAudioBuffer->fltBuffer[i32SampleIndex];
     i32TrigIndex = i32SampleIndex;
     best_delta = 0.0;
     for(x=0; x<w; ++x)
      { f2 = pAudioBuffer->fltBuffer[i32SampleIndex];
        if( f1<=0.0 && f2>0.0 )
         { delta = f2 - f1;
           if( delta > best_delta )
            { i32TrigIndex = i32SampleIndex;
              best_delta = delta;
            }
         }
        f1 = f2;
        i32SampleIndex += nChannels;
        if( i32SampleIndex >= AUDIO_BUFFER_SIZE )
         {  i32SampleIndex = 0;
         }
      }
     // Beginning at i32TrigIndex, draw the waveform
     i32SampleIndex = i32TrigIndex;
     factor = -0.5 * (float)h / range;
     offset = 0.5 * h;
     for(x=0; x<w; ++x)
      {
        f1 = pAudioBuffer->fltBuffer[i32SampleIndex] * factor + offset;
        i32SampleIndex += nChannels;
        if( i32SampleIndex >= AUDIO_BUFFER_SIZE )
         {  i32SampleIndex = 0;
         }
        if(f1<0.0) f1=0.0;
        if(f1>h)   f1=h;
        y = y1 + (int)f1;
        if( x==0 ) MoveToEx(hDC, x+x1, y, NULL);
        else       LineTo( hDC, x+x1, y);
      }

     SelectObject(hDC, hOldPen);
     DeleteObject(hPen);      // <<< never forget this (when using the ugly windoze GDI) to avoid leaks

   }
} // end ShowWaveform()

//---------------------------------------------------------------------------
void ShowSpectrum( HDC hDC, int x1, int y1, int w, int h )
{
  T_AudioSourceInfo *pSourceInfo = GetAudioSourceInfoForGUI(); // NEVER returns NULL - pointer ALWAYS valid
  T_AudioBuffer   * pAudioBuffer = GetAudioBufferForGUI(); // may return NULL (!)
  T_ChunkInfo     * pChunkInfo   = AudioBuf_GetLatestChunkInfo( pAudioBuffer );
  #define L_DISPLAY_FFT_SIZE 1024 // the sample-buffer will ALWAYS be larger than this !
  float re[L_DISPLAY_FFT_SIZE], im[L_DISPLAY_FFT_SIZE]; // working area for the FFT
  long i32SampleIndex;
  int  i,x,y,nChannels;
  double d,range,factor,min_dB;
  BOOL ok = FALSE;
  // Ugly windows GDI stuff:
  HPEN hPen, hOldPen;

  if( pAudioBuffer!=NULL && pChunkInfo!=NULL && pSourceInfo!=NULL)
   { nChannels = pSourceInfo->nChannelsPerSamplePoint;
     range     = pSourceInfo->dblSampleValueRange;
     if( (range>0.0) && (nChannels>0) && (pAudioBuffer->fltBuffer!=NULL) )
      { ok = TRUE;
      }
   }


  if( ! ok )
   { SimpleTextOut(hDC, "Signal not available", 0, y1 );
   }
  else
   {
     // Erase the curve area, using a bit block transfer (fast!)
     BitBlt( hDC, x1, y1, w, h,
           NULL,       // source bitmap device context
           0, 0,       // start x,y on source bitmap
           WHITENESS); // blit method

     // Draw a very crude grid (vertical with 10 dB / div) :
     min_dB = -120.0;    // minimum displayed "dB over full scale" (max always 0 dBfs)
     hPen = CreatePen(PS_SOLID, 1, RGB(240,240,240) );
     hOldPen = SelectObject(hDC, hPen);
     for(i=0; i>=(int)min_dB; i-=10/*dB*/ )  // amplitude grid ...
      {
        y = (int)( (h-1) * i / (int)min_dB );  // similar amplitude scaling as for the logarithmized FFT
        if( y>=0 && y<h )
         { MoveToEx(hDC, x1,   y, NULL);
           LineTo(  hDC, x1+w, y);
         }
      }
     for(i=0; i<=10; ++i )  // frequency grid ... 10 divisions, 11 lines
      {
        x = (int)( (w-1) * i / 10 );
        MoveToEx(hDC, x, y1, NULL);
        LineTo(  hDC, x, y1+h );
      }
     SelectObject(hDC, hOldPen);
     DeleteObject(hPen);      // <<< never forget this (when using the ugly windoze GDI) to avoid leaks
     hPen = CreatePen(PS_SOLID, 1, RGB(0,0,0) );
     SelectObject(hDC, hPen);


     // Always show the MOST RECENT samples, and one sample per pixel.
     // Step a bit 'back in time' to collect enough samples for an FFT
     //  with 2^N points:
     i32SampleIndex = (long)pAudioBuffer->dwHeadIndex - nChannels*L_DISPLAY_FFT_SIZE;
     while( i32SampleIndex < 0 )
      {  i32SampleIndex += AUDIO_BUFFER_SIZE;
      }
     for(i=0; i<L_DISPLAY_FFT_SIZE; ++i)
      { re[i] = pAudioBuffer->fltBuffer[i32SampleIndex];
        im[i] = 0.0;
        i32SampleIndex += nChannels;
        if( i32SampleIndex >= AUDIO_BUFFER_SIZE )
         {  i32SampleIndex = 0;
         }
      }
#if(0)  // TEST ?   fill the sample block with a 'clean sinewave' of know amplitude
     for(i=0; i<L_DISPLAY_FFT_SIZE; ++i)
      { re[i] = pow(10.0, -110.0/*dB*/ / 20.0) * range * sin( 0.1*(double)i );
      }
#endif
     MultiplyHanningWindow( re, L_DISPLAY_FFT_SIZE );
     CalcRealFft(
       L_DISPLAY_FFT_SIZE, // number of points in the time domain (input), 2^N
       re,  // the real input signal, also used as result (real part)
       im); // output, imaginary part
     //  Note: The output signals run from  0...iNrOfPoints/2 !
     //        A "1024 point REAL FFT" produces 513(!) POINTS in pfltRe[]
     //                                and 513(!) POINTS in pfltIm[] .
     // A pure input sin wave ... Asin(wt)... will produce an fft output
     //   peak of (N*A/4)^2  where N is FFT_SIZE (for HANN window, arvg=0.5).
     // Get the conversion factor to normalize amplitude spectrum to [0..1] :
     factor = 4.0 / ( (double)L_DISPLAY_FFT_SIZE * range/*A*/ );
     // Test this with a "clean" sine wave of maximal amplitude !
     for(i=0; i<L_DISPLAY_FFT_SIZE/2; ++i)
      {
        d = re[i]*re[i] + im[i]*im[i];
#if(0)
        d = pow( L_DISPLAY_FFT_SIZE * range / 4.0,  2.0);  // TEST -> 0 dBfs ?
#endif
        d = sqrt( d ) * factor;
        d = 20*log10( d );  // -> "decibel over full scale" (thus always negative)
        x = x1 + (2 * i * w) / L_DISPLAY_FFT_SIZE;
        y = (int)( (double)(h-1) * d / min_dB );  // -> display range 0 dBfs (max) ... -120 dBfs (min)
        if( y<0 ) y=0;
        if( y>h ) y=h;
        y = y1 + y;
        if( i==0 ) MoveToEx(hDC, x, y, NULL);
        else       LineTo( hDC, x, y);
      }
     SelectObject(hDC, hOldPen);
     DeleteObject(hPen);      // <<< never forget this (when using the ugly windoze GDI) to avoid leaks
   }
} // end ShowSpectrum()


//---------------------------------------------------------------------------
LONG OnWmPaint(HWND hWnd)
{ // reaction to WM_PAINT message
  PAINTSTRUCT ps;
  SIZE text_size;
  HDC hDC;
  HFONT hOldFont, hNormalFont, hBoldFont;
  RECT rect;
  BOOL fShowBargraphs;
  int  i,x,y,yp,ye,x2,y2,iChn,nChannels, curve_area_width, curve_area_height;
  int  iAudioSource, iReaderIndex, nReaders;
  int  iUnixDayNumber,iYear,iMonth,iDay,iHour,iMin,iSec;
  double dbl;
  char sz80[81], *cp;
  float fltPeakValue;
  T_AudioSourceInfo *pSourceInfo = GetAudioSourceInfoForGUI(); // NEVER returns NULL - pointer ALWAYS valid
  T_ControlPanelInfo *pCPInfo = GetControlPanelInfoForGUI(); // ptr always valid
  T_AudioBuffer * pAudioBuffer = NULL;
  T_AudioReader * pReader;
  T_ChunkInfo   * pChunkInfo;


  hDC = BeginPaint(hWnd,&ps);
  if( g_pSharedMemoryData != NULL )
   {  // No "lazy return" within this block !

     GetClientRect(hWnd,&rect);
     SetBkColor(hDC, GetSysColor(MY_BACKGROUND_COLOR) );
     y = 4;  // painting begins at the top, at this pixel position


     // DONT FORGET TO 'UNSELECT' the custom fonts later and DELETE the font objects:
     hNormalFont = CreateFont(
                 14/*nHeight*/,      0/*best width*/,
                 0/*nEscapement*/,  0/*nOrientation*/,
                 FW_NORMAL/*fnWeight*/,
                 FALSE/*fdwItalic*/,    FALSE/*fdwUnderline*/,
                 FALSE/*fdwStrikeOut*/,	ANSI_CHARSET/*fdwCharSet*/,
                 OUT_DEFAULT_PRECIS , // fdwOutputPrecision
                 CLIP_DEFAULT_PRECIS, // fdwClipPrecision
                 DEFAULT_QUALITY,     // fdwQuality
                 DEFAULT_PITCH | FF_DONTCARE, // fdwPitchAndFamily
                 "Arial");           // pointer to typeface name string
     hBoldFont = CreateFont(
                 14/*nHeight*/,      0/*best width*/,
                 0/*nEscapement*/,  0/*nOrientation*/,
                 FW_BOLD/*fnWeight*/,
                 FALSE/*fdwItalic*/,    FALSE/*fdwUnderline*/,
                 FALSE/*fdwStrikeOut*/,	ANSI_CHARSET/*fdwCharSet*/,
                 OUT_DEFAULT_PRECIS , // fdwOutputPrecision
                 CLIP_DEFAULT_PRECIS, // fdwClipPrecision
                 DEFAULT_QUALITY,     // fdwQuality
                 DEFAULT_PITCH | FF_DONTCARE, // fdwPitchAndFamily
                 "Arial");           // pointer to typeface name string
     hOldFont = (HFONT)SelectObject(hDC, hNormalFont );
     GetTextExtentPoint32(hDC, "Q", 1, &text_size); // -> text_size.cx=8, cy=14
     g_TextHeight_px = text_size.cy; // to simplify passing this to subroutines called below
     SetBkMode(hDC, OPAQUE );

     // Get the width of the usable 'curve area' :
     nChannels = pSourceInfo->nChannelsPerSamplePoint;
     if( nChannels>AUDIO_IO_MAX_CHANNELS )
      {  nChannels=AUDIO_IO_MAX_CHANNELS;
      }
     switch( pCPInfo->iDisplayMode )
      { // Special display modes WITHOUT the vertical bargraphs ?
        case DISPLAY_MODE_SHOW_EVT_HISTORY:
           curve_area_width  = MainWnd_rParamDisplay.right;
           curve_area_height = MainWnd_rParamDisplay.bottom;
           rect.bottom= MainWnd_rParamDisplay.bottom;
           rect.right = MainWnd_rParamDisplay.right;
           fShowBargraphs = FALSE;
           break;
        default: // any display-mode WITH the vertical bargraphs visible:
           curve_area_width  = MainWnd_rParamDisplay.right - 18*(nChannels+1) - 2;
           curve_area_height = MainWnd_rParamDisplay.bottom- MainWnd_rParamDisplay.top - 2;
           rect.bottom= MainWnd_rParamDisplay.bottom;
           rect.right = MainWnd_rParamDisplay.right - 16*pSourceInfo->nChannelsPerSamplePoint;
           fShowBargraphs = TRUE;
           break;
       }

     SetTextAlign(hDC,TA_LEFT|TA_TOP);
#if(0) // TEST ONLY:  is the "parameter display area" ok ?
      FillRect(hDC, &rect, (HBRUSH)COLOR_ACTIVECAPTION );
#else  // normal operation
      FillRect(hDC, &rect, (HBRUSH)(MY_BACKGROUND_COLOR+1) ); // ex: COLOR_WINDOW
      // why add one to the colour value ? just another windoze-annoyance..
#endif

     ye = MainWnd_rParamDisplay.bottom - text_size.cy;
     //  'ye' is the bottom / "endstop" for the following displays
     switch( pCPInfo->iDisplayMode )
      { case DISPLAY_MODE_SHOW_CONNECTIONS:
           // Show the first few ACTIVE SOURCES and DESTINATIONS :
           for( iAudioSource=0; (iAudioSource<AUDIO_MAX_SOURCES) && (y<ye) ; ++iAudioSource)
            { pAudioBuffer = g_pAudioBuffers[iAudioSource];
              pSourceInfo = &g_pSharedMemoryData->source[iAudioSource];
              nReaders = 0;
              for(iReaderIndex=0; iReaderIndex<AUDIO_MAX_READERS; ++iReaderIndex)
               { pReader = &pSourceInfo->reader[iReaderIndex];
                 if( pReader->dwFlags & AUDIO_READER_OPENED )
                  { ++nReaders;
                  }
               }
              if(   (pAudioBuffer!=NULL) || (nReaders>0)
                 || (pSourceInfo->dwWriterFlags&AUDIO_WRITER_OPENED) )
               {
                 sprintf(sz80,"Stream: %s, %.1f Hz", pSourceInfo->sz40AudioStreamID, pSourceInfo->dblSampleRate ) ;
                 i = 0; // is this audio stream opened (by its 'writer') ?
                 if ( pSourceInfo->dwWriterFlags & AUDIO_WRITER_OPENED )
                  { i = 1;
                  }
                 if( pAudioBuffer != NULL )
                  { if( pAudioBuffer->iWriterDllInstanceIndex == iThisInstancesDllInstanceIndex)
                     { i |= 2;  // "THIS instance's stream"
                     }
                  } // end if( pAudioBuffer != NULL )
                 if((i&1)==0) strcat(sz80," (closed)");
                 SelectObject(hDC, (iAudioSource==iSelectedAudioSource) ? hBoldFont : hNormalFont );
                 SimpleTextOut(hDC, sz80, MainWnd_rParamDisplay.left, y );
                 y += text_size.cy;
                 yp = y;
                 for(iReaderIndex=0; (iReaderIndex<AUDIO_MAX_READERS) && (y<ye); ++iReaderIndex)
                  { pReader = &pSourceInfo->reader[iReaderIndex];
                    if(   (pReader->dwFlags & AUDIO_READER_OPENED)
                       && (pReader->sz40ReaderID[0] > 0 )  )
                     { sprintf(sz80,"R%d: %s",(int)iReaderIndex,(char*)pReader->sz40ReaderID);
                       x  = MainWnd_rParamDisplay.left+4;
                       x2 = MainWnd_rParamDisplay.left+10;
                       y2 = y+text_size.cy/2;
                       MoveToEx( hDC, x, yp, NULL );
                       LineTo(   hDC, x, y2 );
                       LineTo(   hDC, x2,y2 );
                       yp = y2;
                       TextOut(hDC,x2+2, y, sz80, strlen(sz80) );
                       y+=text_size.cy;
                     }
                  } // end for < all readers of this audio-source >
               } // end  if( (pAudioBuffer!=NULL) || (pSourceInfo->dwWriterFlags ..
            } // end for( iAudioSource .. )

           // Was there 'any sign of life from Winamp' yet ?
           y = MainWnd_rParamDisplay.bottom - text_size.cy;
           if( g_pSharedMemoryData->sz255WinampLastEvent[0] != 0 )
            { SimpleTextOut(hDC,  g_pSharedMemoryData->sz255WinampLastEvent,
                                        MainWnd_rParamDisplay.left, y );
            }
           break; // end case DISPLAY_MODE_SHOW_CONNECTIONS

        case DISPLAY_MODE_SHOW_CHUNK_INFO:
           // Show the contents of the most recent AUDIO CHUNK INFO :
           pSourceInfo = GetAudioSourceInfoForGUI(); // NEVER returns NULL - pointer ALWAYS valid
           pAudioBuffer = GetAudioBufferForGUI();    // may return NULL (!)
           if( pAudioBuffer==NULL )
            { SimpleTextOut(hDC,"No source selected", MainWnd_rParamDisplay.left, y );
              y+=text_size.cy;
            }
           else
            { sprintf( sz80, "Chunk-info from stream '%s':", pSourceInfo->sz40AudioStreamID );
              SimpleTextOut(hDC, sz80, MainWnd_rParamDisplay.left, y );
              y+=text_size.cy;
              pChunkInfo = AudioBuf_GetLatestChunkInfo( pAudioBuffer );
              if( pChunkInfo==NULL )
               { SimpleTextOut(hDC,"No info available", MainWnd_rParamDisplay.left, y );
                 y+=text_size.cy;
               }
              else // Show the contents of the latest 'chunk info' in text form:
               { sprintf( sz80, " %ld S/chunk, total = %ld",
                        (long)pChunkInfo->dwNrOfSamplePoints,
                        (long)pChunkInfo->i64TotalSampleCounter );
                 SimpleTextOut(hDC, sz80, MainWnd_rParamDisplay.left, y );
                 y+=text_size.cy;
                 if( (pChunkInfo->dwValidityFlags & CHUNK_INFO_SAMPLE_RATE_VALID) == 0 )
                  { cp = "invalid";
                  }
                 else if( (pChunkInfo->dwValidityFlags & CHUNK_INFO_SAMPLE_RATE_CALIBRATED) == 0 )
                  { cp = "uncalibrated";
                  }
                 else
                  { cp = "calibrated";
                  }
                 sprintf( sz80, " %.1lf S/sec, %s",
                        (double)pChunkInfo->dblPrecSamplingRate, cp );
                 SimpleTextOut(hDC, sz80, MainWnd_rParamDisplay.left, y );
                 y+=text_size.cy;
                 if( pChunkInfo->dwValidityFlags & CHUNK_INFO_TIMESTAMPS_VALID )
                  { // Only display the timestamps if the writer flagged them as "valid" :
                    iUnixDayNumber = (int)(pChunkInfo->ldblUnixDateAndTime / 86400.0/*seconds per day*/);
                    UnixDaysToYearMonthDay( iUnixDayNumber, &iYear, &iMonth, &iDay );
                    dbl = fmod(pChunkInfo->ldblUnixDateAndTime, 86400.0 ); // -> seconds of day
                    dbl = dbl / 3600.0;   // -> hours of day, fractional
                    iHour = (int)dbl; dbl=60.0*(dbl-iHour);  // -> hours of day
                    iMin  = (int)dbl; dbl=60.0*(dbl-iMin);   // -> minutes of hour
                    iSec  = (int)dbl; dbl -= iSec;           // -> seconds of minute
                    // remainder in 'dbl' : "fractions of a second"
                    sprintf( sz80, "TS1: %04d-%02d-%02d %02d:%02d:%02d.%02d",
                             iYear,iMonth,iDay,iHour,iMin,iSec,(int)(100.0*dbl) );
                    SimpleTextOut(hDC, sz80, MainWnd_rParamDisplay.left, y );
                    y+=text_size.cy;
                    sprintf( sz80, "TS2: %8.3lf",
                             (double)pChunkInfo->rtRecTime );
                    SimpleTextOut(hDC, sz80, MainWnd_rParamDisplay.left, y );
                    y+=text_size.cy;
                  }
               }
              // Show the last error code (hopefully "no error"),
              y = MainWnd_rParamDisplay.bottom - text_size.cy;
              if( pAudioBuffer->iLastErrorCode!= 0 )
               { strcpy(sz80, AIO_ErrorCodeToString(pAudioBuffer->iLastErrorCode) );
                 SimpleTextOut(hDC,  sz80, MainWnd_rParamDisplay.left, y );
               }
            } // end else < pAudioBuffer!=NULL >
           break; // end case DISPLAY_MODE_SHOW_CHUNK_INFO

        case DISPLAY_MODE_SHOW_WAVEFORM:
           ShowWaveform( hDC, 0/*x1*/, MainWnd_rParamDisplay.top/*y1*/,
                                curve_area_width, curve_area_height );
           break;

        case DISPLAY_MODE_SHOW_SPECTRUM:
           ShowSpectrum( hDC, 0/*x1*/, MainWnd_rParamDisplay.top/*y1*/,
                                curve_area_width, curve_area_height );
           break;

        case DISPLAY_MODE_SHOW_EVT_HISTORY:
           ShowEventHistory( hDC, 0/*x1*/, MainWnd_rParamDisplay.top/*y1*/,
                                curve_area_width, curve_area_height );
           break;


        default:  // implement 'other display modes' above ;o)
           break;

      } // end  switch( MainWnd_iDisplayMode )

     if( fShowBargraphs )
      { // Draw vertical bar(s) for the audio volume indicator(s)
        pSourceInfo = GetAudioSourceInfoForGUI(); // NEVER returns NULL - pointer ALWAYS valid
        nChannels = pSourceInfo->nChannelsPerSamplePoint;
        if( nChannels>AUDIO_IO_MAX_CHANNELS )
         {  nChannels=AUDIO_IO_MAX_CHANNELS;
         }
        for(iChn=0; iChn<nChannels; ++iChn)
         { // there may be one or two audio channels..
          fltPeakValue = pSourceInfo->fltAudioPeakValue[iChn]; // already normalized
          x = MainWnd_rParamDisplay.right - 16*nChannels + 16*iChn;
          // Draw the "peak value indicator" at the right window edge...
          DrawVerticalIndicatorBar(  hDC,        // device context for output
             x, MainWnd_rParamDisplay.top,
             x+17,MainWnd_rParamDisplay.bottom,   // drawing area: left,top,right,bottom
             (fltPeakValue>0.7)?   // Color of the indicator (RGB-mix)
                            RGB(0xFF,0x00,0x00) : RGB(0x00,0xFF,0x00),
             RGB(0xFF,0xFF,0xFF),  // Color of the background (RGB-mix)
             100.0*fltPeakValue,   // percentage value (0..100)
             (char*)( (nChannels>1)?    // caption, here: L,R channel
                      ((iChn==0)?"vl":"vr") : "vo" )  );
         } // end for <all audio channels>

        // Draw another vertical bar which shows the current buffer usage:
        x = MainWnd_rParamDisplay.right - 18*(nChannels+1) - 2;
        i = pSourceInfo->iDisplayedBufferUsage_0to255;
        if(i<0)i=0; if(i>255)i=255; // i: 0=empty->BLUE, 255=full->RED
        DrawVerticalIndicatorBar(  hDC,        // device context for output
          x, MainWnd_rParamDisplay.top,
          x+17,MainWnd_rParamDisplay.bottom, // drawing area: left,top,right,bottom
          // Color of the indicator (RGB-mix) : empty buffer=BLUE-ish, full=RED
          (i<127)? RGB( 127  , 170      , 255  )  // buffer looks good, BLUE below 50%
                 : RGB( i    , 340-4*i/3, 255-i), // turning red above 50% buffer usage
          RGB(0xFF,0xFF,0xFF),  // Color of the background (RGB-mix)
          (float)i/2.55,        // percentage value (0..100) ,
          "bu" );               // caption, here "bu"=buffer
       } // end if( fShowBargraphs )

      SelectObject(hDC,hOldFont);
      DeleteObject(hNormalFont);
      DeleteObject(hBoldFont);
   }

  EndPaint(hWnd,&ps);
  return 0L;
} // end OnWmPaint


//---------------------------------------------------------------------------
LONG OnWmCommand(HWND hWnd,
           WORD wNotifyCode, // notification code
           WORD wID,         // item, control, or accelerator identifier
           HWND hwndCtl )    // handle of control
{ // reaction to WM_COMMAND message
  (void)hWnd;  // 'unused parameter' - so what..
  (void)wID;

  // Edit field events...
  if(hwndCtl==MainWnd_hEditNcoFreq)
     {  OnNcoEdit(wNotifyCode);      return 0L; }
  if(hwndCtl==MainWnd_hEditTestFreq)
     {  OnTestFreqEdit(wNotifyCode); return 0L; }

  // Button- and combo-list events...
  switch( wID )
   {
     case ID_COMBO_DISP_MODE:
       return 0L;
     default:
       break;
   }

  return 0L;
} // end OnWmCommand()

//---------------------------------------------------------------------------
LONG OnWmSysCommand(HWND hWnd,
       WORD uCmdType,     // (wParam)         type of system command requested
       WORD xPos,         // (LOWORD(lParam)) horizontal postion, in screen coordinates
       WORD yPos )        // (HIWORD(lParam)) vertical postion, in screen coordinates
{ // reaction to WM_SYSCOMMAND message
  T_ControlPanelInfo *pCPInfo = GetControlPanelInfoForGUI(); // ptr always valid
  (void)hWnd; (void)xPos; (void)yPos;
  switch(uCmdType)
   {
    case IDM_SHOW_ABOUT_BOX  :
         // ex: WA_About( g_hwndControlPanel );
         // CAUTION: This "About"-box, implemented in the DLL, is usually
         //    a normal 'blocking' windoze message box [ MessageBox() ] .
         //    But when this 'about' box was opened from a BUTTON in the DLL's
         //    non-modal control panel, the 'about' box remained open while
         //    the application unloaded the DLL  ->   BANG !
         //    There is no easy way to close such a dialog programmatically
         //    (automatically) when unloading the DLL.
         if( g_hwndControlPanelsParent != NULL )
          { // Block the PARENT WINDOW while the 'About' box is displayed;
            // otherwise it may unload the DLL while this stupid 'MessageBox' is open:
            WA_About( g_hwndControlPanelsParent );
          }
         else if( g_hwndControlPanel != NULL )
          { WA_About( g_hwndControlPanel );
          }
         return 0;

    case IDM_CLEAR_UNUSED:   // clears unused (i.e. "currently closed") entries
                             // in the table of audio stream writers & readers
         RedrawWindow( hWnd, NULL, NULL, RDW_INVALIDATE );
         return 0L;
    case IDM_SHOW_CONNECTIONS:
         pCPInfo->iDisplayMode = DISPLAY_MODE_SHOW_CONNECTIONS;
         return 0L;
    case IDM_SHOW_CHUNK_INFO :
         pCPInfo->iDisplayMode = DISPLAY_MODE_SHOW_CHUNK_INFO;
         return 0L;
    case IDM_SHOW_WAVEFORM   :
         pCPInfo->iDisplayMode = DISPLAY_MODE_SHOW_WAVEFORM;
         return 0L;
    case IDM_SHOW_SPECTRUM   :
         pCPInfo->iDisplayMode = DISPLAY_MODE_SHOW_SPECTRUM;
         return 0L;
    case IDM_SHOW_EVT_HISTORY:
         pCPInfo->iDisplayMode = DISPLAY_MODE_SHOW_EVT_HISTORY;
         return 0L;
    case IDM_CLEAR_EVT_HISTORY:
         ClearEventHistory();
         return 0L;

#if( SWI_ASIO_SUPPORTED )
    case IDM_ASIO_ENABLED:    // toggle ASIO-emulation ON/OFF
         return 0L;
    case IDM_ASIO_SEL_STREAM: // select which of the streams shall be used for ASIO-I/O
         pCPInfo->iDisplayMode = DISPLAY_MODE_SHOW_CONNECTIONS;
         pCPInfo->iMouseMode   = MOUSE_MODE_SEL_ASIO_STREAM;
         return 0L;
    case IDM_ASIO_REGISTER  : // register this Audio-I/O-DLL as an ASIO-driver
         // Not sure if this is a full replacement of "regsvr32.exe in_AudioIO.dll" !
         // regsvr32.exe DOES call DllRegisterServer(),
         //        but it possibly does more than this:
         if( DllRegisterServer() == S_OK ) // implemented in ASIO_dummy.c
          { if( ! fWaitingInMessageBox )
             { fWaitingInMessageBox = TRUE;
               MessageBox( NULL, // handle of owner window : UNKNOWN ?
                           "Successfully registered this AUDIO-I/O-DLL\n"
                           "as an ASIO driver.",
                  MY_DLL_NAME, // title of message box
                  MB_OK );     // style of message box
               fWaitingInMessageBox = FALSE;
             }
          }
         return 0L;
    case IDM_ASIO_UNREGISTER: // unregister this Audio-I/O-DLL as an ASIO-driver
         // Not sure if this is a full replacement of "regsvr32.exe -u in_AudioIO.dll" !
         // regsvr32.exe -u <DLL-Name>   DOES call DllRegisterServer(),
         //        but it possibly does more than this:
         if( DllUnregisterServer() == S_OK ) // implemented in ASIO_dummy.c
          { if( ! fWaitingInMessageBox )
             { fWaitingInMessageBox = TRUE;
               MessageBox( NULL, // handle of owner window : UNKNOWN ?
                           "Successfully UN-registered this AUDIO-I/O-DLL\n"
                           "as an ASIO driver.",
                  MY_DLL_NAME, // title of message box
                  MB_OK );     // style of message box
               fWaitingInMessageBox = FALSE;
             }
          }
         return 0L;
#endif // SWI_ASIO_SUPPORTED

    default:
         break;
   } // end switch (wParam) for our own additions to the system menu
  return 1L;  // message has not been handled
} // end OnWmSysCommand()



//---------------------------------------------------------------------------
LONG OnWmCreate(HWND hWnd, LPCREATESTRUCT lpcs)
  // Handler for WM_CREATE message.
  // Notes:
  //  * This is low-level Win32 programming in all it's ugliness, but the
  //    the resulting code is compact, requires no frameworks, resources, etc.
{
  HMENU hMenu, hSubMenu;
  RECT rect;
  int  iCtrlWidth, iCtrlHeight, iCtrlTop;
  // float flt;
  char sz40[44];
  T_AudioSourceInfo *pSourceInfo = GetAudioSourceInfoForGUI(); // NEVER returns NULL - pointer ALWAYS valid

   (void)lpcs;

   // Append some line to the "window menu" alias "system menue" alias "control menu".
   // This only works properly in non-VCL applications.
   hMenu = GetSystemMenu( hWnd, FALSE/*revert*/ );
   // GetSystemMenu(hWnd, bRevert) :
   //    hWnd Identifies the window that will own a copy of the window menu.
   // bRevert Specifies the action to be taken. If this parameter is FALSE,
   //         GetSystemMenu returns the handle of the copy of the window menu
   //         currently in use. The copy is initially identical
   //         to the window menu, but it can be modified.
   AppendMenu( hMenu, MF_SEPARATOR, 0, NULL); // separator between the "normal system menu" and our additions
   AppendMenu( hMenu, MF_STRING,  IDM_SHOW_ABOUT_BOX, "About..." );

   // To insert a menu with SUBMENU into this 'system menu'-thing,
   //    1. Create empty menu;  2. Insert items (and submenus);  3. Assign to item being created.
   hSubMenu = CreateMenu();    // << creates a new, "empty" menu; actually the new sub-menu
   AppendMenu( hSubMenu,       // handle to menu to be changed
          MF_STRING,           // UINT uFlags, menu-item flags
          IDM_SHOW_CONNECTIONS, // menu-item identifier or handle of drop-down menu or submenu
          "Show Connections"); // lpNewItem, menu-item content
   AppendMenu( hSubMenu, MF_STRING,  IDM_SHOW_CHUNK_INFO, "Show Chunk Info");
   AppendMenu( hSubMenu, MF_STRING,  IDM_SHOW_WAVEFORM  , "Show Waveform");
   AppendMenu( hSubMenu, MF_STRING,  IDM_SHOW_SPECTRUM  , "Show Spectrum (10 dB / div)");
   AppendMenu( hSubMenu, MF_STRING, IDM_SHOW_EVT_HISTORY, "Show Event History");
   AppendMenu( hSubMenu, MF_SEPARATOR, 0, NULL);
   AppendMenu( hSubMenu, MF_STRING, IDM_CLEAR_EVT_HISTORY, "Clear Event History");
   AppendMenu( hMenu, MF_POPUP/*uFlags*/, (UINT)hSubMenu, "Display Mode");

#if( SWI_ASIO_SUPPORTED ) // defined in SWITCHES.H (project-specific options)
   hSubMenu = CreateMenu();    // << creates a new, "empty" menu; actually the new sub-menu
   AppendMenu( hSubMenu, MF_STRING,  IDM_ASIO_ENABLED   , "Enabled");
   AppendMenu( hSubMenu, MF_STRING,  IDM_ASIO_SEL_STREAM, "Select Stream for ASIO");
   AppendMenu( hSubMenu, MF_STRING,  IDM_ASIO_REGISTER  , "Register this DLL as ASIO driver");
   AppendMenu( hSubMenu, MF_STRING,  IDM_ASIO_UNREGISTER, "Unregister this DLL as ASIO driver");
   AppendMenu( hMenu, MF_POPUP/*uFlags*/,(UINT)hSubMenu,  "ASIO Emulation" );
#endif // SWI_ASIO_SUPPORTED ?

   // Append other items to the 'system menu' (which are not / don't have sub-menus):
   AppendMenu( hMenu, MF_STRING, IDM_CLEAR_UNUSED,  "Clear unused (closed) entries");

   GetClientRect(hWnd,&rect); // retrieve coordinates of a window's client area

   // get the area for the parameter display in the upper part of the window
   MainWnd_rParamDisplay.top = 0; // ex: C_COMBO_HEIGHT;
   MainWnd_rParamDisplay.bottom = rect.bottom - (C_EDIT_HEIGHT + 2*C_VERT_SEPARATION);
   MainWnd_rParamDisplay.left = 5;
   MainWnd_rParamDisplay.right= rect.right-5;

   // Initialize application-specific stuff...
   // some parameters must be known before the visible controls are created !

#if(0)
   // Ex: Create a combo-box in the upper part, which allows the user
   //  to select different 'views' (see OnWmPaint). This eliminates the need
   //  for a tabbed control window (which would be a nightmare to develop
   //  without some graphic user interface editor like Borland C++Builder,
   //  or -better- QT, or "the microsoft visual stuff" ).
   CreateStaticText(hWnd, "Show:", 0/*style*/, 0/*x*/, 0/*y*/, 40/*w*/, C_COMBO_HEIGHT/*h*/ );
   MainWnd_hDisplayModeCombo = CreateComboBox(hWnd, "DisplayMode", CBS_DROPDOWNLIST,
      ID_COMBO_DISP_MODE, // user-defined CONSTANT (suitable for case-lists)
      40/*x*/, 0/*y*/, 120/*width*/, C_COMBO_HEIGHT+200/*height including drop-down list*/);
   AddStringToComboBox( MainWnd_hDisplayModeCombo, "Connections" );
   AddStringToComboBox( MainWnd_hDisplayModeCombo, "Chunk Info"  );
#endif

   // Create edit fields for (software-)"gain" and "test-signal"- frequency :
   iCtrlHeight = C_EDIT_HEIGHT;   // height for EDIT FIELDS on lower part
   iCtrlWidth = (rect.right-C_HORZ_SEPARATION*C_HORZ_EDITS)/C_HORZ_EDITS;
   iCtrlTop   = rect.bottom - C_EDIT_HEIGHT - C_VERT_SEPARATION;
   CreateStaticText(hWnd, "fc/Hz",
      0/*style*/, 2,iCtrlTop+2,C_STATIC_TEXT_WIDTH,C_EDIT_HEIGHT);
   MainWnd_hEditNcoFreq = CreateEditField(hWnd, "0", WS_BORDER, // text,style
      2+C_STATIC_TEXT_WIDTH+(C_HORZ_SEPARATION+iCtrlWidth)*0,iCtrlTop, // x,y
      iCtrlWidth-C_STATIC_TEXT_WIDTH, C_EDIT_HEIGHT);    // width, height
   CreateStaticText(hWnd, "ft/Hz", 0/*style*/,
      2+(C_HORZ_SEPARATION+iCtrlWidth)*1,iCtrlTop+2,
      C_STATIC_TEXT_WIDTH,C_EDIT_HEIGHT);
   sprintf(sz40, "%.1f", (float)pSourceInfo->fltTestGeneratorFrequency );
   MainWnd_hEditTestFreq = CreateEditField(hWnd, sz40, WS_BORDER, // text,style
      2+C_STATIC_TEXT_WIDTH+(C_HORZ_SEPARATION+iCtrlWidth)*1,iCtrlTop, // x,y
      iCtrlWidth-C_STATIC_TEXT_WIDTH, C_EDIT_HEIGHT);    // width, height

   SetTimer( // creates a timer with the specified time-out value
         hWnd,    // handle of window for timer messages
         0,       // timer identifier
         100,     // UINT uElapse = time-out value, in milliseconds
         NULL );  // lpTimerFunc  = address of timer procedure

   return 0L;
} // end OnWmCreate()


//---------------------------------------------------------------------------
void OnWmClose(HWND hWnd)
{
 int i;
 char sz40Section[44];
 T_ControlPanelInfo *pCPInfo = GetControlPanelInfoForGUI(); // ptr always valid

  // If the window is not minimized, save the current window position LATER..
  if((i=GetWindowPosX1(hWnd))>0)
   { pCPInfo->iX1 = i;
   }
  if((i=GetWindowPosY1(hWnd))>0)
   { pCPInfo->iY1 = i;
   }

  // Save the current window state, and other data which is only relevant
  // for the DLL's CONTROL PANEL (but not for the processing & connections) :
  strcpy(sz40Section,"AudioIOControlPanel");
  WriteIntegerToIniFile( (char*)APPL_szIniFileName, sz40Section,
                    "x1"/*key*/, pCPInfo->iX1/*value*/  );
  WriteIntegerToIniFile( (char*)APPL_szIniFileName, sz40Section,
                    "y1"/*key*/, pCPInfo->iY1/*value*/ );
  WriteIntegerToIniFile( (char*)APPL_szIniFileName, sz40Section,
                  "DisplayMode", pCPInfo->iDisplayMode );
  WritePrivateProfileString( sz40Section, "d", "", APPL_szIniFileName);  // separator-dummy


  // If there's a PARENT WINDOW for the control panel,
  // let the parent window know we have been closed :
  if( g_hwndControlPanelsParent != NULL )
   { // The same old question: "Send" or better "Post" the message here ?
     // Because neither SendMessage() nor PostMessage() seemed to work
     //  for a CONSOLE APPLICATION (the "Audio I/O Test Client"),
     // a stupidly simple system of CALLBACK FUNCTIONS is now used
     // instead of the windoze message system .
 // ex: SendMessage(g_hwndControlPanelsParent/*receiver*/,
 //        WM_AUDIO_IO_CONTROL,         /* Msg, here: WM_APP plus something */
 //        WM_AUDIO_IO_FC_PANEL_CLOSED, /* wParam, here: "function code"    */
 //        0 /*lParam*/   );
     if( pCallbackForControlPanel != NULL )
      { // The callback function is declared as below in AudioIO.H :
        // > typedef INT32 _stdcall (*tpCallbackFromDLL)(INT32 i32FunctionCode, INT32 iParam1, INT32 iParam2);
        (*pCallbackForControlPanel)(pCallbackUserDataForControlPanel, WM_AUDIO_IO_FC_PANEL_CLOSED, 0, 0 );
      }
   }

} // end OnWmClose()


/*==============================================================================
       WndProc

   This was once the 'window procedure of a normal windoze application,
   but HERE, in the DLL, there is no 'WinMain' function with a message loop !
   Instead, the "window in the DLL" must use the *APPLICATION'S* message loop.
   In other words, there must not be TWO message loops .
   Trying to circumvent this can turn into a nightmare. Some examples :
   -  http://www.codeproject.com/KB/DLL/hooks.aspx ,  "Hooks and DLLs"

==============================================================================*/
LRESULT CALLBACK _export WndProc(HWND hWnd, UINT message, UINT wParam, LONG lParam)
{
  LRESULT result;
  DOBINI();    // -> DLL_iSourceLine (optional, WATCH this after stopping the debugger)
  switch (message)
   {
    case WM_CREATE: // sent when application requests that a window be created
      return OnWmCreate(hWnd, (LPCREATESTRUCT) lParam);
    case WM_PAINT:  // request to paint a portion of an application's window
      return OnWmPaint(hWnd);
    case WM_COMMAND: // handle button press etc etc etc
      return OnWmCommand(hWnd,
              HIWORD(wParam), // wNotifyCode: notification code
              LOWORD(wParam), // wID: item, control, or accelerator identifier
              (HWND) lParam); // hwndCtl:  handle of control
    case WM_SYSCOMMAND: // handle events from the "system menu" (alias "window menu")
      if (!OnWmSysCommand(hWnd,
              wParam,  // uCmdType = type of system command requested
              LOWORD(lParam),  // xPos = horizontal postion, in screen coordinates
              HIWORD(lParam))) // yPos = vertical postion, in screen coordinates
        return 0L;     // .. this message has been handled
      else
        return DefWindowProc (hWnd, message, wParam, lParam) ;
    case WM_INITMENU: // sent when a menu is about to become active (including the "system" menu)
      UpdateMyMenu();
      return 0L;             // .. this message has been handled

    case WM_CLOSE:
      if( hWnd==g_hwndControlPanel )
       { OnWmClose(hWnd);
         DestroyWindow(hWnd);   // perform wm_destroy
         return 0L;
       }
      break;
    case WM_DESTROY:
  // ex: PostQuitMessage( 0 );  // Is this such a good idea ? What does PostQuitMessage() do ?
      // > The PostQuitMessage function posts a WM_QUIT message to the thread's
      // > message queue and returns immediately; the function simply indicates
      // > to the system that the thread is requesting to quit at some time
      // > in the future.
      // > When the thread retrieves the WM_QUIT message from its message queue,
      // > it should exit its message loop and return control to Windows.
      // > The exit value returned to Windows must be the wParam parameter
      // > of the WM_QUIT message.
      // WB: The thread which receives this messages will, for example,
      //     be SpecLab.
      //  BUT THIS IS NOT WHAT WE WANT: Terminate the APPLICATION !
      //  (because the receiving thread would be the application's MAIN THREAD)
      // Just to make sure that no-one tries to send anything
      // to the (now invalid) DLL control panel's window handle:
      if( hWnd==g_hwndControlPanel )  // is this really "my" window ?
       { g_hwndControlPanel = NULL;  // yet, it is (well, it WAS) ..
#      if( 0 && L_ALLOW_TESTING )
         MessageBeep(-1);   // beep beep - "I'm out of business"
         Sleep(200);        // Test result: got here when closing the panel,
         MessageBeep(-1);   // after it had been opened by winamp .
         // Also got here when winamp closed it's own plugin control window !
#      endif
       }
      return 0L;

    //-------------- optional message handlers ----------------------------
    case WM_TIMER:
      OnWmTimer(hWnd, (WORD)wParam/*wTimerID*/ ); // called on WM_TIMER
      return 0L;
  //case WM_COPYDATA:
  //  return OnWmCopydata( hWnd,  // handle of THIS window
  //               (HWND)wParam,  // handle of SENDING window
  //   (PCOPYDATASTRUCT)lParam);  // pointer to a COPYDATASTRUCT (read-only!)

   }
  DOBINI();
  result = DefWindowProc (hWnd, message, wParam, lParam) ;
  DOBINI();
  return result;
} // end WndProc()



//---------------------------------------------------------------------------
BOOL CheckBeforeWndCreation(void)
   // Called by the application skeleton BEFORE any window is created.
   // May be used to detect if another instance is already running.
   // Return value (evaluated somewhere in WinMain() ):
   //    FALSE = exit program, don't create a window, don't call OnWmCreate
   //    TRUE  = let me run, create the main window, then call OnWmCreate
{
 // LRESULT result;

  // Check if "this" window is already running somewhere in another instance.
  // If so, DO NOT open a second window !
  HWND hWnd=FindWindow( APPL_szWindowClassName/*"SND_INPUT"*/, NULL);
  if(hWnd!=NULL)
   { // oops.. "I" am already running on this system !
     // PostMessage(hWnd, WM_SETFOCUS, 0,0); // no
     ShowWindow( hWnd, SW_SHOWNORMAL );
     BringWindowToTop(hWnd);
     return FALSE; // don't create a window for this instance !
   }

  return TRUE;  // ok, create a window and a message loop for this application
} // end CheckBeforeWndCreation()



//----------------------------------------------------------------------
__declspec( dllexport ) INT32 AIO_ShowControlPanel( // NON-MODAL (not blocking) !
       HWND   hwndParent,            // [in] handle of the caller's main window
                                     //      (may be NULL if the caller is a console application)
       HWND   *phwndControlPanel,    // [out] handle of the DLL's control panel
       INT32  iDisplayOptions,       // AIO_DISPLAY_SHOW_WINDOW, AIO_DISPLAY_HIDE_WINDOW, etc
       void  *pvUserDataForCallback,         // [in,optional] address of some user data passed to the callback
       tpCallbackFromDLL pCallbackFunction ) // [in,optional] address of a callback function, may be NULL
  // Shows the DLL's own "control panel" box as a non-modal (= non-blocking) window.
  // Certain Audio-I/O-DLLs may provide some additonal CONFIGURATION here.
  // Return value: AIO_NO_ERROR, if the control panel could successfully be opened.
  // Notes:
  //  * This is low-level Win32 programming in all it's ugliness, but the
  //    the resulting code is compact, requires no frameworks, resources, etc.
  //  * Don't try to open the control panel multiple times.
  //  * For debugging, see AudioIO.c (="DLL host"), AIO_Host_ShowControlPanel() .
{
  WNDCLASS wndclass ;  // is it ok to use a short-living automatic variable for this ?
  T_ControlPanelInfo *pCPInfo = GetControlPanelInfoForGUI(); // ptr always valid

#if( 0 && L_ALLOW_TESTING )
  MessageBeep(-1);   // TEST: beep - "I should be in business"
       // TEST result: Winamp called this function with each new click
       //   on the 'Configure' button on Winamp's plugin screen.
       //   But the damned DLL's control panel only opened for the FIRST click.
#endif // L_ALLOW_TESTING ?

  if( phwndControlPanel != NULL )
   { *phwndControlPanel = NULL;  // let the caller know we have NO VALID window handle yet
   }
  if( g_pSharedMemoryData==NULL )  // problem during DLL-init ?
   { return AIO_ERROR_INVALID_PARAMETER;
   }


  if( iDisplayOptions & AIO_DISPLAY_SHOW_WINDOW )
   { if( g_hwndControlPanel == NULL )
      { // The 'control window' has not been created yet, so create it NOW .

#if( 0 && L_ALLOW_TESTING )
        MessageBeep(-1);   // TEST: beep - "I should be in business"
        // TEST result: Winamp called this function with each new click
        //   on the 'Configure' button on Winamp's plugin screen.
        //   But the damned DLL's control panel only opened for the FIRST click.
#endif // L_ALLOW_TESTING ?



        // To avoid trouble with "About"-box (see OnHelpButton() ),
        // use the *PARENT'S* window handle for blocking it
        // just as if the PARENT WINDOW had opened the dialog :
        g_hwndControlPanelsParent = hwndParent;

        // Register the control panel's WINDOW CLASS (old-style Win32 API):
        wndclass.style         = CS_HREDRAW | CS_VREDRAW ;
        wndclass.lpfnWndProc   = WndProc ;
        wndclass.cbClsExtra    = 0 ;
        wndclass.cbWndExtra    = 0 ;
        wndclass.hInstance     = APPL_hInstance;
        wndclass.hIcon         = NULL;
        wndclass.hIcon         = LoadIcon(APPL_hInstance,TEXT("PROGRAM_ICON"));
        wndclass.hCursor       = LoadCursor(NULL, IDC_ARROW);
        wndclass.hbrBackground = (HBRUSH)(MY_BACKGROUND_COLOR+1); // don't ask what the "+1" is for !
        wndclass.lpszMenuName  = NULL ;
        wndclass.lpszClassName = APPL_szWindowClassName; // window name, here used as "class name"
        if( ! fRegisteredClassName )   // Only call RegisterClass() when necessary:
         { if( RegisterClass(&wndclass) == 0 )
            { // Couldn't register the window class ! Already registered ? Shit happened, because:
              // > No window classes registered by a DLL are unregistered
              // > when the DLL is unloaded. A DLL must explicitly unregister
              // > its classes when it is unloaded.
              DOBINI();
              // ex: return AIO_GENERAL_ERROR;
              // Not sure if this RegisterClass-thingy is 'system global'.
              // If it is, RegisterClass will fail if the control panel
              // is ALREADY OPENED FOR ANOTHER APPLICATION.
              // That would NOT be an error, so do NOT give up here yet !
            }
         }
        else
         { fRegisteredClassName = TRUE; // don't forget to 'Unregister' later
         }

        // now create the window :
        DOBINI();
        g_hwndControlPanel = CreateWindow(
             APPL_szWindowClassName, // lpClassName,    pointer to registered class name
             APPL_szWindowCaption,   // "lpWindowName", pointer to window name, effectively the TITLE
             // DWORD dwStyle,        window style. Tried a lot ...
             WS_POPUPWINDOW | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX | WS_OVERLAPPED,  // normal window
             // WS_POPUP | WS_DLGFRAME ,   // window with no title, quite unusual as main window
             // WS_POPUPWINDOW | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX | WS_SIZEBOX, // sizeable
             // WS_CAPTION,   // with title, but no icon and no system menu nor close symbol
             pCPInfo->iX1, pCPInfo->iY1, // initial position of the window
             240,            // window width
             180,            // window height
             hwndParent,     // HWND hWndParent, handle to parent or owner window
             NULL,            // HMENU hMenu,     handle to menu or child-window identifier
             APPL_hInstance,  // HANDLE hInstance, handle to application instance
             NULL);           // LPVOID lpParam,   pointer to window-creation data
        // Note: CreateWindow() already calls the window procedure a few times,
        //       even if there is no 'message pump' anywhere in the application .
        DOBINI();
      }
     if( g_hwndControlPanel != NULL ) // CreateWindow() was successfull ...
      {
        if( phwndControlPanel != NULL ) // let the caller know the window handle
         { *phwndControlPanel = g_hwndControlPanel;
         }
        pCallbackUserDataForControlPanel = pvUserDataForCallback;
        pCallbackForControlPanel = pCallbackFunction;

        // The 'control window' seems to exist now; make it visible:
        DOBINI();
        ShowWindow( g_hwndControlPanel, SW_SHOWNORMAL );
        // Note: ShowWindow() also calls the window procedure a dozen times,
        //       even if there is no 'message pump' anywhere in the application .
        //       But without a 'message pump' the window remains empty (gray),
        //       and doesn't respond to user input. This happened 2011-07-05,
        //       see details in c:\cbproj\AudioIO\AudioIO_Test_Client.cpp !
        DOBINI();  // 2011-07-05: successfully got here, but the window was "unresponsive"
                   //  Fixed by adding a small 'message pump' in the test client.
      }
     else  // CreateWindow() failed, why ?
      { return AIO_GENERAL_ERROR;
      }
   }
  else // don't SHOW but HIDE the window:
   { if( g_hwndControlPanel != NULL )
      { DOBINI();
        ShowWindow( g_hwndControlPanel, SW_HIDE );
        DOBINI();
      }
   }
  return AIO_NO_ERROR;
} // end AIO_ShowControlPanel()




// EOF <aio2winamp.cpp>





